########################################################################################################
# Script to import Mpmas output into R (or transform it into more readable text files using R) 
# automatically generated by mpmasql
#
# (c)  Christian Troost, Universität Hohenheim, 2012
#######################################################################################################
# Part I: Settings
# These settings might be adapted by the user


#Directories
dir_root<-"./"
dir_out<-"./out/"
dir_lab<-"./out/"
dir_file_out<-"./R/data/"

#Scenarios & periods to import
prefix = "example"
scenarios<-c("_")
periods<-c(1)

#What to do with them
keep_in_workspace<-TRUE
save_as_tab<-FALSE
lean<-TRUE
join_raw<-TRUE
join_group<-TRUE
nogroup<-FALSE
filter<-FALSE
filter_vars<-c(1)

#
########################################################################################################
# Part II:  The actual script 

#Constants
header_all<-c("DataType","AgentID", "CatchmentID", "SectorID","FarmsteadID","PopulationID","ClusterID","NetworkID","SegmentID","TypeOfExpectations","ColorInMaps","Xcoordinate","Ycoordinate","NumberOfPlots")
header_end_all<-c("period","scenario","exit")
p_header<-c("TotalGrossMargin", "Cashflow", "Income",
					"ValueAdded", "EquityCapital", "PaymentToLand",
					"RelativeFactorPayments", "OnFarmLabor", "OnFarmCapital",
					"InterestsPaid", "DistanceCosts",
					"AdditionalTransportCosts", "WithdrawalsForConsumption",
					"LiquidMeans", "MigrationDecision",
					"LastLpSolved", "FixCosts",
					"TranspCosts", "Deposits", "Depreciation",
					"LiqMeansUsedForInvestment", "PrincipalPaid")
notfirst14 <-array(,dim=14)
notfirst14[1:14]<-FALSE
notheaderend<-c(FALSE, FALSE, FALSE)

#Make output directory unless it exists
if(save_as_tab) {
	dir.create(paste(dir_file_out,"tab/",prefix,"/", sep=""), recursive=TRUE)
}
if(keep_in_workspace) {
	dir.create(paste(dir_file_out,"workspaces/", sep=""), recursive=TRUE)
}
#auxiliary flag
first_period<-TRUE

#loop over scenarios
for (s in scenarios) {
#loop over periods
	for (p in periods) {

##Read files and split by content type
### k
		a<-file.info(paste(dir_out, prefix,s,"_k", sprintf("%02d",p),".out", sep = ""))
		if (! is.na(a$size) && a$size > 0) {
			tmp<-read.table(paste(dir_out, prefix,s,"_k", sprintf("%02d",p),".out", sep = ""))
			tmp$period = p
			tmp$scenario = s
			tmp$exit = FALSE
			rhs_<-tmp[ tmp[1] == 4, ]
			lhs_<-tmp[ tmp[1] == 5, ]
			rm(tmp)
		}
		else {
			rhs_ <- NULL
			lhs_ <- NULL
		}
### xk
		a<-file.info(paste(dir_out, prefix,s,"_xk", sprintf("%02d",p),".out", sep = ""))
		if (! is.na(a$size) && a$size > 0) {
			tmp<-read.table(paste(dir_out, prefix,s,"_xk", sprintf("%02d",p),".out", sep = ""))
			tmp$period = p
			tmp$scenario =  s
			tmp$exit = TRUE
			rhsx_<-tmp[ tmp[1] == 4, ]
			lhsx_<-tmp[ tmp[1] == 5, ]
			rm(tmp)
		}
		else {
			rhsx_ <-NULL
			lhsx_ <- NULL
		}
### u		
		a<-file.info(paste(dir_out, prefix,s,"_u", sprintf("%02d",p),".out", sep = ""))
		if (! is.na(a$size) && a$size > 0) {
			tmp<-read.table(paste(dir_out, prefix,s,"_u", sprintf("%02d",p),".out", sep = ""))
			tmp$period = p
			tmp$scenario = s
			tmp$exit = FALSE
			sol_<-tmp[ tmp[1] == 2, ]
			prices_<-tmp[ tmp[1] == 3, ]
			rm(tmp)
		}
		else {
			sol_ <- NULL
			prices_ <- NULL
		}
### xu
		a<-file.info(paste(dir_out, prefix,s,"_xu", sprintf("%02d",p),".out", sep = ""))
		if (! is.na(a$size) && a$size > 0) {
			tmp<-read.table(paste(dir_out, prefix,s,"_xu", sprintf("%02d",p),".out", sep = ""))
			tmp$period = p
			tmp$scenario =  s
			tmp$exit = TRUE
			solx_<-tmp[ tmp[1] == 2, ]
			pricesx_<-tmp[ tmp[1] == 3, ]
			rm(tmp)
		}
		else {
			solx_ <-NULL
			pricesx_ <- NULL
		}		
### p
		a<-file.info(paste(dir_out, prefix,s,"_p", sprintf("%02d",p),".out", sep = ""))
		if (! is.na(a$size) && a$size > 0) {
			tmp<-read.table(paste(dir_out, prefix,s,"_p", sprintf("%02d",p),".out", sep = ""))
			tmp$period = p
			tmp$scenario = s
			tmp$exit = FALSE
			perf_<-tmp[ tmp[1] == 1, ]
			rm(tmp)
		}
		else {
			perf_ <- NULL
		}
### xp
		a<-file.info(paste(dir_out, prefix,s,"_xp", sprintf("%02d",p),".out", sep = ""))
		if (! is.na(a$size) && a$size > 0) {
			tmp<-read.table(paste(dir_out, prefix,s,"_xp", sprintf("%02d",p),".out", sep = ""))
			tmp$period = p
			tmp$scenario =  s
			tmp$exit = TRUE
			perfx_<-tmp[ tmp[1] == 1, ]
			rm(tmp)
		}
		else {
			perfx_ <-NULL
		}


#Merge with other periods
		if(first_period) {
			rhs<-rbind(rhs_, rhsx_)
			lhs<-rbind(lhs_, lhsx_)
			solution<-rbind(sol_, solx_)
			prices<-rbind(prices_, pricesx_)
			performance<-rbind(perf_, perfx_)
			first_period<-FALSE
		}	
		else {	
			rhs<-rbind(rhs, rhs_, rhsx_)
			lhs<-rbind(lhs, lhs_, lhsx_)
			solution<-rbind(solution, sol_, solx_)
			prices<-rbind(prices, prices_, pricesx_)
			performance<-rbind(performance, perf_, perfx_)
		}
		rm(rhs_, lhs_, rhsx_, lhsx_, a, prices_, pricesx_, sol_, solx_, perf_, perfx_)

	}
#Set column names
	constraints<-read.table(paste(dir_lab, prefix,s,"_ConstraintInfo.txt", sep = ""), header=TRUE, sep="	", quote="")
	nonzero_constraints<-constraints[constraints$MpmasqlType != "MASzero" &    constraints$MpmasqlType != "MASforget" ,]

	activities<-read.table(paste(dir_lab, prefix,s,"_ActivityInfo.txt", sep = ""), header=TRUE, sep="	", quote="")
	nonzero_activities<-activities[activities$MpmasqlType != "MASzero" &    activities$MpmasqlType != "MASfuture" ,]

	names(performance)<-c(header_all, p_header, header_end_all)
	names(solution)<-c(header_all, as.character(nonzero_activities[, "Code"]), header_end_all)
	names(prices)<-c(header_all, lapply(as.character(nonzero_activities[, "Code"]), function(x) paste("prices_",x, sep="")), header_end_all)
	names(rhs)<-c(header_all, as.character(nonzero_constraints[, "Code"]), header_end_all)
	names(lhs)<-c(header_all, lapply(as.character(nonzero_constraints[, "Code"]), function(x) paste("lhs_",x, sep="")), header_end_all)


	info_block<-performance[,c("FarmsteadID" ,header_end_all)]
	if (join_group) {
		userset<-performance
	}
#Column aggregation
	if (!nogroup) {
#Aggregate columns and create result variable datasets for constraints
		cgroups<-names(constraints)
		for (g in cgroups[6: (length(cgroups) -1 ) ] ) {
			res<-as.data.frame(sapply( unique(nonzero_constraints[nonzero_constraints[[g]] != "-1",g ]), function (t) rowSums(as.matrix(rhs[, c(notfirst14, nonzero_constraints[[g]] == as.character(t), notheaderend ) ] ))))
			names(res)<-lapply( unique(nonzero_constraints[nonzero_constraints[[g]] != "-1", g]), function(x) paste(g, "_",x, sep="")) 

			if(join_group) {
				userset<-cbind(userset, res)
			}

			res<-cbind(info_block,res)

			if (keep_in_workspace) {
				assign(paste(s, "_rhs_", g, sep=""), res)
			}
			if (save_as_tab) {
				write.table(res, paste(dir_file_out,"tab/",prefix,"/",s,"_rhs_",g,".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)	
			}
			
			rm(res)

			res<-as.data.frame(sapply( unique(nonzero_constraints[nonzero_constraints[[g]] != "-1",g]), function (t) rowSums(as.matrix(rhs[, c(notfirst14, nonzero_constraints[[g]] == as.character(t), notheaderend) ] ))))
			names(res)<-lapply(unique(nonzero_constraints[nonzero_constraints[[g]] != "-1",g]), function(x) paste(g, "_lhs_",x, sep="")) 

			if(join_group) {
				userset<-cbind(userset, res)
			}

			res<-cbind(info_block,res)
			if (keep_in_workspace) {
				assign(paste(s, "_lhs_", g, sep=""), res)
			}
			if (save_as_tab) {
				write.table(res, paste(dir_file_out,"tab/",prefix,"/",s,"_lhs_",g,".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)	
			}
			rm(res)

		}

#Aggregate columns and create result variable datasets for activities
		agroups<-names(activities)
		for (g in agroups[6: (length(agroups)) ] ) {
			res<-as.data.frame(sapply( unique(nonzero_activities[nonzero_activities[[g]] != "-1",g]), function (t) rowSums(as.matrix(solution[, c(notfirst14, nonzero_activities[[g]] == as.character(t), notheaderend) ] ))))
			names(res)<-lapply(unique(nonzero_activities[nonzero_activities[[g]] != "-1",g]), function(x) paste(g, "_",x, sep="")) 
		
			if(join_group) {
				userset<-cbind(userset, res)
			}

			res<-cbind(info_block,res)
			if (keep_in_workspace) {
				assign(paste(s, "_solution_", g, sep=""), res)
			}
			if (save_as_tab) {
				write.table(res, paste(dir_file_out,"tab/",prefix,"/",s,"_solution_",g,".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)								
			}
			rm(res)
		}
	}

#Make a combined dataset of raw data
	if (join_raw || (lean && !join_group)) {
		raw_joint<-cbind(performance, solution, rhs)
		if (keep_in_workspace) {
			assign(paste(s, "_combined", sep=""), raw_joint)
		}
		if (save_as_tab) {
			write.table(raw_joint, paste(dir_file_out,"tab/",prefix,"/",s,"_combined",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)								
		}
	}
	if (join_group) {
		if (keep_in_workspace) {
			assign(paste(s, "_userset", sep=""), userset)
		}
		if (save_as_tab) {
			write.table(userset, paste(dir_file_out,"tab/",prefix,"/",s,"_userset",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)								
		}		
	}
	if (join_group && filter) {
		if (keep_in_workspace) {
			assign(paste(s, "_userset_filtered", sep=""), userset[, c(header_end_all, filter_vars)])
		}
		if (save_as_tab) {
			write.table(userset[, c(header_end_all, filter_vars)], paste(dir_file_out,"tab/",prefix,"/",s,"_userset_filtered",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)								
		}

	}

#Attach scenario prefix to raw datasets, in case the user chose to keep them in the workspace
	if (keep_in_workspace && !lean) {
		assign(paste(s, "_rhs", sep=""), rhs)
		assign(paste(s, "_solution", sep=""), solution)
		assign(paste(s, "_lhs", sep=""), lhs)
		assign(paste(s, "_performance", sep=""), performance)
		assign(paste(s, "_orow", sep=""), prices)
		assign(paste(s, "_nz_constraints", sep=""), nonzero_constraints)
		assign(paste(s, "_nz_activities", sep=""), nonzero_activities)
	}

#Save raw datasets as tab-separated text files if the user requests this
	if (save_as_tab && !lean) {
		write.table(rhs, paste(dir_file_out,"tab/",prefix,"/",s,"_rhs",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)				
		write.table(lhs, paste(dir_file_out,"tab/",prefix,"/",s,"_lhs",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)				
		write.table(solution, paste(dir_file_out,"tab/",prefix,"/",s,"_solution",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)				
		write.table(performance, paste(dir_file_out,"tab/",sprefix,"/",,"_performance",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)				
		write.table(prices, paste(dir_file_out,"tab/",prefix,"/",s,"_orow",".txt", sep=""), sep="	", quote=FALSE, row.names=FALSE)				
	}



#Remove temporary objects
	rm(rhs, lhs, solution, performance, prices, constraints, activities, agroups,cgroups, g, userset, raw_joint, info_block, nonzero_activities, nonzero_constraints)
#Set firstperiod to TRUE
	first_period<-TRUE
}

#Remove auxiliary objects
rm(s,p,first_period)
rm(save_as_tab,dir_lab, dir_out, dir_root, join_raw, join_group, lean, nogroup, header_all, header_end_all, p_header)

#Save workspace
if (keep_in_workspace) {
    save.image(paste(dir_file_out,"workspaces/",prefix, ".RData", sep=""))
}
rm(keep_in_workspace, dir_file_out)
###### end of script ######################################################################################################################################################
