#!/usr/bin/perl
use strict;
use warnings;
use lib qw(/usr/local/lib/mpmasql/modules);

use Statistics::R;
use IndexedTable;
my $verbose = 0;
my $infile;
for my $element (@ARGV) {
	if (not $element =~ /^-/) {
		# statement file name
		$infile = $element
	}
}

open (my $INI,"<", $infile) or die "Error when trying to read  $infile\n $!\n";
system(`mkdir -p /tmp/mpmasql`);
my $r_temp = "/tmp/mpmasql/toR.tab";

my $solutions;
my $reference;
my $variablelist;
my $scenarioinfo;
my @comparisontable;
my $temp;
my $R = Statistics::R->new() ;

my $status = "n";
my $oldstatus = "n";
my %omitted;
while ( <$INI>) {
	chomp;
	my $line = $_;
	next if ($line =~ /^\s*$/);
	next if ($line =~ /^'/);
	next if ($status eq "c" and not $line =~ /\s*end\s+comment/);
	
	if ($line =~ /\s*end\s+(\w+)/) {
		if ($1 eq "comment") {
			$status = $oldstatus;
		}
	}
	elsif ($line =~ /\s*start\s+(\w+)/) {
		if ($1 eq "comment") {
			$oldstatus = $status;
			$status = "c",

		}
	}
	
	print "$line\n";
	if ($line  =~ /^\s*compare\s*"([^"]+)"\s*with\s*"([^"]+)"\s*using\s+variablelist\s*"([^"]+)"(\s*,.*)?/  ) {
			$solutions = IndexedTable->create_from_ascii($1,3,"solution",'\t',0);
			$reference = IndexedTable->create_from_ascii($2,2,"reference",'\t',1);
			$variablelist = IndexedTable->create_from_ascii($3,1,"variables",'\t',1);
			my $options = $4;
			 undef %omitted;
			my @comparisontable = undef;
			if (defined $options and $options =~ /scenario_factors\s*\(\s*"([^"]+)"\s*\)/) {
				$scenarioinfo = IndexedTable->create_from_ascii($1,2,"sceninfo",'\t',1);
			}
			open ($temp, ">", $r_temp) || die"Could not open temp file: $!\n";
			print $temp join("\t", ("scenario","id","variable","result","observed"))."\n";
			for my $s ($solutions->query([],0) ) {
				for my $o ($solutions->query([$s],0) ){
					my $divisor_sol = 1;
					my $divisor_ref = 1;
					if (defined $options and $options =~ /shares/) {
						$divisor_sol = 0;
						$divisor_ref = 0;
						for my $v ($variablelist->query([],0) ) {
							$divisor_sol += $solutions->query([$s, $o, $v],4) ||0;
							$divisor_ref += $reference->query([$o, $v],3) ||0;

						}
						if ($divisor_sol == 0) {
							print "ID $o has sum of zero in solution\n" if ($verbose);
							$omitted{$o} = 1;
							next ;
						}
						if ($divisor_ref == 0) {
							print "ID $o has sum of zero in reference\n" if ($verbose);
							$omitted{$o} = 1;
							next ;
						}
					}
					
					for my $v ($variablelist->query([],0) ) {
						my $sol = $solutions->query([$s, $o, $v],4) ||0;
						my $ref = $reference->query([$o, $v],3) ||0;
						push @comparisontable, [($s,$o,$v,$sol/$divisor_sol ,$ref/$divisor_ref)];
						print $temp join("\t", ($s,$o,$v,$sol/$divisor_sol,$ref/$divisor_ref))."\n";
					}	
				}
			}
			close $temp;
			if (defined $options and $options =~ /save\s*\("([^"]+)"\)/) {

				system("cp \"$r_temp\" \"$1\" ");#"
			}
	}
	elsif ($line =~ /^\s*show\s+fit\s+by\s+(\w+)(\s*,.*)?/ ) {
		my $by = $1;
		my $options = $2;
		#print "Optional output: $options\n";

		my $omitstring = " ";
		if (defined $options && $options =~ /omit\s*\(([\w\s,]+)\)/) {

			my @vars_to_omit = split /\s*,\s*/, $1;
			for my $i (@vars_to_omit) {
				$omitstring .= " & comtab\$variable != '$i'"
			}
		}
		$R->startR ;  

		$R->send("library(hydroGOF)");
		$R->send(qq`comtab <- read.table("$r_temp", header = TRUE)` );
		if ($by eq "scenario") {
			my %scenfit;
			print "scenario\tshcorrect\tnse\trsr\tpbias\trmse\tolsni_slope\tolsni_r2\tR2\tbR2\td\n";
			for my $scen (sort { $a cmp $b } $solutions->query([],0) ) {
				my ($coef, $r2, $is_above_cutoff) = calib_ols_noconst($R, "comtab[comtab\$scenario == \"$scen\" $omitstring,])", -1 );
				my %gof = gof($R, "comtab\$scenario == \"$scen\" $omitstring");
				my $shccp = share_of_correctly_classified_plots($R, "comtab\$scenario == \"$scen\" $omitstring");
				#print %gof;
				print "$scen\t$shccp\t$gof{NSeff}\t$gof{RSR}\t$gof{PBIAS}\t$gof{RMSE}\t$coef\t$r2\t$gof{R2}\t$gof{bR2}\t$gof{d}\n";
				$scenfit{$scen}->{shcorrect} = 	$shccp;			
				$scenfit{$scen}->{nse} = 	$gof{NSeff};
				$scenfit{$scen}->{olsni_slope} = 	$coef;
			}
			$R->stopR() ;
			
			if (defined $options and $options =~ /tabforsens\s*\(\s*([\w]+)\s*,\s*"([^"]+)"\s*\)/) {
				my $yvar = $1;	
				my $fname = $2;
				print "Saving tabforsens to $fname\n";
				my $cls = 0;
				my %varindex;
				my @vartab;
				my @yvec;
				for my $s ($solutions->query([],0)) {
					my @row;
					push @yvec, $scenfit{$s}->{$yvar};
					#push @yvec, $s;
					for my $v( $scenarioinfo->query([$s],0) ){
						
						$varindex{$v} = $cls++ if (not exists $varindex{$v});
						$row[$varindex{$v}] = $scenarioinfo->query([$s, $v],3);

					}
					push @vartab, \@row;
				}			
				open ($temp, ">", $fname) || die"Could not open temp file: $!\n";
				print $temp "Scenario\t".join("\t", sort {$varindex{$a} <=> $varindex{$b} } keys %varindex)."\n";
				for my $j (0..$#vartab) {
					print $temp "$yvec[$j]\t";
					#print $temp "$s\t";
					foreach my $k (@{$vartab[$j]}) {
						print $temp (defined $k ?$k : "NA")."\t"; 
						
					}
					print $temp "\n";
				}
				close($temp);
				print "done\n";
				#$R->send("library(sensitivity)");
				#$R->send(qq` scenvars<- read.table("$r_temp", header = TRUE)` );
				#print $R->read();
				#for my $v (keys %varindex) {
				#	$R->send("min(scenvars\$$v) == max(scenvars\$$v)");
				#	$R->send("scenvars\$$v<- NULL") if ($R->read() =~ /TRUE/);
				#}
				#$R->send(qq`y <- c(`. join(",",@yvec).")");
				#print $R->read();
				#$R->send(qq`src(scenvars,y, nboot = 100)`);
				#print $R->read();

			}



		}
		if ($by eq "id") {
			my $cutoff = -1;
			my $nsec = -1;			
			if (defined $options and $options =~ /\%_coef_dev_less_than\s*\(\s*([\d.-]+)\s*\)/) {
				$cutoff = $1;	
			}
			if (defined $options and $options =~ /\%_nse_greater\s*\(\s*([\d.-]+)\s*\)/) {
				$nsec = $1;	
			}
			my $onlysummary = (defined $options  && $options=~ /summarize/);
			print "scenario\tid\tshcorrect\tnse\trsr\tpbias\trmse\tolsni_slope\tolsni_r2\tR2\tbR2\td\n";
			for my $scen (sort { $a cmp $b } $solutions->query([],0) ) {
				my $id_above_cutoff = 0;
				my $nse_above_cutoff = 0;
				my $count = 0;
				for my $id ( sort { $a <=> $b } $solutions->query([$scen],0) ) {
						next if (exists $omitted{$id});
						$count++;
						my ( $coef, $r2, $is_above_cutoff) = calib_ols_noconst($R, qq`comtab[comtab\$scenario == "$scen" & comtab\$id == "$id" $omitstring,])`, $cutoff );
						$id_above_cutoff += $is_above_cutoff;

						my %gof = gof($R, "comtab\$scenario == \"$scen\" & comtab\$id == \"$id\" $omitstring");
						#print %gof;
						my $shccp = share_of_correctly_classified_plots($R, "comtab\$scenario == \"$scen\" & comtab\$id == \"$id\" $omitstring");
						$nse_above_cutoff++ if ($nsec > -1 and not $gof{NSeff} eq "NA" and $gof{NSeff} >  $nsec);
						unless ($onlysummary) {
							print "$scen\t$id\t$shccp\t$gof{NSeff}\t$gof{RSR}\t$gof{PBIAS}\t$gof{RMSE}\t$coef\t$r2\t$gof{R2}\t$gof{bR2}\t$gof{d}\n ";	
						}
				}
					
				print "Coef summary for $scen\t". sprintf("%.1f", ($id_above_cutoff / $count) * 100 )."\% deviate less than $cutoff from 1\n" if ($cutoff > -1);
				print "NSE summary for $scen\t". sprintf("%.1f", ($nse_above_cutoff / $count) * 100 )."\%  above $nsec \n" if ($nsec > -1);
			}
			
			#$R->stopR() ;
		}
		if ($by eq "variable") {
			my $cutoff = -1;
			
			print "scenario\tvariable\tshcorrect\tnse\trsr\tpbias\trmse\tolsni_slope\tolsni_r2\tR2\tbR2\td\n";
		
			for my $scen (sort { $a cmp $b } $solutions->query([],0) ) {
				for my $v ( sort { $a cmp $b } $variablelist->query([],0)  ) {
					my ( $coef, $r2, $is_above_cutoff) = calib_ols_noconst($R, qq`comtab[comtab\$scenario == "$scen" & comtab\$variable == "$v" ,])`, $cutoff );
					my %gof = gof($R, "comtab\$scenario == \"$scen\" & comtab\$variable == \"$v\" $omitstring");
					my $shccp = share_of_correctly_classified_plots($R, "comtab\$scenario == \"$scen\" & comtab\$variable == \"$v\" $omitstring");
					#print %gof;
					print "$scen\t$v\t$shccp\t$gof{NSeff}\t$gof{RSR}\t$gof{PBIAS}\t$gof{RMSE}\t$coef\t$r2\t$gof{R2}\t$gof{bR2}\t$gof{d}\n ";
				}
			}
					
	
			
			
			
		}
		$R->stopR() ;
	}
}
#$R->stopR() ;
close ($INI);

sub calib_ols_noconst {
	my $R = shift;
	my $regstring = shift,
	my $cutoff = shift;
	$R->send("options(warn = -1)");
	$R->send("reg <- lm('result ~ 0 + observed',". $regstring);
	$R->send("print(coef(reg))") ;
	my @ll = split /\n/, $R->read();
	my $coef = "NA";
	my $is_above_cutoff = 0;
	if (defined $ll[1]) {
		$ll[1] =~ s/([\d.]+)/$1/;
		$coef = $ll[1];
		
		$is_above_cutoff = 1 if ($cutoff >= 0 && $ll[1] =~ /\d/ && ($ll[1] > 1 - $cutoff and $ll[1] < 1 + $cutoff ) );
	}
	$R->send(qq`print( summary(reg))`) ;
	@ll = split /\n/, $R->read();;
	my $r2;
	foreach (@ll) {
		next unless ($_ =~ /Multiple R-squared:\s*([\d.-]+),\s*Adjusted R-squared:\s*([\d.-]+)/);
		$r2 = $1;

	}
	$R->send("remove(reg)");
	$r2 = "NA" unless (defined $r2);
	$coef = "NA" unless (defined $coef);
	return  $coef, $r2, $is_above_cutoff;
}
sub gof {
	my $R = shift;
	my $selectstring = shift,

	
	print $R->read();
	$R->send("options(warn = -1)");
	$R->send("print( max(comtab[$selectstring,\"observed\"]) == min(comtab[$selectstring,\"observed\"]))");
	my %gof;
	unless ($R->read() =~ /TRUE/) {
		$R->send("gof <-gof(comtab[$selectstring,\"result\"],comtab[$selectstring,\"observed\"])" );
		print $R->read();
		$R->send("print(gof)");
		my @ll = split /\n/, $R->read();
		shift @ll; #throw away [,1] header
	
		foreach (@ll) {
			$_ =~ s/\%//;
			my ($key,$val) = split /\s+/, $_;
			#print "$key\t$val\n";
			$gof{$key} = $val;

		}
	}
	foreach my $k (qw(NSeff RSR PBIAS RMSE R2 bR2 d)) {
		$gof{$k} = "NA" if (not defined $gof{$k});
	}
	return  %gof;
}
sub share_of_correctly_classified_plots {
	my $R = shift;
	my $selectstring = shift,
	$R->send("options(warn = -1)");
	$R->send("print(sum(comtab[$selectstring,\"observed\"]))");
	my @ll;
	unless ($R->read() =~ /^\s*\[1\]\s+0[^\d]*$/) {
		$R->send("print(sum(pmin(round(comtab[$selectstring,\"result\"]) , comtab[$selectstring,\"observed\"])) /sum(comtab[$selectstring,\"observed\"]))");
		@ll = split /\s+/, $R->read();
	}
	else {
		$ll[1] = "NA";
	}
	return ($ll[1] eq "NA" ?"NA": sprintf("%.3f", $ll[1]) );
}
