#!/usr/bin/perl
use strict;
use warnings;
use lib qw(/usr/local/lib/mpmasql/modules );
use IndexedTable;
use SimpleObjectRegistry;
use Storable qw/dclone/;
use Term::ReadLine;
use Term::ReadKey;
use MpmasResults;


##Version:
my $thisversion = "0.3 alpha";
my $lastchange = "25-May-2011";
my $longwelcome = <<WC;
####################################################################
####		MP-MAS Result File Explorer $thisversion	####
####								####
####	(c)	Christian Troost, Univ. Hohenheim, 2010-2011	####
####	Version: $thisversion		$lastchange		####
####################################################################
WC

my $shortwelcome = <<WC2;
MP-MAS-QL Result File Explorer $thisversion		
(c)	Christian Troost, Universität Hohenheim, 2009-2011
WC2

my $batchloop = 0;
my $shelloop = 0;
my $ifh;
my $curDir = "./";
my $term ;
#Parse commandline arguments
while (@ARGV) {
	my $el = shift @ARGV;
	if  ($el =~ /^-v$/ ) {
		print $longwelcome;
		exit 0;
	}
	elsif  ($el =~ /^-b$/ ) {
		$batchloop = 1;
		my $ifn = shift @ARGV;
		if (defined $ifn and not $ifn =~ /^-/) {
			open ($ifh, "<", $ifn) or (print "Error when trying to open $ifn: $!\n" and exit (1));
		}
		else {
			unshift( @ARGV, $ifn )if (defined $ifn);
			open ($ifh, "<&STDIN");
		}
	}
	elsif ($el =~ /^-h/ or $el =~ /^-\?/ or $el =~ /^--help/) {
		print $shortwelcome;
		print <<HELP;
Usage:	
	mqlresult [options]

Options:
	-b [<filename>]	start in batchmode and process commands in 
			<filename>. If filename is not specified, read
			commands from STDIN.	
	-h, -?, --help	print this help
	-v		print version information
	
HELP
	exit(0);
}
	else {
		$curDir = $el;
	}	
		
}
if (not $batchloop) {
	$term = Term::ReadLine->new('MPMASQL Result Explorer');
	print "MPMASQL Result $thisversion\nInteractive mode (using ".$term->ReadLine.")\n";
	$shelloop = 1;

}

my $curResultset;

MAINLOOP: while ($shelloop or $batchloop) {
	my $commandstring;
	my @commands;
	if ($shelloop) {		
		my $prompt = "MpmasQL Result: $curDir>";
		$commandstring = $term->readline($prompt); #ReadLine(0) ;
		$term->addhistory($commandstring);
	}
	if ($batchloop) {
		if (eof ($ifh) ) {
			$batchloop = 0;
			 last MAINLOOP ;
		}
		$commandstring = <$ifh>;
		print "$commandstring";
	}
	if (defined $commandstring) {
		chomp $commandstring;
		$commandstring =~ s/^\s*//;
		$commandstring =~ s/\s*$//;
		
		
		if ($commandstring =~ /"/) {
			while ($commandstring) {
				if ($commandstring =~ /^\s*"([^"]*)"\s*/){
					push @commands, $1;	
					$commandstring =~ s/^[\s]*"$1"[\s]*//;
				}
				elsif ($commandstring =~ /^\s*([\w^"])\s+/){
					push @commands, $1;	
					$commandstring =~ s/^\s*$1\s+//;
				}
				elsif ($commandstring =~ /^\s*([\w^"])\s*$/){
					push @commands, $1;	
					$commandstring =~ s/^\s*$1\s+//;
				}
			}
			
		}
		else {
			@commands = split /\s/, $commandstring;	
		}
		#print ": $_\n" foreach @commands;
		#sleep 2;
	}


	my $action = shift @commands;
	$action = "" unless (defined $action); 

	if ($action eq "quit" or $action eq "exit") {
		$shelloop = 0;
		
		exit 0;		
	}
	elsif ($action eq "setdir") {
		$curDir = join(" ", @commands);
	}
	elsif ($action eq "load") {
			my $scenario = shift @commands;
			my $options = join (" ", @commands);

			$curResultset = undef;
			$curResultset = MpmasResults->create($curDir, $scenario, $options);
	}
	elsif ($action eq "actgrouptotal") {
			my $classification = shift @commands;
			if ( defined $classification) {
				my $resulthashtab = $curResultset->total_solution_by_group($classification);
				print "\n";
				print "\t\tPeriod\n";
				print "Group\t";
				for my $i ($curResultset->periods) {
					print "\t$i";
				}
				print "\n";
				for my $j (sort {$a cmp $b} keys %{$resulthashtab}) {
					print "$j\t";
					for my $i ($curResultset->periods ) {
						print "\t".$resulthashtab->{$j}{$i};
					}
					print "\n";
				}
			}
			else {
				print "Error: incomplete arguments. Usage: actgrouptotal <classification> \n";
			}
	}
	elsif ($action eq "rhsgrouptotal") {
			my $classification = shift @commands;
			if ( defined $classification) {
				my $resulthashtab = $curResultset->total_rhs_by_group($classification);
				print "\n";
				print "\t\tPeriod\n";
				print "Group\t";
				for my $i ($curResultset->periods) {
					print "\t$i";
				}
				print "\n";
				for my $j (sort {$a cmp $b} keys %{$resulthashtab}) {
					print "$j\t";
					for my $i ($curResultset->periods ) {
						print "\t".$resulthashtab->{$j}{$i};
					}
					print "\n";
				}
			}
			else {
				print "Error: incomplete arguments. Usage: rhsgrouptotal <classification> \n";
			}
	}
	elsif ($action eq "assetdist") {
			my $classification = shift @commands;
			my $period	= shift @commands;
			my $fn =  shift @commands;
			if (defined $period and defined $classification) {
				if (not defined $fn) {
					print "Please enter a filename:";
					$fn = ReadLine(0);
					chomp($fn);
					print "\n";
				}
				my $resulthashtab = $curResultset->rhsgroup_distribution_for_period($classification, $period);
				
				open (my $fx, ">", $fn) or die "$fn :". $! ."\n";

				print "Writing RHS Distribution of Group $classification for Period $period to $fn\n";
				print $fx "Pop\tClu\tAsset\tUbound\tValue\n";
				for my $p (sort {$a cmp $b} keys %{$resulthashtab}) {
					for my $c (sort {$a cmp $b} keys %{$resulthashtab->{$p}}) {	
						for my $g (sort {$a cmp $b} keys %{$resulthashtab->{$p}->{$c}}) {
							for my $l (sort {$a <=> $b} keys %{$resulthashtab->{$p}->{$c}->{$g}}) {
								print $fx "$p\t$c\t$g\t".$resulthashtab->{$p}->{$c}->{$g}->{$l}."\t$l\n";
							}
						}
					}
				}
				close ($fx);
			}
			else {
				print "Error: incomplete arguments. Usage: assetdist <classification> <period>\n";
			}
	}
	elsif ($action eq "lt2totab") {
			my $fn = shift @commands;
			my $options = join (" ", @commands);
			my $dbheader = 0;
			$dbheader = 1 if (defined $options and $options =~ /-h/);
			if (not defined $fn) {
					print "Please enter a filestub:";
					$fn = ReadLine(0);
					chomp($fn);
					print "\n";
				}
			my $fn_hd = $fn ."_hhmembers.tab";			
			my $fn_ad = $fn ."_assetdist.tab";
			my $fn_ai = $fn ."_agentinfo.tab";
			$curResultset->lt2totab($fn_ad, $fn_hd, $fn_ai, $dbheader );
	}
	elsif ($action eq "lt2sum") {
			my $options = join (" ", @commands);
			my $sumtab = $curResultset->lt2sum();	
			my $fh;
			if ($options =~ /-f\s+"([^"]+)/ or $options =~ /-f\s+([\w.\/]+)/) {
				open ($fh, ">", $1) or (print STDERR "Error when trying to open $1: $!\n" and next);
			}
			else {
				open ($fh, ">&STDOUT");
			}
			if (defined $options  and $options =~ /\s+-pif\s*\(\s*([\w.\/]+|"[^"]+)\s*,\s*([\w.\/]+|"[^"]+)\s*\)/ ) {
				my $template = $1;
				my $link = $2;
				$template =~s/^"//;
				$template =~s/"$//;
				
				$link =~ s/^"//;
				$link =~ s/"$//;
				my $tfh;
				open ($tfh , "<", $template) or print STDERR " Error: Cannot open $template: $!\n" and exit(1);
				my $linktab = IndexedTable->create_from_ascii($link,1);	
				while (<$tfh>) {
					my $in = $_;
					next if ($in =~ /^$/ or $in =~ /^\s*pif/ );
					$in =~ /l1\s+!(\w+)!/;

					local $| = 1;
					eval {
					local $SIG{__WARN__} = sub {
						print STDERR "Problem with $1\n";
					};
					my $p = $linktab->query([$1],2);
					my $c = $linktab->query([$1],3);
					my $i = $linktab->query([$1],4);
					my $res = 0;
					if (exists $sumtab->{$p}->{$c}->{$i}) {
						$res = $sumtab->{$p}->{$c}->{$i};
					}
					print $fh $res."\n";
					};
					
	
				}
				close ($tfh);
			}
			else {
				#default: tab-separated ascii
				
				my $filter;
				if (defined $options  and $options =~ /\s+-r\s*(\w+)/ ) {
					$filter = $1;		
				}
				my $compare;
				if (defined $options  and $options =~ /\s+-cmp\s*\(\s*([\w.\/]+|"[^"]+)\s*\)/ ) {
					$compare = IndexedTable->create_from_ascii($1, 3);		
				}
				print $fh "pop\tclu\tasset\tsum";
				print $fh "\tref\tdif" if (defined $compare); 
				print $fh "\n";
				if (defined $options  and $options =~ /\s+-h/ ) {
					print  $fh "INT\tINT\tVARCHAR(100)\tDOUBLE";
					print $fh "\tDOUBLE\tDOUBLE" if (defined $compare); 
					print $fh "\n";
					print  $fh "k\tk\tk\tn";
					print $fh "\tn\tn" if (defined $compare); 
					print $fh "\n";
				}
				for my $p (sort {$a <=> $b} keys %$sumtab) {
					for my $c (sort {$a <=> $b}  keys %{$sumtab->{$p} } ) {
						for my $i (sort {$a cmp $b}  keys %{$sumtab->{$p}->{$c} } ) {
							next if (defined $filter and not $i =~ /$filter/);
							print  $fh "$p\t$c\t$i\t".$sumtab->{$p}->{$c}->{$i};
							if (defined $compare) {
								print $fh "\t".$compare->query([$p,$c,$i],4);
								my $dif = $sumtab->{$p}->{$c}->{$i} - $compare->query([$p,$c,$i],4);
								print $fh "\t".$dif ;
							}
							print $fh "\n";
						}
					}
				}	
			}	
			close ($fh);
	}
	elsif ($action eq "lt2dist"){
			my $options = join (" ", @commands);
			my $sumtab = $curResultset->lt2dist();	
			my $fh;
			if ($options =~ /-f\s+"([^"]+)/ or $options =~ /-f\s+([\w.\/]+)/) {
				open ($fh, ">", $1) or (print STDERR "Error when trying to open $1: $!\n" and next);
			}
			else {
				open ($fh, ">&STDOUT");
			}

			my $filter;
			if (defined $options  and $options =~ /\s+-r\s*(\w+)/ ) {
				$filter = $1;		
			}
			my $compare;
			if (defined $options  and $options =~ /\s+-cmp\s*\(\s*([\w.\/]+|"[^"]+)\s*\)/ ) {
				$compare = IndexedTable->create_from_ascii($1, 4);		
			}
			print $fh "pop\tclu\tasset\tlevel\tn";
			print $fh "\tref\tdif" if (defined $compare); 
			print $fh "\n";
			if (defined $options  and $options =~ /\s+-h/ ) {
				print  $fh "INT\tINT\tVARCHAR(100)\tDOUBLE\tINT";
				print $fh "\tDOUBLE\tDOUBLE" if (defined $compare); 
				print $fh "\n";
				print  $fh "k\tk\tk\tk\tn";
				print $fh "\tn\tn" if (defined $compare); 
				print $fh "\n";
			}
			for my $p (sort {$a <=> $b} keys %$sumtab) {
				for my $c (sort {$a <=> $b}  keys %{$sumtab->{$p} } ) {
					for my $i (sort {$a cmp $b}  keys %{$sumtab->{$p}->{$c} } ) {
						next if (defined $filter and not $i =~ /$filter/);
						for my $l (sort {$a <=> $b}  keys %{$sumtab->{$p}->{$c}->{$i} } ) {
							print  $fh "$p\t$c\t$i\t$l\t".$sumtab->{$p}->{$c}->{$i}->{$l};
							if (defined $compare) {
								print $fh "\t".$compare->query([$p,$c,$i,$l],4);
								my $dif = $sumtab->{$p}->{$c}->{$i}->{$l} - $compare->query([$p,$c,$i,$l],5);
								print $fh "\t".$dif ;
							}
							print $fh "\n";
						}
					}
				}
			}		
			close ($fh);
	}
	else {
		if ($shelloop) {
			print "\nUnknown Command\n"
		}
	}
} ;

close ($ifh) if (defined $ifh);
exit 0;


