#!/usr/bin/perl
use strict;
use warnings;
use lib qw(/usr/local/lib/mpmasql/modules);
use IO::Handle;
use Term::ReadLine;

use MpmasInputfile;
use MpmasMILP;
use CtUtilities qw(is_in_list);
use Term::ReadKey;
##Version:
my $thisversion = "v2.9";
my $lastchange = "16-09-2013";

=head1 mqlmatrix 

Perl script to examine, change and solve MP-MAS matrix files in ASCII format (.mtx, .mpx or .mip, .dat)
(c)	Christian Troost, Univ. Hohenheim, 2009-2013



=cut

my $longwelcome = <<WC;
####################################################################
####		mpmasql Matrix Browser $thisversion		####
####	Perl script to examine, change and solve MP-MAS 	####
####	matrix files in ASCII format (.mtx, .mpx or .mip, .dat)	####
####								####
####	(c)	Christian Troost, Univ. Hohenheim, 2009-2013	####
####	Version: $thisversion		$lastchange		####
####################################################################
WC

my $shortwelcome = <<WC2;
mpmasql Matrix Browser $thisversion		
(c)	Christian Troost, Universität Hohenheim, 2009-2013

WC2

my @tempfiles;
my $pager = "less";
$pager = "most" if ( `which most`);
my $spreadsheet = "oocalc";
$spreadsheet ="gnumeric" if ( `which gnumeric`);


my $matrixfile;
my $actdesname;
my $condesname;
my $mxmode;
my $mpxformat = 0;
my $mpmas_subformat;
my $tableau_compact = 0;
my $desdir = "xlsInput/";
my $milpcheck = "/usr/local/bin/MilpCheck";
my $milpcheck_options = "";
system('mkdir -p /tmp/mpmasql/');
my $outdir = "/tmp/mpmasql/";
my $shell = 1;
my $import_solution = 0;
my @children;
my $counter = 0;
my $ini = "mqlmatrix.conf";
my $ra;
if (-e $ini) {
	open (my $INI,"<", $ini) or die "Error when trying to read configuration:$ini\n $!\n";
	while (<$INI>) {
		my $line = $_;
		if ($line =~ /^\s*outdir\s*=\s*(.*)$/){
			$outdir = $1;
			$outdir =~ s/\s*$//;
		}
		elsif ($line =~ /^\s*milpcheck\s*=\s*(.*)$/){
			$milpcheck = $1;
			$milpcheck =~ s/\s*$//;
		}
		elsif ($line =~ /^\s*infofiles\s*=\s*(.*)$/){
			$desdir = $1;
			$desdir =~ s/\s*$//;
		}
		elsif ($line =~ /^\s*milpcheck_options\s*=\s*(.*)$/){
			$milpcheck_options = $1;
			$milpcheck_options =~ s/\s*$//;
		}
	}
	close ($INI);
}

#Parse commandline arguments
while(@ARGV) {
	my $element = shift @ARGV;
	 if (not $element =~ /^-/) {
			# solutionfile name
			$matrixfile = $element;
			}
	elsif ($element =~ /^-A/) {
			# activity description file 
			$actdesname = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}
	elsif ($element =~ /^-O/) {
			# output directory
			$outdir = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}
	elsif ($element =~ /^-I/) {
			#directory of constraint and activity description files
			$desdir = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}
	elsif ($element =~ /^-C/) {
			# constraint description file 
			$condesname = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}			  
	elsif ($element =~ /^-M/) {
			#MilpCheck
			$milpcheck = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}
	elsif ($element =~ /^-F/) {
			$mxmode = "Failed";
			}			  
	elsif ($element =~ /^-S/) {
			$mxmode = "Single";
	}	
	elsif ($element =~ /^-D/) {
			$mxmode = "Debug";
	}
	elsif ($element =~ /^-V/) {
			$mxmode = "Mpmas";
			$mpmas_subformat = shift @ARGV if ($ARGV[0] =~ /^(\d+)/);
			print "Error: Please specify a valid subformat for option -V.\n(Run mqlmatrix --help for more information.)\n" and exit(1) unless (defined $mpmas_subformat and $mpmas_subformat >= 0 and $mpmas_subformat <= 2);
	}
	elsif ($element =~ /^--compact/ or $element =~ /^-c/) {
			$tableau_compact = 1;
	}
	elsif ($element =~ /^--risk_aversion/ ) {
			$ra = shift @ARGV unless ($ARGV[0] =~ /^-/);
	}
	elsif ($element =~ /^-p/) {
			$shell = 0;
	}
	elsif ($element =~ /^-R/) {
			$import_solution = 1;
	}
      elsif ($element =~ /^-XP/) {
			$pager = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}
      elsif ($element =~ /^-XS/) {
			$spreadsheet = shift @ARGV unless ($ARGV[0] =~ /^-/);
			}
	elsif ($element =~ /^-h/ or $element =~ /^-\?/ or $element =~ /^--help/) {
		print $shortwelcome;
		print <<HELP;
Usage:
	mqlmatrix <filename> [options]

Options:
	File format:
	-F		expect matrix in failed format (.mip/.mpx)
	-S		expect matrix in single agent format (.mtx)
	-D		expect matrix in debug format (.mip/.mpx)
	-V <subformat>	expect matrix in mpmas input file format (.dat) 
			<subformat>
			0: basic format 
			1: with second production stage ("harvest milp")
			2: with advanced consumption model
			11: TSPC, with second production stage (not implemented)
			12: TSPC, with advanced consumption model (not implemented)
	(specifying one of these formats is obligatory)

	--compact, -c	expect tableau in compact format (only in combination with -V and -S)
	

	Override default settings:
	-A <filename>	specify file containing activity description
	-C <filename>   specify file containing constraint description
	-O <directory>	override output directory
	-I <directory>	specify directory containing description files
	(if not specified, the directories are read from mqlmatrix.conf in the working directory
	and names of description files are guessed from the matrix file name)

	-M <filename>	filename of MilpCheck executable
			(default: /usr/local/bin/MilpCheck)
	-XP <command>	set program for paging (defaults: most | less)
	-XS <command>	set spreadsheet program for table views (defaults: gnumeric | oocalc)
	
	Others:	
	-p			print matrix in single agent format (.mtx) and exit
	-R			import solution vector
	-h, -?, --help		print this help
	-v			print version information
	--risk_aversion <value> set risk aversion coefficient for evaluation of Q matrix to <value>
HELP
		exit (0);
	}
	elsif ($element =~ /^-v/ ) {
		print $longwelcome;
		exit(0);
	}
}

print STDERR "Error: No matrix specified\n" and exit(1) unless (defined $matrixfile);
print STDERR "Error: You did not indicate the matrix file format\n(-F: Failed MILP,\n -S: Single Matrix, -D: Debugging MILP, -V <subformat>: Mpmas Input File)\n" and exit(1) unless (defined $mxmode);


my $scen;
my $ag = "X";
my $type;
my $milpnumber;
my $solutionfile;
if ($mxmode eq "Failed") {
	print STDERR "Warning: expecting .mip or .mpx file when using format option -F\n" unless ($matrixfile =~/\.(mip|mpx)$/); 
	#$matrixfile =~ /Failed_(\w+?)_(\d+)_(...)Loaded(\d*)\w*\.(mip|mpx)/;
	$matrixfile =~ /(\w+?)_\d_(\d+)_MILP\d+_Failed(...)Loaded(\d*)\w*\.(mip|mpx)/;

	$scen = $1;
	$ag = $2;
	$type = $3;
	$milpnumber = $4;
	$mpxformat = 1 if ($5 eq "mpx");
}
elsif ($mxmode eq "Debug") {
	print STDERR "Warning: expecting .mip or .mpx file when using format option -D\n" unless ($matrixfile =~/\.(mip|mpx)$/); 
	$matrixfile =~ /(\w+?)_(\d+)_(\d+)MILP(\d*)\.(mip|mpx)/;
	$scen = $1;
	my $period = $2;
	$ag = $3;
	$mpxformat = 1 if ($5 eq "mpx");
	$type = "Dbg";
	$milpnumber = $4;
	$mxmode = "Failed";
}
elsif ($mxmode eq "Single") {
	print STDERR "Warning: expecting .mtx file when using format option -S\n" unless ($matrixfile =~/\.mtx$/); 
	$matrixfile =~ /\/?(\w+)MILP_(\w+)\.mtx$/;
	$scen = $1;
	$ag = $2;
	$type = "RE";
	$milpnumber = "";
}
elsif ($mxmode eq "Mpmas") {
	print STDERR "Warning: expecting .dat file when using format option -V\n" unless ($matrixfile =~/(\.cpt|\.dat)$/); 
	$matrixfile =~ /\/?(\w+)MILP(\w+)?(\.cpt|\.dat)$/;
	$scen = $1;
	$ag = $2 || "blank";
	$type = "RE";
	$milpnumber = "";
}
$solutionfile = $matrixfile;
$solutionfile =~ s/(.mtx|.mpx|.mip)\s*$/.pri/;


print $shortwelcome."\n";
print "Pager:\t$pager\n"; 
print "Spreadsheet:\t$spreadsheet\n";

print "\nMatrix file:\t\t". $matrixfile."\n";  
print "Identified Scenario $scen, agent $ag\n";
$actdesname = $desdir.$scen."ActivityInfo.txt" if (not defined $actdesname);
$condesname = $desdir.$scen."ConstraintInfo.txt" if (not defined $condesname);	
print "Constraint Description file:\t\t". $condesname."\n";  
print "Activity Description file:\t\t". $actdesname."\n";  
MpmasInputfile->set_global_references($scen, $outdir, $outdir, undef, undef, undef, undef, undef, undef,undef );
my $matrix = MpmasMILP->recreate_from_file($matrixfile, $mxmode, $actdesname, $condesname, $mpxformat, $mpmas_subformat, $tableau_compact);
if ($import_solution) {
    print "Solution imported from $solutionfile\n";
    $matrix->import_solution_vector($solutionfile, $ra) ;
}
my $status = -1;

unless ($shell) {
	print_matrix($matrix, $ag, $type);
	exit 0;
}
else {
	my $term = Term::ReadLine->new('Mpmas QL Matrix');
	print "Interactive mode (using ".$term->ReadLine.")\n";
	while ($shell) {
		my $prompt = "mqlmatrix>";
		my $commandstring = $term->readline($prompt); #ReadLine(0) ;
		#$term->addhistory($commandstring);
		$commandstring =~ s/^\s*//;
		$commandstring =~ s/\s*$//;
		if ($commandstring eq "exit" or $commandstring eq "quit") {
			$shell = 0;
		}
		elsif ($commandstring =~ /^help(\s.*)?/) {
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1);
			push @children, $pid if $pid;
			list_help($out);
			close $out;
		}
		elsif ($commandstring eq "info") {
			print "Number of rows:\t".$matrix->milpinfo("NumRow")."\n";
			print "Number of columns:\t".$matrix->milpinfo("NumCol")."\n";
			print "Integers:\t".$matrix->milpinfo("IntegAct")."\n";
			print "Quadratic variables:\t".$matrix->milpinfo("QuadraticAct")."\n";
			print "Objective value:\t".$matrix->milpinfo("TargetValue")."\n" if (defined $matrix->milpinfo("TargetValue") );
			print "Objective value (linear part):\t".$matrix->milpinfo("ObjectiveValueLin")."\n" if (defined $matrix->milpinfo("ObjectiveValueLin") );
			print "Objective value (quadratic part):\t".$matrix->milpinfo("ObjectiveValueQuad")."\n" if (defined $matrix->milpinfo("ObjectiveValueQuad") );
			print "Variance in risk model (r = ". $matrix->milpinfo("RiskAversionCoef")  ."):\t".$matrix->milpinfo("VarianceRiskModel")."\n" if (defined $matrix->milpinfo("VarianceRiskModel") );
			print "Risk aversion coefficient now set to:\t $ra\n" if (defined $ra);
			if ($status eq "-1") {
				  print "Status:\tno attempt to solve yet\n" ;
			}
			else {
				  print "Status:\t$status\n" 
			}
		}
		elsif ($commandstring =~ /^printtxt\s*(\w*)/) {
			$type = $1 if (defined $1);
			print_matrix($matrix,$ag, $type, 1);
			print "\n";
		}
		elsif ($commandstring =~ /^printmtx\s*(\w*)/) {
			$type = $1 if (defined $1);
			print_matrix($matrix,$ag, $type, 0);
			print "\n";
		}
		elsif ($commandstring eq "reload") {
			$matrix = undef;
			$matrix = MpmasMILP->recreate_from_file($matrixfile, $mxmode, $actdesname, $condesname, $mpxformat, $mpmas_subformat, $tableau_compact);
			$matrix->import_solution_vector($solutionfile, $ra) if $import_solution;
		}
		elsif ($commandstring =~ /^set_solver_options\s*(.*)/) {
			if ($1 eq "NULL") {
				$milpcheck_options = "";
			}
			else {
			    $milpcheck_options = $1;
			    chomp($milpcheck_options);
			}
		}
		elsif ($commandstring =~ /^set_risk_aversion\s*([\d.]*)/) {
			if ($1 eq "NULL") {
				$ra = undef;
			}
			else {
			    $ra = $1;
			}
		}
		elsif ($commandstring =~ /^append_solver_options\s*(.*)/) {
		        $milpcheck_options .= " ".$1;
			chomp($milpcheck_options);
		}
		elsif ($commandstring =~ /^solver_options/) {
		      if (defined $milpcheck_options) {
			    print $milpcheck_options."\n";
		      }
		      else {
			    print "No solver options defined\n";
		      }
		}
		elsif ($commandstring =~ /^solve\s*(\d)?\s*(-test)?/) {
			print STDERR "Warning: You have loaded a matrix indicating -V, which stands for the MPMAS input file format (.dat).\nAttempting to solve this format usually leads to unexpected results, because of missing agent_specific RHS, missing yields, investment coefficients and cost, missing dynamic coefficients for perennials, et cetera.\n" if ($mxmode eq "Mpmas");
			my $modestring = $1;
			my $teststring = $2;
			my $type = "_temp";
			$type .= ++$counter;
			
			if (defined $modestring) { 
			      $modestring = " -mode ".$modestring;
			      print "Using$modestring\n";
			};
			my $fn = print_matrix($matrix,$ag, $type, 0);
			$fn =~ s/.mtx$//;
			my $string = $milpcheck." ".$fn;
			$string .= " ".$teststring if (defined $teststring);
			$string .= $modestring if (defined $modestring);
			$string .= " ".$milpcheck_options if (defined $milpcheck_options);
			
			print "$string\n";
			system($string);
			$fn .= ".pri";
			eval {
				local $SIG{__WARN__}=sub{}; #temporarily suppress warnings
				$matrix->import_solution_vector($fn, $ra);
			};
			if ($@) {
				print "$@";
				$status = "Last solution failed, ".`date +%T`. $@;
			}
			else {
				print "Solution has been imported.\n"
				
			}
			$status = "Last solution successful, ".`date +%T` ;
		}
		elsif ($commandstring =~ /^impsol\s*(.*)/) {
			
			$solutionfile = $1 if (defined $1);
			print "Importing solution from $solutionfile\n";
			eval {
				local $SIG{__WARN__}=sub{}; #temporarily suppress warnings
				$matrix->import_solution_vector($solutionfile, $ra);
			};
			print "$@" if ($@);
			$status = "Solution imported ".`date +%T` ;
		}
		elsif ($commandstring =~ /^col\s+(\d+)/) {
			my $k = $matrix->activity_at_col($1);
			print "Error: Column $1 does not exist\n" and next unless (defined $k);
			print_info($k);	
		}
		elsif ($commandstring =~ /^Q\s+(\d+)/) {
			my $k = $matrix->activity_at_col($1);
			print "Error: Column $1 does not exist\n" and next unless (defined $k);
			if ($k->attr("quadratic")) {
			  print "Q matrix for activity ". $k->id. "\n";
			  for my $i ($k->subelement_fields("quadratic")) {
			      print sprintf("%30.27s\t%f\n", $i, $k->subelement_field("quadratic", $i));
			  }
			}
			else {
			  print "Column $1 is not quadratic\n" 
			}
			
		}
		elsif ($commandstring =~ /^multiply\s+Q\s+([-\d.E+e]+)/) {
			my $m = $1;
			foreach my $j ($matrix->sorted_quadratic_activities() ) {
			    my $k = $matrix->activity($j);
			    foreach my $i ($k->subelement_fields("quadratic")) {
				  $k->modify_subelement_field("quadratic", $i, $k->subelement_field("quadratic", $i) * $m);
			    }
			  }
		}
		elsif ($commandstring =~ /^divide\s+Q\s+([-\d.E+e]+)/) {
			my $m = $1;

			foreach my $j ($matrix->sorted_quadratic_activities() ) {
			    my $k = $matrix->activity($j);
			    foreach my $i ($k->subelement_fields("quadratic")) {
				  $k->modify_subelement_field("quadratic", $i, $k->subelement_field("quadratic", $i) / $m);
			    }
			  }
		}
		elsif ($commandstring =~ /^coef\s+col\s+(\d+)(\s+.+)?/) {
			
			my $k = $matrix->activity_at_col($1);
			print "Error: Column $1 does not exist\n" and next unless (defined $k);
			my ($out, $pid, $full, $nf, $typefilter) = check_options($2, undef, \@tempfiles);
			push @children, $pid if $pid;
			print $out "Tableau Coefficients of Col $1 (Activity ".$k->id."):\n";
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $c = $matrix->constraint_at_row($i);
				print $out sprintf("%d\t%30s\t%.4f\t%s\n", $i, $c->id, $k->coeff($c->id), $c->attr("name")) if (defined $k->coeff($c->id)) ;
			} 
			close ($out);
		}
		elsif ($commandstring =~ /^coef\s+row\s+(\d+)(\s+.+)?/) {
			my $k = $matrix->constraint_at_row($1);
			print "Error: Row $1 does not exist\n" and next unless (defined $k);
			my ($out, $pid, $full, $nf, $typefilter) = check_options($2, undef, \@tempfiles);
			push @children, $pid if $pid;
			print $out "Tableau Coefficients of Row $1 (Constraint ". $k->id."):\n";
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $c = $matrix->activity_at_col($i);
				print $out sprintf("%d\t%30s\t%.4f\t%s\n", $i, $c->id, $c->coeff($k->id), $c->attr("name")) if (defined $c->coeff($k->id));
			} 
			close ($out);
		}
		elsif ($commandstring =~ /^row\s+(\d+)/) {
			my $k = $matrix->constraint_at_row($1);
			print "Error: Row $1 does not exist\n" and next unless (defined $k);
			print_info($k);	
		}
		elsif ($commandstring =~ /^con\s+(\w+)/) {
			my $k = $matrix->constraint($1);
			print "Error: Constraint $1 does not exist\n" and next unless (defined $k);
			print_info($k);	
		}
		elsif ($commandstring =~ /^act\s+(\w+)/) {
			my $k = $matrix->activity($1);
			print "Error: Activity $1 does not exist\n" and next unless (defined $k);
			print_info($k);	
		}
		elsif ($commandstring =~ /^modify\s+row\s+(\d+)\s*\(\s*(\w+)\s*=\s*([-.\w]+)\s*\)/) {
			my $k = $matrix->constraint_at_row($1);
			print "Error: Row $1 does not exist\n" and next unless (defined $k);
			$k->set_attr("$2", $3);
			print "\n";
			print_info($k);
		}
		elsif ($commandstring =~ /^modify\s+rows\s+(\d+):(\d+)\s*\(\s*(\w+)\s*=\s*([-.\w]+)\s*\)/) {
			for my $i ($1..$2) {
			    my $k = $matrix->constraint_at_row($i);
			    print "Error: Row $i does not exist\n" and next unless (defined $k);
			    $k->set_attr("$3", $4);
			    print "\n";
			    print_info($k);
			}
		}
		elsif ($commandstring =~ /^modify\s+col\s+(\d+)\s*\(\s*(\w+)\s*=\s*([-.\w+]+)\s*\)/) {
			my $k = $matrix->activity_at_col($1);
			print "Error: Column $1 does not exist\n" and next unless (defined $k);
			$k->set_attr("$2", $3);
			print "\n";
			print_info($k);
			
		}
		elsif ($commandstring =~ /^modify\s+cols\s+(\d+:\d+)\s*\(\s*(\w+)\s*=\s*([-.\w+]+)\s*\)/) {
		    for my $i ($1..$2) {
			my $k = $matrix->activity_at_col($i);
			print "Error: Col $i does not exist\n" and next unless (defined $k);
			$k->set_attr("$3", $4);
			print "\n";
			print_info($k);
		    }
		}
		elsif ($commandstring =~ /^modify\s+coef\s+(\d+)\s*(\d+)\s+([-.\d]+)/) {
			my $k = $matrix->activity_at_col($1);
			my $l = $matrix->sorted_constraints($2);
			print "Error: Column $1 does not exist\n" and next unless (defined $k);
			print "Error: Row $2 does not exist\n" and next unless (defined $l);
			$k->set_coeff("$l", $3);
			print "Tableau Coefficients of Col $1 (Activity ".$k->id."):\n";
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $c = $matrix->constraint_at_row($i);
				print sprintf("%d\t%30s\t%.4f\t%s\n", $i, $c->id, $k->coeff($c->id), $c->attr("name")) if (defined $k->coeff($c->id)) ;
			}
		}
		elsif ($commandstring =~ /^rhs\s+(\d+)/) {
			my $k = $matrix->constraint_at_row($1);
			print "Error: Row $1 does not exist\n" and next unless (defined $k);
			print sprintf("%d\t%20s\t%.4f\t%20s\n", $1, $k->id, $k->attr("value"), $k->attr("name"));
		}
		elsif ($commandstring =~ /^lhs\s+(\d+)/) {
			my $k = $matrix->constraint_at_row($1);
			print "Error: Row $1 does not exist\n" and next unless (defined $k);
			print "LHS of Row($1) based on Solution\n";
			print $matrix->calculate_lhs_of_constraint($k)."\n";
		}
		elsif ($commandstring =~ /^lhslb\s+(\d+)/) {
			my $k = $matrix->constraint_at_row($1);
			print "Error: Row $1 does not exist\n" and next unless (defined $k);
			print "Minimum LHS of Row($1) based on Lower Bounds of Columns\n";
			print $matrix->calculate_lhs_of_constraint($k, "lbound")."\n";
		}
		elsif ($commandstring =~ /^list\s+act(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				print $out sprintf("%d\t%30.27s\t%20s\t%s\n", $i, $k->id, $k->attr("type"), $k->attr("name") ) if (not defined $typefilter or $typefilter eq $k->attr("type") ) ;
			}
			close($out);
			
		}
		elsif ($commandstring =~ /^list\s+con(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $k = $matrix->constraint_at_row($i);
				print $out sprintf("%d\t%30.27s\t%20s\t%s\n", $i, $k->id, $k->attr("type"), $k->attr("name") ) if (not defined $typefilter or $typefilter eq $k->attr("type") );
			}
			close($out);
		}
		elsif ($commandstring =~ /^list\s+==(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $k = $matrix->constraint_at_row($i);
				print $out sprintf("%d\t%30.27s\t%20s\t%s\n", $i, $k->id, $k->attr("type"), $k->attr("name") ) if ($k->attr("eqtype") == 3) and (not defined $typefilter or $typefilter eq $k->attr("type") );
			}
			close($out);
		}
		elsif ($commandstring =~ /^list\s+>=(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $k = $matrix->constraint_at_row($i);
				print $out sprintf("%d\t%30.27s\t%20s\t%s\n", $i, $k->id, $k->attr("type"), $k->attr("name") ) if ($k->attr("eqtype") == 2) and (not defined $typefilter or $typefilter eq $k->attr("type") );
			}
			close($out);
		}
		elsif ($commandstring =~ /^list\s+rhs(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				
				my $k = $matrix->constraint_at_row($i);
				print $out sprintf("%d\t%30.27s\t$nf\t%s\n", $i, $k->id, $k->attr("value"), $k->attr("name") ) if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and ( $full or sprintf($nf,$k->attr("value")) != 0));
			}
			close($out);
		}
		elsif ($commandstring =~ /^list\s+lhslb(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			
			
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $k = $matrix->constraint_at_row($i);
				my $lhs = $matrix->calculate_lhs_of_constraint($k, "lbound");
				print $out sprintf("%d\t%30.27s\t$nf\t%s\n", $i, $k->id, $lhs, $k->attr("name") ) if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and ($full or sprintf($nf, $lhs) != 0));
			}
			close($out);
		}
		elsif ($commandstring =~ /^list\s+lhs(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $k = $matrix->constraint_at_row($i);
				if (not defined $typefilter or $typefilter eq $k->attr("type") ) {
				      my $lhs = $matrix->calculate_lhs_of_constraint($k);
				      print $out sprintf("%d\t%30.27s\t$nf\t%s\n", $i, $k->id, $lhs, $k->attr("name") ) if (($full or sprintf($nf, $lhs) != 0));
				}
			}
			close($out);
		}
		elsif ($commandstring =~ /^list\s+sol(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter, $integs) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				print $out sprintf("%d\t%30.27s\t$nf\t%s\n", $i, $k->id, $k->attr("solution") ?  $k->attr("solution") : 0, $k->attr("name") ) if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and ($integs == 0 or $k->attr("integ") == 1 ) and ( $full or ( defined $k->attr("solution") and sprintf($nf, $k->attr("solution")) != 0 and sprintf($nf, $k->attr("solution")) != -1 )));
			}
			close($out);
			
		}
		elsif ($commandstring =~ /^list\s+orow(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				print $out sprintf("%d\t%30.27s\t$nf\t%s\n", $i, $k->id, $k->attr("orow") ?  $k->attr("orow") : 0, $k->attr("name") ) if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and ( $full or ( defined $k->attr("orow") and sprintf($nf, $k->attr("orow")) != 0 and sprintf($nf, $k->attr("orow")) != -1 )));
			}
			close($out);
			
		}
		elsif ($commandstring =~ /^list\s+int(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			
			print $out "Integer Activities\n";
			print $out "Col\tCode\tPriority\tName\n";
			my $cc = 0;
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				(++$cc and print $out sprintf("%d\t%30.27s\t%d\t%s\n", $i, $k->id, $k->attr("priority"), $k->attr("name") ) ) if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and (defined $k->attr("integ") and $k->attr("integ") == 1));
				
			}
			print $out "$cc Marked Integers\t(".$matrix->milpinfo("IntegAct")." according to Matrix Info)\n";
			close ($out);
		}
		elsif ($commandstring =~ /^list\s+quadratic(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			
			print $out "Quadratic activities\n";
			print $out "Col\tCode\tName\n";
			my $cc = 0;
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				(++$cc and print $out sprintf("%d\t%30.27s\t%s\n", $i, $k->id,  $k->attr("name") ) ) if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and (defined $k->attr("quadratic") and $k->attr("quadratic") == 1));
				
			}
			close ($out);
		}
		elsif ($commandstring =~ /^list\s+lb(\s.*)?/){
			my ($out, $pid, $full, $nf, $typefilter) = check_options($1,undef, \@tempfiles);
			push @children, $pid if $pid;
			
			print $out "Activities with Lower Bounds > 0\n";
			print $out "Col\tCode\tLBound\tName\n";
			my $cc = 0;
			for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				print $out sprintf("%d\t%30.27s\t%e\t%s\n", $i, $k->id, $k->attr("lbound"), $k->attr("name") )  if ((not defined $typefilter or $typefilter eq $k->attr("type") ) and (defined $k->attr("lbound") and $k->attr("lbound") >0 ));
			}
			close $out;
		}
		elsif ($commandstring =~ /^tab\s+(\d+)\s*\.\.\s*(\d+)\s*:\s*(\d+)\.\.(\d+)(\s*-.+)?/) {
			my ($astart, $aend, $cstart, $cend) = ($1,$2,$3,$4);
			if ($astart < 0 or $aend > $matrix->milpinfo("NumCol") -1 or $cstart < 0 or $cend > $matrix->milpinfo("NumRow") -1) {
			    print "Invalid dimensions\n";
			}
			else {
			      my ($out, $pid, $full, $nf, $typefilter) = check_options($5, 3, \@tempfiles);
			      push @children, $pid if $pid;
			      print_tableau_part($out, [($astart..$aend)],[($cstart..$cend)], $nf);
			      close $out;
			}
		}
		elsif ($commandstring =~ /^tab\s+([\w\s,]+)\s+:\s+([\w\s,]+)(\s*-.+)?/) {
			my @atypes = split /\s*,\s*/, $1;
			my @ctypes = split /\s*,\s*/, $2;
			my $options = $3;
			$ctypes[$#ctypes] =~ s/\s*$//;
			$atypes[$#atypes] =~ s/\s*$//;
			#print join("|", @atypes)."\n";
			#print join("|", @ctypes)."\n";
			#print $options."\n";
			my ($out, $pid, $full, $nf, $typefilter) = check_options($options, 3, \@tempfiles);
			push @children, $pid if $pid;
			my (@alist ,@clist);
			 for my $i (0..$matrix->milpinfo("NumCol")-1) {
				my $k = $matrix->activity_at_col($i);
				if (is_in_list($k->attr("type"), \@atypes))  {
				    push @alist, $i;
				}
			 }
		         for my $i (0..$matrix->milpinfo("NumRow")-1) {
				my $k = $matrix->constraint_at_row($i);
				if (is_in_list($k->attr("type"), \@ctypes))  {
				    push @clist, $i;
				}
			 }
			 print_tableau_part($out, \@alist,\@clist, $nf);
                         close $out;                         
		}
		else {
			print "unrecognized command\n"
		}
	}
}

waitpidall(\@children, \@tempfiles);
unlink('/tmp/mpmasql/*');
exit(0);

sub print_tableau_part {
      my ($out, $alist, $clist, $nf) =@_;
      my @temptab;
      for my $i (@$alist) {
	      my $k = $matrix->activity_at_col($i);
	      push @{$temptab[0]}, $k->name;
	      push @{$temptab[1]}, $k->id;
	      push @{$temptab[2]}, $k->attr("type");
	      push @{$temptab[3]}, $k->attr("integ");
	      push @{$temptab[4]}, $k->attr("lbound");
	      push @{$temptab[5]}, $k->attr("ubound");
	      push @{$temptab[6]}, $k->lp_column;
	      push @{$temptab[7]}, $k->attr("orow");
      }
      for my $i (@temptab) {
	      print $out "\t\t\t\t\t\t".join("\t",@$i)."\n";
      }
      for my $j (@$clist) {
	      my $l = $matrix->constraint_at_row($j);
	      print $out join("\t", $l->name, $l->id, $l->attr("type"), $l->attr("showsign"), $l->attr("value"),$j);
	      for my $i (@$alist) {
		    my $k = $matrix->activity_at_col($i);
		    print $out "\t";
		    print $out sprintf("$nf",$k->coeff($l->id)) if (defined $k->coeff($l->id));
	      }
	      print $out "\n";
      }
}

sub check_options {
      my $options = shift;
      my $force_pipe = shift;
      my $tempfilesref = shift;
      my $nf = "%.4f";
      $nf = "%e" if (defined $options and $options =~ /-e/);
      $nf = "%.$1f" if (defined $options and $options =~ /-d\s*(\d+)/);
      my $full = (defined $options and $options =~ /-f/);
      my $pipe;
      my $file;
      
      if (defined $force_pipe) {
	    $pipe = $force_pipe
      }
      elsif (defined $options and $options =~ /-l/) {
	  $pipe = 1;
      }
      elsif (defined $options and $options =~ /-o\s*([^-]+)/) {
	  $pipe = 2;
	  $file = $1;
      };
      
      my ($out, $pid) = less_fork( $pipe,$file,$tempfilesref);
      my $typefilter;
      $typefilter = $1 if (defined $options and $options =~ /-t\s*(\w+)/);
      my $integs = 0;
      $integs = 1 if (defined $options and $options =~ /-i/);
      return $out, $pid, $full, $nf, $typefilter, $integs;
}

  sub list_help {
	  my $out = shift;
	  print $out <<HELP;
	  Commands:
	  
	  GENERAL:
	  exit			quits the interactive session
	  help			prints this help
	  quit			quits the interactive session

	  I/O
	  printmtx {<suffix>}	saves the matrix as .mtx into a file with the indicated <suffix>
	  printtxt {<suffix>}	saves the commented matrix into a file with the indicated <suffix>
	  reload		reloads the matrix from the original file
	  solve {<mode>}	tries to solve the matrix and to import the solution vector if successful 
	  impsol {<filename>}	import solution vector from file
	  
	  COLUMN/ROW INFORMATION
	  info				prints general information about the Matrix
	  act <id>			prints information about activity with mpmasql identifier <id>
	  con <id>			prints information about constraint with mpmasql identifier <id>
	  col <C>			prints information about matrix column no. <C>
	  row <R>			prints information about matrix row no. <R>
	  coef col <C> {-l}		prints all coefficients <> 0 of matrix column no. <C>
	  coef row <R> {-l}		prints all coefficients <> 0 of matrix row no. <R>
	  rhs <R>			prints the rhs for row <R>
	  lhs <R>			calculates the LHS for row <R> based on the current solution
	  lhslb <R>			calculates the lhs for row <R> based on the lbounds of columns
	  Q <C>				show Q matrix entries for column C
	  multiply Q <value>		multiply Q matrix by <value>
	  divide Q <value>		divide Q matrix by <value>

	  LISTING
	  list act {-l -o -t}		list all activities
	  list con {-l -o -t}		list all constraints
	  list == {-l -o -t}		list all equality constraints
	  list lhs {-l -f -e -d -o -t}	prints all LHS values <> 0 
	  list rhs {-l -f -e -d -o -t} 	prints all RHS values <> 0 
	  list sol {-l -e -d -o -t}	lists the solution for all activities where it is <> 0
	  list orow {-l -e -d -o -t}	lists the objective function coefficient for all activities where it is <> 0
	  list int {-l -o -t}		lists all integer activities
	  list lb {-l -o -t}		lists all activities with lower bounds > 0
	  list lhslb {-l -f -e -d -o -t}lists the LHS based on lbounds where LHS <> 0

	  Options for list: 
		  -l 			redirects the output to the less pager in a separate window
		  -o <filename>		redirects the output to <filename>
		  -e			scientific number format
		  -d <n>		number of digits (default <n> = 3) 	
		  -f			prints all elements, not only those <> 0
		  -t <type>		list only constraints/activities of mpmasql type <type>
														  
	  TABLE VIEW
	  tab <C1>..<C2> : <R1>..<R2>	view tableau between columns C1 and C2 and rows R1 and R2 in spreadsheet
	  tab <typelistA> : <typelistC>	view tableau for activities/constraint of types given in spreadsheet, 
					where <typelist> is comma-separated list of mpmasql types
	  MODIFICATIONS		
	  modify row <R> (<attr> = <new value>)		sets attribute <attr> of row <R> to <new value>
	  modify col <C> (<attr> = <new value>)		sets attribute <attr> of column <C> to <new value>
	  modify rows <R1:R2> (<attr> = <new value>)	sets attribute <attr> of rows <R1> to <R2> to <new value>
          modify cols <C1:C2> (<attr> = <new value>)	sets attribute <attr> of column <C1> to <C2> to <new value>
	  modify coef <C> <R> <new value>  		sets coefficient in column <C> and row <R> to <new value>

	  SOLVER OPTIONS
	  solver_options		show options passed to MilpCheck
	  append_solver_options		append to options passed to MilpCheck
	  set_solver_options		change options passed to MilpCheck

	  OTHER OPTIONS
	  set_risk_aversion <value>	set risk aversion coefficient for evaluation of quadratic part of objective values
					(Note: has no effect on current Q matrix)
HELP
  };
  sub print_matrix {
	  my $matrix = shift;
	  my $ag = shift;
	  my $type = shift;
	  my $mode = shift;
	  $matrix->blank_file;
	  $matrix->prepare_special(1, $ag.$type, {});
	  my $fn = $matrix->write($mode, $ag.$type,{},1);
	  return $fn;
  }
  sub print_info {
	  my $element = shift;
	  my @temp = @{$element->list_of_attribute_fields};
	  foreach my $i (@temp) {
		  print sprintf("%20s\t%s\n", $i, $element->attr($i)) if (defined $element->attr($i));
	  }

  }
  sub less_fork {
	  my $dofork = shift;
	  my $file = shift;
	  my $tempfilesref = shift;
	  if (not $dofork) {
		  open (my $out, ">&STDOUT" );
		  return $out;
	  }
	  elsif ($dofork == 1) {
	      pipe (PIPE_FROM_MAIN, PIPE_TO_LESS);
	      PIPE_TO_LESS->autoflush(1);
	      my $pid = fork();
	      if ($pid) {
		      close (PIPE_FROM_MAIN);
		      my $fn = "mpmasql".sprintf ("%03d", scalar @$tempfilesref).".temp";
		      push @$tempfilesref, $fn;
		      return (*PIPE_TO_LESS{IO}, $pid );
	      }
	      else {
		      close (PIPE_TO_LESS);
		      my $n = $#{$tempfilesref} + 1 ;
		      
		      my $fn = "mpmasql".sprintf ("%03d", scalar @$tempfilesref).".temp";
		      push @$tempfilesref, $fn;
		      
		      open (my $file ,">", "/tmp/mpmasql/$fn") or die "Could not open temp file:\n $!\n";
		      while ( <PIPE_FROM_MAIN> ){
			      
			      print $file $_;
		      };
		      
		      close ($file);
		      exec('gnome-terminal -e "'.$pager.' /tmp/mpmasql/'.$fn.'"');
		      
		      exit (0);
	      }
	 }
	elsif ($dofork == 2) {
		  open (my $out, ">", $file ) or (print "Error $! when trying to open $file\n" and return ">&STDOUT");
		  print "Output is written to file $file\n";
		  return $out;
	}
	elsif ($dofork == 3) {
	      pipe (PIPE_FROM_MAIN, PIPE_TO_LESS);
	      PIPE_TO_LESS->autoflush(1);
	      my $pid = fork();
	      if ($pid) {
		      close (PIPE_FROM_MAIN);
		      my $fn = "mpmasql".sprintf ("%03d", scalar @$tempfilesref).".temp";
		      push @$tempfilesref, $fn;
		      return (*PIPE_TO_LESS{IO}, $pid );
	      }
	      else {
		      close (PIPE_TO_LESS);
		      my $fn = "mpmasql".sprintf ("%03d", scalar @$tempfilesref).".temp";
		      push @$tempfilesref, $fn;
		      open (my $file ,">", "/tmp/mpmasql/$fn") or die "Could not open temp file:\n $!\n";
		      while ( <PIPE_FROM_MAIN> ){
			      
			      print $file $_;
		      };
		      
		      close ($file);
		      exec($spreadsheet.' /tmp/mpmasql/'.$fn);
		      
		      exit (0);
	      }
	 }
}
sub waitpidall {
	my $arrayref = shift;
	my $filearrayref = shift;
	for my $f (@$filearrayref) {
	    unlink("/tmp/mpmasql/$f");
	}
	waitpid($_, 0) foreach @{$arrayref};
}
