#!/usr/bin/perl
use strict;
use warnings;
use lib qw(/usr/local/lib/mpmasql/modules);

use Term::ReadLine;
use IndexedTable;
##Version:
my $thisversion = "v1.0 beta";
my $lastchange = "09-May-2011";

my $longwelcome = <<WC;
####################################################################
####	MP-MAS-QL Matrix Solving Batch $thisversion		####
####	Perl-script to solve a batch of MP-MAS matrices (.mtx)	####
####								####
####	(c)	Christian Troost, Univ. Hohenheim, 2011		####
####	Version: $thisversion		$lastchange		####
####################################################################
WC

my $shortwelcome = <<WC2;
MP-MAS-QL Matrix Solving Batch $thisversion		
(c)	Christian Troost, Universität Hohenheim, 2011
WC2

my $indir = "./";
my $actdesname;
my $condesname;
my $desdir = "./";
my $milpcheck = "/usr/local/bin/MilpCheckExe";
my $outdir = "./";
my $shell = 1;
my $ini = ".mqlmatrix";

if (-e $ini) {
	open (my $INI,"<", $ini) or die "Error when trying to read configuration:$ini\n $!\n";
	while (<$INI>) {
		my $line = $_;
		if ($line =~ /^\s*outdir\s*=\s*(.*)$/){
			$outdir = $1;
			$outdir =~ s/\s*$//;
		}
		elsif ($line =~ /^\s*milpcheck\s*=\s*(.*)$/){
			$milpcheck = $1;
			$milpcheck =~ s/\s*$//;
		}
		elsif ($line =~ /^\s*infofiles\s*=\s*(.*)$/){
			$desdir = $1;
			$desdir =~ s/\s*$//;
		}
		elsif ($line =~ /^\s*samdir\s*=\s*(.*)$/){
			$indir = $1;
			$indir =~ s/\s*$//;
		}

	}
	close ($INI);
}
my $samlist = "sam_list.txt";
my $prefix;
#Parse commandline arguments
my $parselevel = 0;
my @grouplist;
while (@ARGV) {
	my $element = shift @ARGV;
	if ($parselevel == 0) {
		if ($element =~ /^-G/ ){
			$parselevel = 1;
		}
		
		elsif (not $element =~ /^-/) {
				# samlist prefix
				$samlist = $element.$samlist;
				print "\nLooking for agent list in file:\t\t". $samlist."\n";  
		}
		elsif ($element =~ /^-O/) {
				# output directory
				$outdir = shift @ARGV unless ($ARGV[0] =~ /^-/);
		}
		elsif ($element =~ /^-I/) {
				# input directory
				$indir = shift @ARGV unless ($ARGV[0] =~ /^-/);
		}
		elsif ($element =~ /^-D/) {
				# input directory
				$desdir = shift @ARGV unless ($ARGV[0] =~ /^-/);
		}
		elsif ($element =~ /^-M/) {
				#MilpCheckExe
				$milpcheck = shift @ARGV unless ($ARGV[0] =~ /^-/);
		}
		elsif ($element =~ /^-h/ or $element =~ /^-\?/ or $element =~ /^--help/) {
			print $shortwelcome;
			print <<HELP;
Usage:
	mqlsam	<scenario> [options]

	<scenario>	the prefix of the sam_list.txt file containing 
			the list of matrices to process

Options:
	-G <resultgroup1> <resultgroup2> ... [-G]
			The resultgroups (as specified in .cll) to which solution activities will be aggregated.
			(Note: everything after the -G flag will be interpreted as a resultgroup until a second
			-G flag is encountered)

	-I <directory>	input directory (<scenario>_sam_list.txt and .mtx files are expected here)
	-O <directory>	output directory
	-D <directory>  location of activity and constraint description files
	-M <filename>	MilpCheckExe to be used
	(if not specified, the directories are read from .mqlmatrix entries samdir, outdir, infofiles, milpcheck.
	If these do not exist either, they default to the current working directory and 
	milpcheck defaults to /usr/local/bin/MilpCheckExe)

	-h, -?, --help	print this help
	-v		print version information

HELP
			exit (0);
		}
		elsif ($element =~ /^-v/ ) {
			print $longwelcome;
			exit(0);
		}
	}
	elsif ($parselevel == 1) {
		if ($element =~ /-G$/) {
			$parselevel = 0;
		}
		else {
			my @groups = split (/\s*,\s*/, $element);
			push @grouplist ,  @groups ;
		}
	}
}
my $milplist = IndexedTable->create_from_ascii($indir.$samlist, 2, "samlist");
my $solutiontab = IndexedTable->create_empty(3, "solutiontab");
my %grouptabs;
for my $g (@grouplist) {
	$grouptabs{$g} = IndexedTable->create_empty(3, "$g");
}

for my $scen ($milplist->query([],0 )) {
	my $actdesname = $desdir.$scen."_ActivityInfo.txt" if (not defined $actdesname);
	my $condesname = $desdir.$scen."_ConstraintInfo.txt" if (not defined $condesname);
	my $actinfo = IndexedTable->create_from_ascii($actdesname, 1, "activity_description");

	for my $ag ($milplist->query([$scen],0 )) {
		print "Solving $scen, $ag\n";
		my $fn = $indir.$scen."_MILP_".$ag;
		my $string = $milpcheck." ".$fn." -mode 1";
		system($string);
		my $solutionfile = $fn.".pri";
		open (my $fh, "<", $solutionfile) or (print STDERR "Error when trying to open $solutionfile: $!" and exit(1) );
		print "Importing Solution Vector from $solutionfile\n";
		my $cc = 0;
		my @rows;
		while (<$fh>) {
			chomp;
			my $line = $_;
			next if ($line =~ /^\s*$/); #skip empty lines	

			my @colvalues;
			if ($line =~ /\s*Target value:\s*([-\d.E+e]*)/) {
				@colvalues = ($scen, $ag, "orow",$1);
			}
			elsif ($line =~/^\s*?((\s|-)[-\d.E+e]+)/) {
				my $sol = $1;
				 if (sprintf("%.4f", $sol) > 0) {
					my $actcode = $actinfo->query([$cc],2);
					@colvalues = ($scen, $ag, $actcode, sprintf("%.4f", $sol));
					for my $g (@grouplist) {
						my $gcn = $actinfo->column_by_name($g);
						my $isgroup = $actinfo->query([$cc], $gcn);
						unless ($isgroup eq "-1") {
							my $curval = $grouptabs{$g}->query([$scen, $ag, $isgroup],4) || 0;
							$grouptabs{$g}->set_column_value([$scen, $ag, $isgroup],4, sprintf("%.4f",$curval + $sol));
					
						}
					}
				}
				$cc++;
			}
			push @rows, \@colvalues if (scalar @colvalues > 0);
		}
		close $fh;

		$solutiontab->add_rows(\@rows);
		
	}
}
$solutiontab->printme($outdir.$prefix."solution.tab");
for my $k (keys %grouptabs) {
	$grouptabs{$k}->printme($outdir.$prefix."${k}_solution.tab");
}

