#!/usr/bin/perl
use strict;
use warnings;

##Version:
my $thisversion = "v1.2";
my $lastchange = "08-02-2013";

my $shortwelcome = <<WC2;
runcompact $thisversion		
(c)	Christian Troost, Universität Hohenheim, 2012-2013

WC2
my $scen;
my $infolder = "./";
my $com = " ";
my $run = 1;
while (@ARGV) {
  my $el = shift @ARGV;
  if ($el =~ /^-N(\w+)/) {
    $scen = $1;
  }
  elsif ($el =~ /^-I(.+)/) {
    $infolder= $1;
  }
  elsif ($el =~ /^-!/) {
    $run = 0;
    next;
  }
  elsif ($el =~ /^-v/) {
    print $shortwelcome;
    exit(0);
  }
 elsif ($el =~ /^-\?/ or $el =~ /^-h/ or $el =~ /^--help/) {
  print $shortwelcome;
  print <<HELP;
Tool to convert .cpt matrix format (created with mpmasql --compact)
to full MPMAS format .dat before running the model.

Usage:
  runcompact <full MPMAS commandline>

Alternative usage:
  runcompact -! -N<scenario_name> -I<folder where input/dat is located>

(only converts the model, without starting the simulation).
HELP
  exit(0);
  }
  $com .= " ".$el;
}

my $if;
my $of;

my @files = glob($infolder."/input/dat/".$scen."MILP*.cpt");
for my $ifname (@files) { 

    my $ofname = $ifname;
    $ofname =~ s/\.cpt$/.dat/;

    open ($if, "<", $ifname) or print STDERR "$!\n" and exit(1);
    open ($of, ">", $ofname) or print STDERR "$!\n" and exit(1);

    my $part = "pre";
    my ($NumRow, $NumCol, $percfilled, $NumCoef);
    my @tableau;
    my $cc = 0;
    my $ci = 0;
    LINES: while (<$if>) {	
	    my $line = $_;
	    if ($part eq "pre") {
		    print $of $line;
		    next LINES if ($line =~ /^\s*$/);
		    if ($line =~ /^\s*\d*/ ) {
			    chomp($line);
			    $line =~ s/^\t*//;
			    $line =~ s/^\s*//;
			    $line =~ s/\s*$//;
			    ($NumRow, $NumCol, $percfilled) = split /\s/, $line;
			      @tableau = map { [ map { 0 } 0 .. $NumCol -1 ] } 0 .. $NumRow -1;
			    $part = "mpmas_info_block";
		    }
		    else {
			    print STDERR "File does not fit Mpmas format\n" and exit(1);
		    }
		    
	    }
	    elsif ($part eq "mpmas_info_block") {
			    print $of $line;
			    if ($line =~ /^\s*\d+/ ) {
				$cc++;
			    }
			    if ($cc == 6) {
				$cc = 0;
				$part = "shape"
			    }
	    }
	    elsif ($part eq "shape") {
		    #can for now only handle one block
		    print $of $line;
		    next LINES if ($line =~ /^\s*$/);
		    if ($line =~ /^\s*\d/ and $cc == 0) {
			    $line =~ s/^\s*//;
			    $line =~ s/\s*$//;
			    my ($blocks, $trans) = split /\s+/, $line;
			    print STDERR "Blocked matrices and transposed matrices are not yet supported\nRead blocks: $blocks, transposed: $trans\n"  and exit(1) unless ($blocks == 1 and $trans == 0);
			    $part = "orow";
			    $cc= 0;
		    }
	    
	    }
	    elsif ($part eq "orow") {
		    print $of $line;
		    next LINES if ($line =~ /^\s*$/);
		    if ($line =~ /^\s*(-?\d)/) {
			    $part = "tableau";
		    }
	    }
	    elsif ($part eq "tableau" ) {
			    if (not defined $NumCoef and $line =~ /^\s*(\d+)/) {
				  $NumCoef = $1;
			    }
			    elsif ($line =~ /^\s*(-?\d)/) {
				  $line =~ s/^\s+//;
				  my ( $row, $col, $coef) = split /\s+/, $line;
				  $tableau[$row][$col] = $coef;
				  $cc++;
			    }
			    if (defined $NumCoef and $cc >= $NumCoef) {
				    $part = "rest";
				    $cc = 0;
				    print $of "\n";
				    for my $r (@tableau) {
					    print $of "\t\t\t\t\t\t\t\t\t\t\t".join ("\t",@$r)."\n";
				    }
			    }
	    }
	    elsif ($part eq "rest") {
		    print $of $line;      
	    }
    }

    close($if);
    close($of);
}
#rename ($infolder."/input/dat/".$scen."MILP.dat" , $infolder."/input/dat/".$scen."MILP.cpt");
#rename ($infolder."/input/dat/".$scen."MILP.tmp" , $infolder."/input/dat/".$scen."MILP.dat");
exec ($com) if $run;
