#!/usr/bin/perl
package IndexedTable;
use strict;
use warnings;
use SimpleObject;
use SimpleObject;
use Carp qw(croak);
use Storable qw(dclone store retrieve);
use Data::Dumper;
use Hash::Merge qw(merge);

our @ISA = qw(SimpleObject);

sub create_from_SQL {
	my ($class, $connection, $sqlstring, $numkeys, $id) =  @_; 
	# $connection:	DBI object of open connection
	# $sqlstring:	an sql string resulting in a table
	# $numkeys:	the number of primary key columns (have to be the first columns to be returned from the database)
	# $id:		ID with which the table can be referred to in registries
	
	## conditional to load query result into $retrieved_array
	my $retrieved_array; 
	eval{	#exception handling to avoid fatals because table column does not exist 
								
		local $SIG{__WARN__}=sub{}; #Temporarily suppress warnings
		$retrieved_array = $connection->selectall_arrayref($sqlstring); #temporary array to store query result
	};
	if ($@) {
		croak "$@\n For SQL: $sqlstring\n";
	}
							
	my $hash = create_index_structure($retrieved_array, $numkeys, $sqlstring);

	my $self = {};
	$self->{numkeys} = $numkeys;
	$self->{table} = $hash;
	$self->{source} = $sqlstring;
	$self->{id} = $id;
	
	bless $self, $class;
	
}

sub create_index_structure {
	my ($retrieved_array , $numkeys , $sqlstring) = @_;
	my $hash = {};
	for my $j (@{$retrieved_array}){
		my $rt = $hash;
		if (@{$j} < $numkeys) {
			warn "\"$sqlstring\"  results in a table lacking necessary columns.\n" and return undef;
		}
		my $lastcolumn = $#{$j};
		my (@keylist, @columns);
		$numkeys > 1 ? @keylist = @{$j}[0..$numkeys-1] : $keylist[0] = ${$j}[0];
		if ($numkeys < $lastcolumn) {
			@columns = @{$j}[$numkeys..$lastcolumn]
		}
		elsif ($numkeys == $lastcolumn) {
			$columns[0] = ${$j}[$numkeys];
		}
		else {
		      $columns[0] = 1;
		}

		my $lastkey = pop @keylist;
		foreach my $key (@keylist) {
			if (not defined $key) {
				$key = "NA";
				print join(":",@keylist) ." Undefined key set to NA! \n$sqlstring\n";
			};
			if (not exists($rt->{$key})) {
				$rt->{$key} = {};
			}

			$rt = $rt->{$key};
		}
		if (not defined $lastkey) {
			$lastkey = "NA"; 
			print join(":",@keylist) ." Undefined key set to NA! \n$sqlstring\n";
		};	
  		$rt->{$lastkey} = \@columns;	
	}
	return $hash;
}
sub create_from_ascii {
	my ($class, $file, $numkeys, $id, $separator, $header) = @_;
	croak "Error: Create_from_ascii is a class method, you have called it from an object!\n" if (ref $class) and return undef;
	$separator = '\t' unless (defined $separator);
	$header = 1 unless (defined $header);
	
	#print "Reading data from $file...\n";
	open (my $infile, "<", $file) or (croak "Error when opening $file: $! \n\n" and return undef);
	my $i = 0;
	my @table_header;
	my @rows;
	while (<$infile>) {
		chomp;
		my $Line =  $_;
		next if ($Line =~ /^\s*$/); #skip empty lines
		$i++;

		if ($i == 1 and $header) {
			@table_header = split /$separator/, $Line;
		}
		else {
			my @currow = split /$separator/, $Line;
			push @rows, \@currow;
		}
	}
	close ($infile);
	my $hash = create_index_structure(\@rows, $numkeys, $file);

	my $self = {};
	$self->{numkeys} = $numkeys;
	$self->{table} = $hash;
	$self->{id} = $id;
	
	bless $self, $class;
	if ($header){
		$self->add_columnnames(@table_header);
	}
	return $self;
}
sub create_from_hash {
	my ($class, $hash, $numkeys, $id) = @_;
	croak "Error: Create_from_hash is a class method, you have called it from an object!\n" if (ref $class) and return undef;

	my $self = {};
	$self->{numkeys} = $numkeys;
	$self->{table} = $hash;
	$self->{id} = $id;
	bless $self, $class;
	return $self;
}
sub create_empty {
	my $class = shift;
	my $numkeys = shift;
	my $id = shift;
	my $header = shift;
	my @table_header;
	if (defined $header) {
		@table_header = @$header;
	}

	my $hash = {};

	my $self = {};
	$self->{numkeys} = $numkeys;
	$self->{table} = $hash;
	$self->{id} = $id;
	
	bless $self, $class;
	if ($header){
		$self->add_columnnames(@table_header);
	}
	return $self;

}
sub regrouped_copy {
    my $self = shift;
    my $newkeysarray = shift;
    my $id = shift;
    croak "Error: IndexedTable $self->id is empty\n" if ($self->is_empty);
    my $numkeys = scalar @$newkeysarray;    
    my $arrayref = $self->as_array_ref;
    my @colstofront = map {$_ - 1} (@$newkeysarray);
    my @colsatend =  (0..$#{$arrayref->[0]});
    my $cc = 0;
    foreach  (@colstofront) {
	splice @colsatend, $_ - ($cc++), 1;
    }
    my @newarray = map {[( @{$_}[@colstofront],  @{$_}[@colsatend]  ) ]  } @$arrayref;

    my $hash = create_index_structure(\@newarray, $numkeys, "regrouping ". $self->{id});
    my $new = {};
    $new->{numkeys} = $numkeys;
    $new->{table} = $hash;
    $new->{id} = $id;
    bless $new, ref $self;
    return $new;
}
sub add_rows {
	my $self = shift;
	my $rows = shift;
	my $hash_to_add = create_index_structure($rows, $self->numkeys, "");
	if (defined $hash_to_add) {
		eval {	#exception handling
								
			local $SIG{__WARN__}=sub{}; #Temporarily suppress warnings
			my $merge = Hash::Merge->new('LEFT_PRECEDENT');
			my $temp  =   $merge->merge ( $hash_to_add, $self->{table} );
			$self->{table} = $temp;
		};
		if ($@) {
			croak "$@"."\n\n";
		}
	}

}
sub add_row {
	my $self = shift;
	my $rows = shift;
	my @rowarray = ($rows);
	eval {	#exception handling to avoid fatals because table column does not exist 						
		local $SIG{__WARN__}=sub{}; #Temporarily suppress warnings
		$self->add_rows(\@rowarray);
	};
	if ($@) {
		croak "$@"."\n\n";
	}
}

sub query {
	my ($self , $keysref, $column) = @_;
	my @keylist = @{$keysref};
	#if (@keylist == 0) {
	#	die "You did not specify a key when requesting table entry\n" ;
	#}
	if (@keylist > $self->{numkeys}) {
		croak "Table entry referred to by more keys than defined for table\n" ;
	}
	#my $lastkey = pop @keylist;
 	my $rt = $self->{table};
	
	foreach my $key (@keylist) {
		croak "Undefined key in keylist of query to IndexedTable!\n" unless (defined $key);
		 #print "no value found for $key in table.\n" and 
		return undef if (not exists($rt->{$key}));
		$rt = $rt->{$key};
		
	}

	if (@keylist == $self->{numkeys} and $column >= 0) {
		(croak "Column $column is a key field in table $self->{id}\n" and return undef) if ($column - $self->{numkeys} -1 < 0) ;
		
		if (ref $rt eq "ARRAY") {
			(croak "Column $column does not exist in table $self->{id}\n"  and  return undef) if ($column - $self->{numkeys} -1 > $#{ $rt }) ;
			return $rt->[$column -$self->{numkeys} - 1];
		}
		elsif (not ref $rt) {
			return $rt;
		}
	}
	elsif 	(@keylist == $self->{numkeys} and $column == -1) {
		return @$rt;
	}
	elsif (@keylist < $self->{numkeys}) {
		return keys(%{ $rt });
	}
}
sub query_by_columnname {
	my ($self , $keysref, $columnname) = @_;
	my $column = $self->column_by_name($columnname);
	return $self->query($keysref,$column);
}

sub column_by_name {
	my ($self ,  $columnname) = @_;
	(warn "Error: Columns in IndexedTable $self->{id} are not named\n" and return undef) unless (exists $self->{columns});
	(warn "Error: Column $columnname does not exists in IndexedTable $self->{id}\n" and return undef) unless (exists $self->{columns}->{$columnname});
	(warn "Error: Column $columnname is a key column in IndexedTable $self->{id}\n" and return undef) unless ( $self->{columns}->{$columnname}  > $self->{numkeys});
	
	my $column = $self->{columns}->{$columnname};
	return $column;
}
sub exists_columnname {
	my $self = shift;
	my $columnname = shift;
	return (exists $self->{columns}->{$columnname});
}
sub is_empty {
  return 1 if (scalar (keys %{$_[0]->{table}} )== 0);
  return 0;

}

sub add_columnnames {
	my $self = shift;
	my @columnnames = @_;
	$self->{header} = \@columnnames;	
	for my $i (0.. $#columnnames) {
		my $cname = $columnnames[$i];
		$self->{columns}->{$cname} = $i + 1;	
	}
	return 1;
}

sub printtable {
	my $self = shift;
	print Dumper($self->{table});
	
}
sub as_array_ref {
      my $self = shift;
      my @array;
      my @curKeyArray;
      my @PointerAtLevel;
      my $numkeys = $self->numkeys;
	$#PointerAtLevel = $numkeys + 1;
	@PointerAtLevel = (0) x @PointerAtLevel; 
        my $level = 1;
	PRINTOUTER: while ($level > -1) {
		PRINTINNER: while ($level <= $numkeys) {
			my @keys = $self->query(\@curKeyArray,0);

			if ($PointerAtLevel[$level] < @keys) {
				push @curKeyArray, $keys[$PointerAtLevel[$level]];
				$PointerAtLevel[$level] += 1;
				$level++;
			}
			elsif  ($level == 1 and  $PointerAtLevel[$level] >= @keys ) {
				$level = -1;
				last PRINTOUTER;
			}
			else {
				$PointerAtLevel[$level] = 0;
				$level--;
				pop @curKeyArray;
			}			

		}
		my @cols = $self->query(\@curKeyArray,-1);
		push @array, [(@curKeyArray, @cols )];
		pop @curKeyArray;
		$level--;
	}
      return \@array;
}
sub printme {
	my $self = shift;
	my $file = shift;
	my $splitter = shift || "\t";
	my $out;
	if (defined $file) {
		open ($out, ">", $file ) or croak "Error: Could not open $file: $!\n";
	}
	else {
		open ($out, ">&STDOUT" );
	}
	
	map { print $out join($splitter, @{$_})."\n" } @{$self->as_array_ref};

	close($out);
}
sub print_to_fh {
	my $self = shift;
	my $out = shift;
	my $splitter = shift || "\t";
	map { print $out join($splitter, @{$_})."\n" } @{$self->as_array_ref};
}
sub set_column_value {
	my $self = shift;
	my $keysref = shift; #reference to anonymous array with keys
	my $column = shift;
	my $value = shift;
	my @keylist = @{$keysref};
	#if (@keylist == 0) {
	#	die "You did not specify a key when requesting table entry\n" ;
	#}
	if (@keylist > $self->{numkeys}) {
		croak "Table entry referred to by more keys than defined for table\n" ;
	}
	#my $lastkey = pop @keylist;
 	my $rt = $self->{table};
	my $lastkey = pop @keylist;
	foreach my $key (@keylist) {
		 #print "no value found for $key in table.\n" and 
		#croak "No entry found for $key in table.\n" if (not exists($rt->{$key}));
		if (not exists($rt->{$key})) {
			$rt->{$key} = {};
		}
		$rt = $rt->{$key};
	}

	if (@keylist+1 == $self->{numkeys}) {
		(warn "Column $column is a key field in table $self->{id}\n" and return undef) if ($column - $self->{numkeys} -1 < 0) ;
		$rt->{$lastkey} = [] unless (exists $rt->{$lastkey});
		$rt->{$lastkey}->[$column -$self->{numkeys} - 1] = $value;
		
	}
	elsif (@keylist +1 < $self->{numkeys}) {
		croak "Not enough keys supplied\n";
	}
}
sub numkeys {
    return $_[0]->{numkeys};
}

1;
