package MpmasBData;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use MpmasInputfile;
use MpmasInputBlock;
our @ISA = qw(MpmasInputfile);

sub filename {
	return "BData";
}


sub prepare_special {
	my $self = shift;	
	#####Consistency checks
	print "Inconsistent setting of CONTROL switches: OFFINVDELAY > 0 and OFFINV > 0. \n" if ($self->config_entry("OFFINVDELAY") > 0 and $self->config_entry("OFFINV") > 0);
	

	####Prepare Input Blocks
	##General Parameters Block		
	my $block1 = MpmasInputBlock->new(3, 10);  # number of commentary columns, priority
	
	$block1->add_row("THE BASIC DATA FILE");
	$block1->add_row("1","GENERAL PARAMETERS","","");
	$block1->add_row("","1.01","Maximum number of sub-basins (or districts)",$self->config_entry("DISTRICTS"));
	$block1->add_row("","1.02","Maximum number of micro-basins (or villages)",$self->config_entry("VILLAGES"));
	$block1->add_row("","1.03","Number of innovation networks",$self->config_entry("NETWORKS"));
	$block1->add_row("","1.04","Number of segments per network",$self->config_entry("SEGMENTS"));
	$block1->add_row("","1.05","Number of agent populations",$self->config_entry("POPULATIONS"));
	$block1->add_row("","1.06","Number of clusters per population",$self->config_entry("CLUSTERS"));
	$block1->add_row("","1.07","Number of loops before lottery is stopped",$self->config_entry("LOTTERYLOOPS"));
	$block1->add_row("","1.08","Minimum share of water demand to be met in lottery",$self->config_entry("MINWATER"));
	$block1->add_row("","1.09","Number of producer organizations",$self->config_entry("PRODUCER_ORGS")|| 0);
	$block1->add_row("","1.10","Number of regional markets",$self->config_entry("MARKETS") || 1);
	$block1->add_row("","","","");
	$block1->add_row("2","SWITCHES FOR VARIOUS SUB-MODELS","","");
	$block1->add_row("","2.01","Type of consumption model",$self->config_entry("CONSTYPE"));
	$block1->add_row("","2.02","Type of crop growth model",$self->config_entry("CROP_MODEL"));
	$block1->add_row("","2.03","Type of hydrology model",$self->config_entry("LANDSCAPE_MODEL"));
	$block1->add_row("","2.04","Livestock model",$self->config_entry("LIVESTOCK_MODEL"));
	$block1->add_row("","2.05","Permanent crops model",$self->config_entry("PERMCROP_MODEL"));
	$block1->add_row("","2.06","Internal land market",$self->config_entry("LANDMARKET_MODEL"));# 0 = landNoTransfer,1 = landToOtherFarms,	2 = landRentalMarket,	3 = landLongTermMarket
	$block1->add_row("","2.07","Type of lottery input",$self->config_entry("ASSETDISTTYPE")||0);
	$block1->add_row("","2.08","Country specific module",$self->config_entry("COUNTRY"));
	$block1->add_row("","2.09","Use of intermediate products",$self->config_entry("HARVEST_DECISION"));
	$block1->add_row("","2.10","Subversion of biophysical modules",$self->config_entry("BIOVERSION")||0);# 0 -default , 1 - FAO56 for EDIC Chile, 2 - FAO56 with flood factor, 3 - External yields from file
	$block1->add_row("","2.11","Policy intervention",$self->config_entry("POLICY")||0); #0 - none; 1 - irrigation infrastructure projects
	$block1->add_row("","2.12","Irrigation",$self->config_entry("IRRIGATION")||0); #0 - no; 1 - yes
	if ( $self->config_entry("COUNTRY") == 5) {
	      $block1->add_row("","2.13","Demography model",$self->config_entry("DEMOGRAPHY_MODEL")||0); #0 - no; 1 - yes
	      $block1->add_row("","2.14","Exogenous yield expectations",$self->config_entry("EXO_YIELD_EXPECTATIONS")||0); #0 - no; 1 - yes
	}
	$block1->add_row("","","","");
	$block1->add_row("3","SOIL INFORMATION","","");
	$block1->add_row("","3.01","Size of 1 grid cell (=parcel)",$self->config_entry("CELLSIZE"));
	$block1->add_row("","3.02","Unit transport costs",$self->config_entry("TCOST"));
	$block1->add_row("","3.03","Number of soil types",$self->config_entry("NSOIL"));
	$block1->add_row("","3.04","Number of nutrient response units (NRUs)",$self->config_entry("NNRU"));
	$block1->add_row("","3.05","Static assignment of NRUs (optional for TSPC)",$self->config_entry("STATICNRU"));
	$block1->add_row("","","","");





	$block1->add_row("4","INNOVATION PARAMETERS","","");
	$block1->add_row("","4.01","Tentative number of farm assets",$self->config_entry("NFARMASSETS"));
	$block1->add_row("","4.02","Constraints on the random asset assignment","");
	$block1->add_row("","4.03","Maximum age of assets in period 0",$self->config_entry("ASSETMAXAGE"));
	$block1->add_row("","4.04","Minimum age of assets in period 0",$self->config_entry("ASSETMINAGE"));
	$block1->add_row("","","Diffusion model","");
	$block1->add_row("","4.05","Type of communication process",$self->config_entry("COMTYPE"));
	$block1->add_row("","4.06","Cumulative adoption  (\"classical\" diffusion curve)",$self->config_entry("CUMADOPT"));
	$block1->add_row("","4.07","Parameter for expectations, see Brandes et al. (1997: 382)",$self->config_entry("PAREX"));
	$block1->add_row("","","Fine-tuning of innovation diffusion","");
	$block1->add_row("","4.08","Safe rounding factor for indivisible assets",$self->config_entry("SAFEROUND"));
	$block1->add_row("","4.09","Minimum investment in permanent crops [ha]",$self->config_entry("MINPERM"));
	$block1->add_row("","4.10","Overlap of network thresholds",$self->config_entry("OVERLAP"));
	$block1->add_row("","4.11","Liquidity reserve factor",$self->config_entry("LIQRESERVE"));
	$block1->add_row("","4.12","Number of \"symbolic\" investment objects",$self->config_entry("SYMBOLICS"));
	$block1->add_row("","4.13","Currency units",$self->config_entry("CURRENCYSCALE"));
	$block1->add_row("","","","");

	if ($self->config_entry("LANDMARKET_MODEL") >= 2) {
		$block1->add_row("5","RENTAL MARKETS","","");
		$block1->add_row("","5.01","Minimum bid (relative to average shadow prices)",$self->config_entry("LMMINBIDREL"));
		$block1->add_row("","5.02","Minimum bid (absolute value)",$self->config_entry("LMMINBIDABS"));
		$block1->add_row("","5.03","Maximum bid (relative to average shadow prices)",$self->config_entry("LMMAXBIDREL"));
		$block1->add_row("","5.04","Maximum bid (absolute value)",$self->config_entry("LMMAXBIDABS"));
		$block1->add_row("","5.05","Markup for renting-out decisions",$self->config_entry("LMMKUPOUT"));
		$block1->add_row("","5.06","Markup factor for renting-in decisions",$self->config_entry("LMMKUPIN"));
		$block1->add_row("","5.07","Number of renting-out attempts",$self->config_entry("LMTRYOUT"));
		$block1->add_row("","5.08","Number of renting-in attempts",$self->config_entry("LMTRYIN"));
		$block1->add_row("","5.09","Maximum distance for placing bid in grid cells)",$self->config_entry("LMMAXDIST"));
		$block1->add_row("","5.10","Duration of rental contract (years)",$self->config_entry("LMDURATION"));
		$block1->add_row("","5.11","Share of rental payment paid after harvest",$self->config_entry("LMSHPAYEND"));
		$block1->add_row("","5.12", "Number of existing rental contracts", $self->config_entry("LM_numcontracts_start") || 0);
		$block1->add_row("","5.13", "Rental market activity", $self->config_entry("LMTRADING_ON"));
		if( $self->config_entry("COUNTRY") == 5 ) {
		      $block1->add_row("","5.14", "No renting out by farm households", $self->config_entry("LMNORENTOUT" || 0));
		}
		$block1->add_row("","","","");
	}
	if ($self->config_entry("POLICY") == 1) {

		$block1->add_row("6","POLICY PARAMETERS","","");
		$block1->add_row("","6.01","Maximum age of subsidized permanent crops",$self->config_entry("MAXAGEPERM"));
		$block1->add_row("","6.02","Duration of policy measure",$self->config_entry("DURPOLICY"));
		$block1->add_row("","6.03","Initial number of irrigation infrastructure projects",$self->config_entry("N_INFPROJ") || 0);
		$block1->add_row("","","","");
	}

	$block1->add_row("7","PROCESS SWITCHES","","");
	$block1->add_row("","7.01","Debug information follows (yes/no)","1");
	$block1->add_row("","7.02","Number of spin-up rounds",$self->config_entry("SPINUPS"));
	$block1->add_row("","7.03","Switch off updating of liquidity and equity",$self->config_entry("OFFLIQEQ"));
	$block1->add_row("","7.04","Switch off aging of household members",$self->config_entry("OFFHHAGING"));
	$block1->add_row("","7.05","Switch off aging of livestock",$self->config_entry("OFFLIVEAGE"));
	$block1->add_row("","7.06","Switch off aging of assets",$self->config_entry("OFFASSAGE"));
	$block1->add_row("","7.07","Switch off investments (immediately)",$self->config_entry("OFFINV"));
	$block1->add_row("","7.08","Switching off investments (delayed by)",$self->config_entry("OFFINVDELAY"));
	$block1->add_row("","7.09","Switch off updating of soil conditions and processes",$self->config_entry("OFFDYNSOIL"));
	$block1->add_row("","7.10","Switch off dynamic inflows",$self->config_entry("OFFDYNFLOW"));

	$block1->add_row("","","","");
	$block1->add_row("8","FINE-TUNING OF THE SOLVER","","");
	$block1->add_row("","8.01","Minimum objective value",$self->config_entry("MINOBJ"));
			
	$self->add_block($block1);		

	##Milp Debugging block
	my $milpblock = MpmasInputBlock->new(3, 20);

	my $ms = 0;
	if ($self->ini_entry("DBGMILPS") ) {
		for my $msp (@{$self->ini_entry("DBGMILPS")}){
			print "\t>>Warning!!!: Entries in DBGMILPS  have to be integers!!!\n\t\>>Value $msp has been skipped!!\n" and next unless $msp =~ /^-?\d*$/;
			$ms += 1;
			my $nmsp = $msp;
			$nmsp = $nmsp * -1 if ($nmsp > 0);
			$milpblock->add_row("","","",$nmsp);
		}
	}
	$milpblock->add_row_at_beginning("","8.02","Number of MILP problems to be saved","$ms");

	$self->add_block($milpblock);


	## Final Block
	my $lastblock = MpmasInputBlock->new(3, 30);
	
	$lastblock->add_row("","","","");
	
	$lastblock->add_row("","8.03","Fine-tuning information (yes/no)","1");
	$lastblock->add_row("","8.04","Maximum time allowed for solving (in seconds)", $self->config_entry("OSL_MAXSOLVETIME"));
	$lastblock->add_row("","8.05","Maximum iterations allowed for solving", $self->config_entry("OSL_MAXITERATE"));
	$lastblock->add_row("","8.06","Maximum number of nodes allowed for solving", $self->config_entry("OSL_MAXNODES"));
	$lastblock->add_row("","8.07","OSL mode sequencing ", $self->config_entry("OSL_SEQUENCE"));
	if ($self->config_entry("OSL_SEQUENCE") == 4) {
	      $lastblock->add_row("","8.08","Max. number of attempts", $self->config_entry("OSL_M4_ATTEMPTS"));
	      $lastblock->add_comment_row("","","Limits on first attempts");
	      $lastblock->add_row("","8.09","max. integers", $self->config_entry("OSL_M4_ITHRESH"));
	      $lastblock->add_row("","8.10","max. nodes", $self->config_entry("OSL_M4_NTHRESH"));
	      $lastblock->add_row("","8.11","max. feasible solutions", $self->config_entry("OSL_M4_STHRESH"));

	      $lastblock->add_comment_row("","","Presolving of relaxed LP (EKKPRSL)");
	      $lastblock->add_row("","8.12","...from attempt", $self->config_entry("OSL_M4_LP_PRESOLVE"));
	      $lastblock->add_row("","8.13","...if previous attempt infeasible", $self->config_entry("OSL_M4_LP_PRESOLVE_ONFAIL"));
	      
	      $lastblock->add_comment_row("","","Scaling (EKKSCAL)"); 
	      $lastblock->add_row("","8.14","...from attempt", $self->config_entry("OSL_M4_SCALE"));
	      $lastblock->add_row("","8.15","...if previous attempt infeasible", $self->config_entry("OSL_M4_SCALE_ONFAIL"));
	      
	      $lastblock->add_comment_row("","","Crash to find feasible basis (EKKCRSH)"); 
	      $lastblock->add_row("","8.16","OSL crash mode", $self->config_entry("OSL_M4_CRASHMODE"));
	      $lastblock->add_row("","8.17","...from attempt", $self->config_entry("OSL_M4_CRASH"));
	      $lastblock->add_row("","8.18","...if previous attempt infeasible", $self->config_entry("OSL_M4_CRASH_ONFAIL"));

	      $lastblock->add_comment_row("","","Integer presolving (EKKMPRE)"); 
	      $lastblock->add_row("","8.19","OSL integer strategy (Istrategy bit mask)", $self->config_entry("OSL_M4_I_STRATEGY"));
	      $lastblock->add_row("","8.20","OSL number of heuristic passes", $self->config_entry("OSL_M4_I_HEURPASS"));
	      $lastblock->add_row("","8.21","Max rows to add for integer presolving", $self->config_entry("OSL_M4_I_PRE_ADDROWS"));
	      $lastblock->add_row("","8.22","...from attempt", $self->config_entry("OSL_M4_I_PRESOLVE"));
	      $lastblock->add_row("","8.23","...if previous attempt infeasible", $self->config_entry("OSL_M4_I_PRESOLVE_ONFAIL"));

	      $lastblock->add_comment_row("","","Accept first attempt if");
	      $lastblock->add_row("","8.24","difference to best estimated < ", $self->config_entry("OSL_M4_ACCEPT_DIFF"));
	      $lastblock->add_row("","8.25","or rel. difference to best estimated < ", $self->config_entry("OSL_M4_ACCEPT_RELDIFF"));
	      $lastblock->add_row("","8.26","Try again after infeasible attempt ", $self->config_entry("OSL_M4_CONT_ONFAIL"));
	      $lastblock->add_row("","8.27","Continue in mode 3 if mode 4 failed", $self->config_entry("OSL_M4_MODE3_ONFAIL"));
	      $lastblock->add_row("","8.28","Log solving in mode 4", $self->config_entry("OSL_M4_LOG"));

	}
	$self->add_block($lastblock);
}

1;
