package MpmasClusterSet;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use SimpleObjectRegistry;
use MpmasInputBlock;
use SimpleObject;
use IndexedTable;
use CtUtilities qw(queryhash extqueryhash strip_sql);
our @ISA = qw(MpmasInputfileSet);
use MpmasCluster;
use MpmasAssets;

sub class_of_files{
	return "MpmasCluster";
 }

sub determine_files {
	my $self = shift;
	for my $p (0..$self->config_entry("POPULATIONS") - 1) {
		for my $c (0..$self->config_entry("CLUSTERS") - 1) {
			my $fname = "$p"."_"."$c";
			
			$self->add_file($fname,$fname);
			$self->add_file($fname,$fname, "MpmasAssets") if ($self->config_entry("ASSETDISTTYPE") == 1);
		}
	}

}

sub prepare_shared_files {
	my $self = shift;
	my $dbh = $self->get_dbh;
	print STDERR "Error: Missing HOUSEHOLD_MEMBER_TYPES [MPMAS TABLES]\n" and exit(1) unless (defined $self->config_entry("HOUSEHOLD_MEMBER_TYPES"));
	my $sexagedef = $self->config_entry("HOUSEHOLD_MEMBER_TYPES")->transform($dbh, 1);
	
	my $householdhash;
	my $assethash;
	my $agentinfo;
	my $agentsinclu;
	if ($self->config_entry("ASSETDISTTYPE") == 0) {
		  print STDERR "\nError: Missing HOUSHOLD_CDF [MPMAS TABLES].\nThis is required because ASSETDISTTYPE has been set to 0 (random allocation of assets within MPMAS)\n"  and exit(1) unless (defined $self->config_entry("HOUSEHOLD_CDF"));
		  $householdhash = $self->config_entry("HOUSEHOLD_CDF")->transform($dbh, 4);
		  print STDERR "\nError: Missing ASSET_CDF [MPMAS TABLES].\nThis is required because ASSETDISTTYPE has been set to 0 (random allocation of assets within MPMAS)\n"  and exit(1) unless (defined $self->config_entry("ASSET_CDF"));
		  $assethash = $self->config_entry("ASSET_CDF")->transform($dbh,4);
	}
	elsif ($self->config_entry("ASSETDISTTYPE") == 1) {
		  print STDERR "\nError: Missing HOUSEHOLD_COMPOSITION [MPMAS TABLES].\nThis is required because ASSETDISTTYPE has been set to 1 (prepared asset distribution provided)\n"  and exit(1) unless (defined $self->config_entry("HOUSEHOLD_COMPOSITION"));
		  print STDERR "\nError: Missing ASSET_ENDOWMENTS [MPMAS TABLES].\nThis is required because ASSETDISTTYPE has been set to 1 (prepared asset distribution provided)\n"  and exit(1) unless (defined $self->config_entry("ASSET_ENDOWMENTS"));
		  $householdhash = $self->config_entry("HOUSEHOLD_COMPOSITION")->transform($dbh,2);
		  $assethash = $self->config_entry("ASSET_ENDOWMENTS")->transform($dbh,2);
		  if ($self->config_entry("AGENT_INFO_FROM_MAPS")) {
			    my $mapdir = "gis/";
			    $mapdir = $self->config_entry("MAPS") if (defined $self->config_entry("MAPS"));
			    $mapdir = $self->outdir ."/../". $mapdir;
			     $mapdir = $mapdir ."/" unless ($mapdir =~/\/$/);
			     use EnhancedGrid;
			      my $fstmap = EnhancedGrid->create_from_ascii($mapdir."CatchMap00Farm.txt");
			      my $clumap = EnhancedGrid->create_from_ascii($mapdir."CatchMap00Clu.txt");
			      my $popmap = EnhancedGrid->create_from_ascii($mapdir."CatchMap00Pop.txt");
			      my $clutab = IndexedTable->create_from_hash($clumap->tabulate_values_by_grid($fstmap), 2, "clutab");
			      my $poptab = IndexedTable->create_from_hash($popmap->tabulate_values_by_grid($fstmap), 2, "poptab");

			      my %infohash;
			      for my $i ($clutab->query([],0)) {
				    my @p = $poptab->query([$i],0);
				    my @c = $clutab->query([$i],0);
				    $infohash{$i} = [( $p[0], $c[0], -1,-1,-1,-1)];
			      }
			      $agentinfo = IndexedTable->create_from_hash(\%infohash,1,"agentinfo");

			      $self->set_config_entry("MAP_X_EXTENT", $fstmap->ncols() );
			      $self->set_config_entry("MAP_Y_EXTENT", $fstmap->nrows() );
		  }
		  else {
		      print STDERR "\nError: Missing AGENT_INFO [MPMAS TABLES].\nThis is required because ASSETDISTTYPE has been set to 1 (prepared asset distribution provided)\n"  and exit(1) unless (defined $self->config_entry("AGENT_INFO"));

		      $agentinfo = $self->config_entry("AGENT_INFO")->transform($dbh,1);	
		  }
		  print STDERR "\nError: Query for AGENT_INFO returned an empty result.\n" and exit(1) if ($agentinfo->is_empty);
		  $agentsinclu = $agentinfo->regrouped_copy([2,3,1]);
	}
	else {
	    print STDERR "\nError in cfg: Invalid value [". $self->config_entry("ASSETDISTTYPE") ."] for ASSETDISTTYPE" and exit(1);
	}	
	for my $f ($self->files) {
		$f->prepare_shared($sexagedef, $householdhash, $assethash, $agentsinclu);
	}	
	$self->set_sexagedef($sexagedef);
	$self->set_householdhash($householdhash);
	$self->set_assethash($assethash);
	$self->set_agentinfo($agentinfo);
}

sub prepare_special_files {
    my $self = shift;
    my $dbh = $self->get_dbh;
    my $fixcosthash = $self->config_entry("FIXCOSTS")->transform($dbh,3) if ($self->config_entry("FIXCOSTS"));
    for my $f ($self->files) {
	  $f->prepare_special($self->sexagedef, $self->householdhash, $fixcosthash, $self->assethash, $self->agentinfo, $self);
    }
}

sub sexagedef {
    my $self = shift;
    return $self->{sexagedef};
}
sub set_sexagedef {
    my $self = shift;
    $self->{sexagedef}  = shift;
}
sub householdhash {
    my $self = shift;
    return $self->{housholdhash};
}
sub set_householdhash {
    my $self = shift;
    $self->{housholdhash}  = shift;
}
sub assethash {
    my $self = shift;
    return $self->{assethash};
}
sub set_assethash {
    my $self = shift;
    $self->{assethash}  = shift;
}

1;
