package MpmasCropSimilarity;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use List::Util qw[min max];
use CtUtilities qw(queryhash extqueryhash);
use MpmasInputfile;
use MpmasInputBlock;
our @ISA = qw(MpmasInputfile);

sub filename {
	return "CropSimilarity";
}

sub required {
	my $self = shift;
	return ($self->config_entry("CROP_MODEL") > 0);
}
sub prepare_shared {
    my ($self) = @_;

    my @sorted_cropping_activities;
    if    ($self->config_entry("CROP_MODEL") == 3) {
	@sorted_cropping_activities = @{ $self->get_global("Sorted_Xn_Activity_Objects") };
    }
    elsif ($self->config_entry("CROP_MODEL") == 2) {
	@sorted_cropping_activities = @{ $self->get_global("Sorted_CropWat_Activities")  };
    }
    elsif ($self->config_entry("CROP_MODEL") == 1) {
	print STDERR "Error: You set CROP_MODEL == 1 (TSPC), but mpmasql is not able to create TSPC files yet\n";
	exit(1);
    }
    else {
	print STDERR "Error: Invalid value for CROP_MODEL\n";
	exit(1);
    }
    $self->{sorted_cropping_activities} = \@sorted_cropping_activities;
    $self->{similarity_levels} = 0;
    $self->{similarity_codes} = {} ;
    $self->{curmaxcode} = [];
    for my $ca (@sorted_cropping_activities) {
	for my $f ($ca->subelement_fields("yield_expectation") ) {
	    next unless ($f =~ /similarity_(\d+)/);
	    my $l = $1;
	    if ($l > $self->{similarity_levels} ) {
		  $self->{similarity_levels} = $l;
		  $self->{curmaxcode}->[$l] = 0;
	    }
	    unless ( exists $self->{similarity_codes}->{$l}->{ $ca->subelement_field("yield_expectation", $f) }  ) {
		  $self->{similarity_codes}->{$l}->{ $ca->subelement_field("yield_expectation", $f) } = ++ $self->{curmaxcode}->[$l];
	    }
	}
	$ca->add_subelement_field("yield_expectation", "similarity_0", $ca->id);
    }
}
sub prepare_special {
    my ($self) = @_;
    my $block = MpmasInputBlock->new(3,10);
    $block->add_comment_row("THE CROP SIMILARITY FILE");
    $block->add_empty_rows(1);
    $block->add_comment_row(1, "CROP SIMILARITIES");
    $block->add_comment_row(1.1, "BASIC PARAMETERS");
    $block->add_comment_row("","", "Variable","Value");
    $block->add_row("","","Levels of similarity",$self->{similarity_levels});
    $block->add_empty_rows(1);
    

    if ($self->{similarity_levels} > 0) {
	$block->add_comment_row(1.2,"MATRIX OF CROP SIMILARITIES");
	$block->add_comment_row("","", "", "Levels");
	$block->add_comment_row("","Activity code", "Activity description", 1..$self->{similarity_levels});
	for my $ca (@{ $self->{sorted_cropping_activities} }) {
	    my @row;
	    for my $l (1..$self->{similarity_levels}) {
		  my $p = 0;
		  my $g;
		  while ($p <= $l) {
		      my $l2 = $l-$p;
		      my $f = "similarity_" .$l2 ;
		      $g =  $ca->subelement_field("yield_expectation", $f);
		      last if (defined $g);
		      $p++;
		  }
		  unless ( exists $self->{similarity_codes}->{$l}->{ $g }  ) {
			  $self->{similarity_codes}->{$l}->{ $g } = ++ $self->{curmaxcode}->[$l];
		  }
		  push @row, $self->{similarity_codes}->{$l}->{ $g };
	    }
	    $block->add_row("", $ca->id, $ca->attr("name"), @row);
	}
	
    }
    $self->add_block($block);
}