package MpmasDemography;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use CtUtilities qw(queryhash extqueryhash);
use MpmasInputfile;
use MpmasInputBlock;
our @ISA = qw(MpmasInputfile);

sub filename {
	my $self = shift;
	return "Pop".$self->file_suffix;
}

sub required {
	my $self = shift;
	return 1;
}
sub prepare_shared {
	

}
sub prepare_special {
	my $self = shift;
	#my $demohash = shift;
	#my $LBGcount = shift;
	#my $refHHNutrients = shift;
	#my @HHNutrients = @$refHHNutrients;
	my $curPop = $self->argument;
	my $demography = $self->config_entry("CONVERTED_DEMOTAB");
	# Columns of HOUSEHOLD DYNAMICS (== CONVERTED_DEMOTAB):
	# Both demography models:
	# pop career upperyear labgrp labprov mortality fertility 
	# Demography model 1 : p_marriage p_leave p_retire priority_hh mincons
	# Advanced consumption model: energy, other nutrients, ...

	my $household_switching = $self->config_entry("CONVERTED_HOUSEHOLD_SWITCHING");
	# pop orig_career cause new_career probability
	my %desc_switching = (
	 0 => "Career of descendants",
	 1 => "New career after marriage",
	 2 => "Career of partner",
	 3 => "New career after becoming household head",
	 4 => "New career after partner became household head",
	 5 => "New career after retiring" 
	);
	my $succession_control = $self->config_entry("CONVERTED_SUCCESSION_CONTROL");
	# pop control_variable value
	my $invhorizon = $self->config_entry("CONVERTED_INVESTMENT_HORIZON");

	my $invhorizonconstraints = $self->config_entry("CONVERTED_INVESTMENT_HORIZON_CONSTRAINTS");
	# [popID, lowerbound], constraint identifier, rhsvalue
	my $extraIncBalance = $self->config_entry("CONVERTED_SUC_INCOME_BALANCE");


	my $LBGcount = $self->config_entry("LBGcount");
	my @HHNutrients = 0;
	if  ($self->config_entry("CONSTYPE") == 1) {
		@HHNutrients = $self->get_cfg_list("HHNUTRIENTS");
	}
	

	
	my $demotabs = MpmasInputBlock->new(2,30);	
	my $CareerCount = 0;			
	
	my @sorted_careers = sort { $self->config_entry("CAREERS")->{$a}<=>$self->config_entry("CAREERS")->{$b} } $demography->query([$curPop],0);
	for my $curCareer (@sorted_careers) {
		$CareerCount++;
		my @ctabs;
		my @ylist = sort { $a<=>$b } $demography->query([$curPop, $curCareer],0);
		#for my $i (0..$#ylist) {
		#	$ylist[$i] =~ s/^x//;
		#}
		#@ylist = sort{$a  <=> $b } @ylist;
		my $L = $#ylist;
		my $lasty = $ylist[$L];
		
		my $curE = 0;
		$demotabs->add_row("$curCareer");
		$demotabs->add_row("","Career ID",$self->config_entry("CAREERS")->{$curCareer});
		$demotabs->add_row("","Number of different ages", $lasty+1);
		my @temp;
		$temp[1] = "Age";
		for my $y (0..$lasty){
			$temp[2+$y] = $y;
		}
		$demotabs->add_comment_row(@temp);
		$ctabs[0][1] = "Labour span";
		$ctabs[1][1] = "Labour provision per year";
		$ctabs[2][1] = "Mortality";
		$ctabs[3][1] = "Fertility";
		my $firstnutrient = 4;
		if ($self->config_entry("DEMOGRAPHY_MODEL") == 1 ) {
			$ctabs[4][1] = "Prob. of marriage / finding a partner";
			$ctabs[5][1] = "Prob. of leaving household";
			$ctabs[6][1] = "Prob. of (attempting to) retire";
			$ctabs[7][1] = "Priority in succession if household head";
			$ctabs[8][1] = "Minimum consumption";
			$firstnutrient = 9; 
		}
		if ($self->config_entry("CONSTYPE") == 1 ) {
			for my $nut (0.. $#HHNutrients) {
				  $ctabs[$firstnutrient + $nut][1] = $HHNutrients[$nut]." requirement";
			}	
		}
		for my $y (0..$lasty){
			my $curY = $ylist[$curE];
			my $h = $demography->query([$curPop,$curCareer, $curY],4);
			my $lg = $self->config_entry("LABGRPS")->{$h}->{lid};
			print STDERR "Error in demography table: Career $curCareer provides labour of labour group $h,\n but no endowment constraint for such a group has been defined!\nCheck the labgrp fields of your labour endowment constraints.\n" and exit(1) unless (defined $lg);
			$ctabs[0]->[$y+2] = $lg; 
			for my $f (1.. ($firstnutrient - 1) ) {
				$ctabs[$f]->[$y+2] = $demography->query([$curPop,$curCareer, $curY],4 + $f) || 0;
			}
			if ($self->config_entry("CONSTYPE") == 1 ) {
				for my $f ($firstnutrient..$firstnutrient + $#HHNutrients) {
					my $nr = $demography->query([$curPop,$curCareer, $curY],4 + $f); 
					$nr = 0 unless (defined $nr);
					$ctabs[$f]->[$y+2] = $nr;
				}
			}
			if (defined $demography->query([$curPop,$curCareer, $y],4)){
				$curE += 1;
			}
		}
		
		push @ctabs, [()];
		push @ctabs, [()];
		$demotabs->add_table(\@ctabs);

	        if ($self->config_entry("DEMOGRAPHY_MODEL") == 1 ) {
			    $demotabs->add_comment_row("","Spawning and switching probabilities", @sorted_careers);

		#Switching/descendant probabilities
			    for my $c (sort {$a <=> $b} keys %desc_switching) {
				my %temp;
				my @row = ("", $desc_switching{$c});
				my $rsum = 0;
				for my $s (@sorted_careers) {
					my $p = $household_switching->query([$curPop, $curCareer, $c, $s], 5);
					$p = 0 unless (defined $p);
					$temp{$s} = $p;
					$rsum += $p unless ($s eq $curCareer); #Remaining in career is calculated as remainder of switching probabilities, any given entry is ignored!!!
				}
				print STDERR "Error in HOUSEHOLD_SWITCHING [MPMAS TABLES]: Sum of probabilities for " . $desc_switching{$c} . " for career $curCareer is larger than 1\n" and exit(1) if ($rsum > 1);
				print STDERR "Error in HOUSEHOLD_SWITCHING [MPMAS TABLES]: Sum of probabilities for " . $desc_switching{$c} . " for career $curCareer is smaller than 0\n" and exit(1) if ($rsum < 0);
				$temp{$curCareer} = 1 - $rsum;#Remaining in career is calculated as remainder of switching probabilities to other careers
				for my $s (@sorted_careers) {
				    push @row, $temp{$s};
				}
				$demotabs->add_row(@row);
			    }
		}
		$demotabs->add_empty_rows(3);
	}
	$self->add_block($demotabs);
	
	my $demohead = MpmasInputBlock->new(2,10);
	$demohead->add_row("THE DEMOGRAPHY FILE");
	$demohead->add_row("1","HOUSEHOLD MEMBER COURSE OF LIFE");
	$demohead->add_row("1.1","Number of different member categories", $CareerCount);
	$demohead->add_row("1.2","Number of different labor groups", $LBGcount );
	$demohead->add_row("1.3","School Age", $self->config_entry("SCHOOLAGE") || 0 );
	$demohead->add_row("1.4","Adult Age", $self->config_entry("ADULTAGE") || 0);
	$demohead->add_empty_rows(2);
	$self->add_block($demohead);
	

	if ($self->config_entry("DEMOGRAPHY_MODEL") == 1 ) {
		#hash containing settings for succesion control
		my %suc;
		# first initialize hash with general settings from MPMAS PARAMETERS section
		for my $c (qw(SUCCESSION_AD_MINCONS_FACTOR SUCCESSION_AD_MINIMUM_PROFIT SUCCESSION_AD_DECISION_ACTIVITY SUCCESSION_VR_MINCONS_FACTOR SUCCESSION_VR_MINIMUM_PROFIT SUCCESSION_VR_DECISION_ACTIVITY SUCCESSION_VR_MINIMUM_RENTAL SUCCESSION_OR_MINCONS_FACTOR SUCCESSION_OR_MINIMUM_PROFIT SUCCESSION_OR_DECISION_ACTIVITY INVESTMENT_HORIZON_MODEL INVESTMENT_HORIZON_PTHRESHOLD INVESTMENT_HORIZON_WHO INVESTMENT_HORIZON_AFFECT_LIFETIME INCOME_BALANCE_ROW RHS_MIN_CONS RHS_FIX_COST)) {
		  $suc{$c} = $self->config_entry($c);
		}
		# then overwrite with population specific settings if they exist
		if (defined $succession_control) {
		    for my $c ($succession_control->query([$curPop],0)) {
			  $suc{$c} = $succession_control->query([$curPop, $c]) if (defined $succession_control->query([$curPop, $c]));  
		    }
		}
		if (defined $invhorizon) {
		    for my $c ($invhorizon->query([$curPop],0)) {
			  $suc{$c} = $invhorizon->query([$curPop, $c]) if (defined $invhorizon->query([$curPop, $c]));  
		    }
		}
		if (defined $extraIncBalance) {
		    for my $c ($extraIncBalance->query([$curPop],0)) {
			  $suc{$c} = $extraIncBalance->query([$curPop, $c]) if (defined $extraIncBalance->query([$curPop, $c]));  
		    }
		}

		my $succession = MpmasInputBlock->new(2,50);
		$succession->add_empty_rows(2);
		$succession->add_row("2", "SUCCESSION OF HOUSEHOLD HEAD");
		$succession->add_empty_rows(1);		
		$succession->add_row("2.1","Succession if household head dies");
		$succession->add_empty_rows(1);		
		$succession->add_row("","Successors are identified and ranked according to their priority of becoming household head");
		$succession->add_row("","The  person with highest priority (i.e. lowest priority value) is the potential successor");
		$succession->add_row("","This person becomes the new household head if");
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- the farm profit was higher than the combined minimum consumption of the household multiplied by the");
		$succession->add_row("2.1.1","rel. factor for minimum consumption for succession at death", $suc{SUCCESSION_AD_MINCONS_FACTOR});
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- AND the farm income is higher than");
		$succession->add_row("2.1.2","the minimum profit for succession at death", $suc{SUCCESSION_AD_MINIMUM_PROFIT});
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- AND there was a positive value in the LP solution for");
	  
		my $da = -1;

		if (defined $suc{SUCCESSION_AD_DECISION_ACTIVITY} and $suc{SUCCESSION_AD_DECISION_ACTIVITY} ne "-1" and $suc{SUCCESSION_AD_DECISION_ACTIVITY} ne "0" ) {
		    my $dao = $self->activity($suc{SUCCESSION_AD_DECISION_ACTIVITY});
		    if (defined $dao) {
			$da = $dao->lp_column ;
		    }
		    else {
			print STDERR "Error for SUCCESSION_AD_DECISION_ACTIVITY: Activity ". $suc{SUCCESSION_AD_DECISION_ACTIVITY} ."  has not been defined in the model\n" and exit(1);
		    }
		}
		$succession->add_row("2.1.3","the activity indicating potential succession at death", $da);
		$succession->add_row("","Otherwise the farm is shut down.");
		$succession->add_empty_rows(1);		
		$succession->add_row("2.2","Succession if household head attempts to retire");
		$succession->add_empty_rows(1);		
		$succession->add_row("","Successors are identified and ranked according to their priority of becoming household head");
		$succession->add_row("","The  person with highest priority (i.e. lowest priority value) is the potential successor");
		$succession->add_row("","This person becomes the new household head if");
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- the farm profit was higher than the combined minimum consumption of the household multiplied by the");
		$succession->add_row("2.2.1","rel. factor for minimum consumption for succession at voluntary retirement", $suc{SUCCESSION_VR_MINCONS_FACTOR});
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- AND the farm income is higher than");
		$succession->add_row("2.2.2","the minimum profit for succession at voluntary retirement", $suc{SUCCESSION_VR_MINIMUM_PROFIT});
		$succession->add_empty_rows(1);		
		$da = -1;
		if (defined $suc{SUCCESSION_VR_DECISION_ACTIVITY} and $suc{SUCCESSION_VR_DECISION_ACTIVITY} ne "-1" and $suc{SUCCESSION_VR_DECISION_ACTIVITY} ne "0" ) {
		    my $dao = $self->activity($suc{SUCCESSION_VR_DECISION_ACTIVITY});
		    if (defined $dao) {
			$da = $dao->lp_column ;
		    }
		    else {
			print STDERR "Error for SUCCESSION_VR_DECISION_ACTIVITY: Activity ". $suc{SUCCESSION_VR_DECISION_ACTIVITY} ."  has not been defined in the model\n" and exit(1);
		    }
		}
		$succession->add_row("","-- AND there was a positive value in the LP solution for");
		$succession->add_row("2.2.3","the activity indicating potential succession at voluntary retirement", $da);
		$succession->add_empty_rows(1);		
		$succession->add_row("","If no potential successor is available, the farmer checks the expected payments for hiring out his land and selling his assets.");
		$succession->add_row("","He will rent out all his land and shut down the farm,"); 
		$succession->add_row("","if the sum of rental payments and resale revenue annualized of the rest of his lifetime is higher than");
		$succession->add_row("2.2.4","the minimum rental earnings for voluntary shutdown", $suc{SUCCESSION_VR_MINIMUM_RENTAL});
		$succession->add_empty_rows(1);		
		$succession->add_row("","Otherwise he will not retire.");
		$succession->add_empty_rows(2);		
		$succession->add_row("2.3","Succession if household head is required to retire");
		$succession->add_empty_rows(1);		
		$succession->add_row("","Successors are identified and ranked according to their priority of becoming household head");
		$succession->add_row("","The  person with highest priority (i.e. lowest priority value) is the potential successor");
		$succession->add_row("","This person becomes the new household head if");
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- the farm profit was higher than the combined minimum consumption of the household multiplied by the");
		$succession->add_row("2.2.1","rel. factor for minimum consumption for succession at voluntary retirement", $suc{SUCCESSION_OR_MINCONS_FACTOR});
		$succession->add_empty_rows(1);		
		$succession->add_row("","-- AND the farm income is higher than");
		$succession->add_row("2.2.2","the minimum profit for succession at voluntary retirement", $suc{SUCCESSION_OR_MINIMUM_PROFIT});
		$succession->add_empty_rows(1);		
		$da = -1;
		if (defined $suc{SUCCESSION_OR_DECISION_ACTIVITY} and $suc{SUCCESSION_OR_DECISION_ACTIVITY} ne "-1" and $suc{SUCCESSION_OR_DECISION_ACTIVITY} ne "0" ) {
		    my $dao = $self->activity($suc{SUCCESSION_OR_DECISION_ACTIVITY});
		    if (defined $dao) {
			$da = $dao->lp_column ;
		    }
		    else {
			print STDERR "Error for SUCCESSION_OR_DECISION_ACTIVITY: Activity ". $suc{SUCCESSION_OR_DECISION_ACTIVITY} ."  has not been defined in the model\n" and exit(1);
		    }
		}
		$succession->add_row("","-- AND there was a positive value in the LP solution for");
		$succession->add_row("2.2.3","the activity indicating potential succession at voluntary retirement", $da);
			
		$succession->add_empty_rows(1);
		$succession->add_row("","Otherwise the farm is shut down.");
		$succession->add_empty_rows(2);
		$self->add_block($succession);

		
		my $invbeh = MpmasInputBlock->new(2,70);
		$invbeh->add_empty_rows(2);
		$invbeh->add_row("3", "INVESTMENT HORIZON");
		$invbeh->add_empty_rows(1);		
		$invbeh->add_row("3.1", "Influence of household head age on investment behavior", $suc{INVESTMENT_HORIZON_MODEL});
		$invbeh->add_empty_rows(1);		
		if ($suc{INVESTMENT_HORIZON_MODEL} == 1) {
			$invbeh->add_comment_row("3.2","Calculation of remaining lifetime of household");
			$invbeh->add_row("3.2.1", "Retirement probability indicating possible end of household lifetime", $suc{INVESTMENT_HORIZON_PTHRESHOLD});
			$invbeh->add_row("3.2.2", "Whom to consider?", $suc{INVESTMENT_HORIZON_WHO});
			$invbeh->add_empty_rows(1);
			$invbeh->add_row("3.3", "Constraints reflecting remaining lifetime" );
			$invbeh->add_empty_rows(1);		

			my @cons = ();
			@cons =  $invhorizonconstraints->query([$curPop],0) if (defined $invhorizonconstraints);
			$invbeh->add_row("", "Number of constraints", scalar (@cons) );
			if (scalar @cons > 0) {
			    $invbeh->add_comment_row("", "Constraint identifier", "LP Row", "Min. remaining lifetime", "RHS value if HH falls into this range" );
			    for my $lb (sort {$a <=> $b } @cons) {
			      my $cID = $invhorizonconstraints->query([$curPop, $lb],3);
			      my $k = $self->constraint($cID);		  
			      print STDERR "Error in INVESTMENT_HORIZON_CONSTRAINTS [MPMAS TABLES]: Constraint '$cID' has not been defined in the .cll file\n" and exit(1) unless (defined $k);
			      $invbeh->add_row("", $cID, $k->lp_row, $lb, $invhorizonconstraints->query([$curPop, $lb],4));
			    }
			}
			$invbeh->add_empty_rows(1);
			$invbeh->add_row("3.4", "Consider household lifetime max. lifespan in investment decision", $suc{INVESTMENT_HORIZON_AFFECT_LIFETIME});
			$invbeh->add_comment_row("","Note: Does not work yet for perennial crops and livestock.");
		}
		$self->add_block($invbeh);

	    	my $extraIncRow = MpmasInputBlock->new(2,80);
		$extraIncRow->add_empty_rows(2);
		$extraIncRow->add_row("4", "INCOME CHECK IN LP");
		$extraIncRow->add_empty_rows(1);	
		$extraIncRow->add_row("", "Control variables for the income row to check minimum consumption within the LP");
		$extraIncRow->add_row("", "(Note: A minus one indicates, the corresponding value(s) is (are) not entered into the LP)");
		$extraIncRow->add_empty_rows(1);

		$da = -1;
		if (defined $suc{INCOME_BALANCE_ROW} and $suc{INCOME_BALANCE_ROW} ne "-1" and $suc{INCOME_BALANCE_ROW} ne "0" ) {
		    my $dao = $self->constraint($suc{INCOME_BALANCE_ROW});
		    if (defined $dao) {
			$da = $dao->lp_row ;
		    }
		    else {
			print STDERR "Error for INCOME_BALANCE_ROW: Constraint ". $suc{INCOME_BALANCE_ROW} ."  has not been defined in the model\n" and exit(1);
		    }
		}
		$extraIncRow->add_row("", "LP index of income balance row", $da);

		$da = -1;
		if (defined $suc{RHS_FIX_COST} and $suc{RHS_FIX_COST} ne "-1" and $suc{RHS_FIX_COST} ne "0" ) {
		    my $dao = $self->constraint($suc{RHS_FIX_COST});
		    if (defined $dao) {
			$da = $dao->lp_row ;
		    }
		    else {
			print STDERR "Error for RHS_FIX_COST: Constraint ". $suc{RHS_FIX_COST} ."  has not been defined in the model\n" and exit(1);
		    }
		}
		$extraIncRow->add_row("", "LP index of RHS for fix costs",$da);

		$da = -1;
		if (defined $suc{RHS_MIN_CONS} and $suc{RHS_MIN_CONS} ne "-1" and $suc{RHS_MIN_CONS} ne "0" ) {
		    my $dao = $self->constraint($suc{RHS_MIN_CONS});
		    if (defined $dao) {
			$da = $dao->lp_row ;
		    }
		    else {
			print STDERR "Error for RHS_MIN_CONS: Constraint ". $suc{RHS_MIN_CONS} ."  has not been defined in the model\n" and exit(1);
		    }
		}
		$extraIncRow->add_row("", "LP index of RHS for minimum consumption",$da);


		$self->add_block($extraIncRow);
	}
}
