package MpmasInputfile;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use SimpleObjectRegistry;
use MpmasInputBlock;
use SimpleObject;
use Carp;
our @ISA = qw(SimpleObject);

#constants
use constant {
  DEFBASEMILP => 10
};

#Global references for all input files - initialized by class function set_global_references
my $mname;
my $outdir;
my $outdir_cm;
my $inihash;
my $configs;
my $setup;
my $ActivityRegistry;
my $ConstraintRegistry;
my $NetworkObjectRegistry;
my $PerennialObjectRegistry;
my $LivestockObjectRegistry;
my $TPPenaltyObjectRegistry;
my $SOSRegistry;
my $SortedActivityRef;
my $SortedQuadActRef;
my $SortedConstraintRef;
my $SortedObjectRef;
my $MILPinfo;
my $other_class_globals;
my $dbh;
my $backwards;
my $specobjname;
my $specobjindex;
my $specobjtype;
my $tableau_compact;
my %group_scripts;
my $milplist;
my $agentinfo;

#Methods for all files, do not overload with specific class methods
sub set_global_references {
	shift; #ignore invoker
	$mname = shift;
	$outdir = shift;
	$outdir_cm = shift;
	$inihash = shift;
	$configs = shift;
	$setup = shift;
	$ActivityRegistry = shift;
	$ConstraintRegistry = shift;
	$NetworkObjectRegistry = shift;
	$PerennialObjectRegistry = shift;
	$LivestockObjectRegistry = shift;
	$TPPenaltyObjectRegistry = shift;
	$SOSRegistry = shift;
	$dbh = shift;
	$backwards = shift;
	$tableau_compact = shift;
#	$SortedActivityRef = shift;
#	$SortedConstraintRef = shift;
#	$MILPinfo = shift;
}
sub set_global {
	$other_class_globals->{$_[1]} = $_[2];
}
sub get_global {

	return $other_class_globals->{$_[1]};
}
sub get_dbh {
	return $dbh;
}
sub set_milpinfo {
	$MILPinfo = $_[1];
}
sub set_milpinfo_entry {
	$MILPinfo->{$_[1]} = $_[2];
}
sub set_specobj_info {
	$specobjname = $_[1];
	$specobjindex = $_[2];
	$specobjtype = $_[3];
}
sub get_specobj_names {
	return $specobjname;
}
sub get_specobj_indices {
	return $specobjindex;
}
sub get_specobj_types {
	return $specobjtype;
}
sub set_sorted_constraints {
	$SortedConstraintRef = $_[1];
}
sub set_sorted_activities {
	$SortedActivityRef = $_[1];
}
sub set_sorted_quadratic_activities {
	$SortedQuadActRef = $_[1];
}
sub set_sorted_networkobjects {
	shift;
	$SortedObjectRef = shift;
}
sub set_constraint_registry {
	$ConstraintRegistry = $_[1];
}
sub set_activity_registry {
	$ActivityRegistry = $_[1];
}
sub set_networkobject_registry {
	$NetworkObjectRegistry = $_[1];
}
sub set_SOS_registry {
	$SOSRegistry = $_[1];
}
sub get_constraint_registry {
	return $ConstraintRegistry;
}
sub get_activity_registry {
	return $ActivityRegistry;
}
sub get_networkobject_registry {
	return $NetworkObjectRegistry;
}
sub get_penalty_registry {
	return $TPPenaltyObjectRegistry;
}
sub get_SOS_registry {
	return $SOSRegistry;
}
sub config_entry {
	return $configs->{$_[1]};
}
sub config_entry_or_croak {
	croak "Error: $_[1] not defined\n" unless (defined $configs->{$_[1]});
	return $configs->{$_[1]};
}
sub config_entry_or_die {
	print STDERR "Error: $_[1] not defined\n" and exit(1) unless (defined $configs->{$_[1]});
	return $configs->{$_[1]};
}
sub set_config_entry {
	$configs->{$_[1]} = $_[2];
}

sub setup_entry {
	return $setup->{$_[1]};
}

sub ini_entry {
	return $inihash->{$_[1]};
}
sub backwards_compatibility {
	return (defined $backwards and $_[1] >= $backwards );
}
sub tableau_compact {
    return $tableau_compact;
}
sub set_milplist {
  $milplist = $_[1];
}
sub milplist {
  return $milplist;
}
sub agentinfo {
    return $agentinfo;
}
sub set_agentinfo {
    $agentinfo  = $_[1];
}
sub activity {
	return $ActivityRegistry->get_object_by_id($_[1]);
}
sub lp_column_of_activity {
	my $i = $ActivityRegistry->get_object_by_id($_[1]);
	croak "Activity $_[1] not defined!" unless (defined $i);
	return $i->lp_column;
}
sub add_activities {
	shift;
	$ActivityRegistry->add_objects(@_);
}
sub number_of_activities {
	return $ActivityRegistry->counter;
}
sub activities {
	return $ActivityRegistry->object_list;
}
sub sorted_activities {
	return $SortedActivityRef->[$_[1]] if (defined $_[1]);
	return @{$SortedActivityRef};
}
sub sorted_quadratic_activities {
	return $SortedQuadActRef->[$_[1]] if (defined $_[1]);
	return @{$SortedQuadActRef};
}
sub sorted_constraint_attr {
	my ($self,$attr,$thismilp) = @_;
	my @temp;
	for my $i ($self->sorted_constraints) {
		my $h =	$self->constraint($i);
		push @temp, $h->attr($attr, $thismilp);
	}
	return @temp;
}
sub sorted_activity_attr {
	my ($self,$attr,$thismilp) = @_;
	my @temp;
	for my $i ($self->sorted_activities) {
		my $h =	$self->activity($i);
		push @temp, $h->attr($attr, $thismilp);
	}
	return @temp;
}
sub sorted_activity_vtype {
	my ($self,$thismilp) = @_;
	my @temp;
	for my $i ($self->sorted_activities) {
		my $h =	$self->activity($i);
		push @temp, $h->attr("quadratic", $thismilp) * 2 + $h->attr("integ", $thismilp);
	}
	return @temp;
}
sub constraint {
	return $ConstraintRegistry->get_object_by_id($_[1]);
}
sub lp_row_of_constraint {
	my $i = $ConstraintRegistry->get_object_by_id($_[1]);
	croak "Constraint $_[1] not defined!" unless (defined $i);
	return $i->lp_row;
}
sub add_constraints {
	shift;
	$ConstraintRegistry->add_objects(@_);
}
sub number_of_constraints {
	return $ConstraintRegistry->counter;
}
sub constraints {
	return $ConstraintRegistry->object_list;
}
sub sorted_constraints {
	return $SortedConstraintRef->[$_[1]] if (defined $_[1]);
	return @{$SortedConstraintRef};
}
sub constraint_at_row {
	my $h = $_[0]->sorted_constraints($_[1]);
	return $_[0]->constraint($h);
}
sub activity_at_col {
	my $h = $_[0]->sorted_activities($_[1]);
	return $_[0]->activity($h);
}

sub networkobject {
	return $NetworkObjectRegistry->get_object_by_id($_[1]);
}
sub networkobjects {
	return $NetworkObjectRegistry->object_list;
}
sub number_of_networkobjects {
	return $NetworkObjectRegistry->counter;
}
sub sorted_networkobjects {
	return $SortedObjectRef->[$_[1]] if (defined $_[1]);
	return @{$SortedObjectRef};
}
sub perennialobject {
	return $PerennialObjectRegistry->get_object_by_id($_[1]);
}
sub perennialobjects {
	return $PerennialObjectRegistry->object_list;
}
sub number_of_perennialobjects {
	return $PerennialObjectRegistry->counter;
}
sub livestockobject {
	return $LivestockObjectRegistry->get_object_by_id($_[1]);
}
sub livestockobjects {
	return $LivestockObjectRegistry->object_list;
}
sub number_of_livestockobjects {
	return $LivestockObjectRegistry->counter;
}
sub SOS {
	return $SOSRegistry->get_object_by_id($_[1]);
}
sub SOSs {
	return $SOSRegistry->object_list;
}
sub number_of_SOS {
	return $SOSRegistry->counter;
}

sub milpinfo {
	return $MILPinfo->{$_[1]};
}

sub mname {
	return $mname;
}


sub outdir {
	return $outdir;
}

sub outdir_cm {
	return $outdir_cm;
}

sub find_special_constraint {
	my $self = shift;
	my $type = shift;
	my $essential = shift || 0;
	my $return_object_too = shift || 0;
	my $civ;
	my $cname;
	for my $i ($self->sorted_constraints) {
		my $t = $self->constraint($i);
		if (lc($t->attr("type")) eq lc($type)){
			$civ = $t->lp_row ;
			$cname = $t->id;
			last;
		}
	}
	print STDERR "No constraint of type '$type' found\n" and exit (1) if ($essential and not defined $civ);
	return $return_object_too ? ($civ, $cname) : $civ;
}

sub find_special_activity {
	my $self = shift;
	my $type = shift;
	my $essential = shift || 0;
	my $civ;
	for my $i ($self->sorted_activities) {
		my $t = $self->activity($i);
		if (lc($t->attr("type")) eq lc($type)){
			$civ = $t->lp_column ;
			last;
		}
	}
	print STDERR "No activity of type '$type' found\n" and exit (1) if ($essential and not defined $civ);
	return $civ;
}
sub find_special_constraints {
	my $self = shift;
	my $type = shift;
	my $essential = shift || 0;
	my $civ;
	my $ct = 0;
	my $cname;
	for my $i ($self->sorted_constraints) {
		my $t = $self->constraint($i);
		if (lc($t->attr("type")) eq lc($type)){
			$civ = $t->lp_row unless (defined $civ);
			$cname = $t unless (defined $cname);
			$ct += 1;
		}
	}
	print STDERR "No constraint of type '$type' found\n" and exit (1) if ($essential and not defined $civ);
	return $ct, $civ;
}
sub find_special_activities {
	my $self = shift;
	my $type = shift;
	my $essential = shift || 0;
	my $civ;
	my $ct = 0;
	for my $i ($self->sorted_activities) {
		my $t = $self->activity($i);
		if (lc($t->attr("type")) eq lc($type)){
			$civ = $t->lp_column unless (defined $civ);
			$ct += 1;
		}
	}
	print STDERR "No activity of type '$type' found\n" and exit (1) if ($essential and not defined $civ);
	return $ct, $civ;
}
sub get_cfg_list {
    my $s = $_[0]->config_entry($_[1]);
    return @$s if (ref $s eq "ARRAY");
    my $list; ;
    eval {
	$list = MpmasQLFunctionLanguage->translate_and_interprete ($s, {}); 
    };
    croak "Error: when interpreting cfg entry $_[1]\n$@\n" if $@;
    croak "Interpretation of cfg entry $_[1] returned empty list!\n" unless (@$list) ;
    return @$list;
}




sub create {}; # nullify SimpleObject constructor

sub blank_file {
	$_[0]->{InputBlocks}->clear;	

}
## subs to be adapted only if necessary
sub extension {
	return ".dat";
}

sub extension_cm {
	return ".txt";
}

sub leftmargin {
	return "\t\t";
}

sub priority{
	return $_[0]->{priority};
}
sub file_suffix {
	return $_[0]->{file_suffix};
}
sub argument {
	return $_[0]->{argument};
}



sub new {
	my ($class,$priority, $argument , $file_suffix) = @_;
	my $self = {};
	$self->{InputBlocks} = SimpleObjectRegistry->create();
	$self->{priority} = $priority;
	my $id = $class;
	if (defined $argument) {
		$id .= "_".$argument;
		$self->{argument} = "$argument";
	}
	$self->{id} = $id;
	$self->{file_suffix} = "$file_suffix" if (defined $file_suffix);
	
	bless $self, $class;
	return $self;
}

sub add_block {
	$_[0]->{InputBlocks}->add_objects(@_[1..$#_]);
}

sub write {
	my ($self,$mode,$cur,$valuetab) = @_;
	#print "in Inputfile: $self,$mode,$cur, $valuehash\n";
	my $out;		
	my $dir = $self->outdir();
	

	my $ext = $self->extension();
	my $mn = $self->mname();
	
	if ($mode == 1) {
		$dir = $self->outdir_cm();
		$ext = $self->extension_cm();
	}
	my $fn = $self->full_filename($dir, $mn, $self->filename(), $ext);
	open($out, ">", $fn ) or croak "Cannot write to $fn: $!\n";
	print "\e[K \r";
	print "Writing $fn...\r";
	my @SortedBlocks = sort {$a->priority <=> $b->priority } $self->{InputBlocks}->object_list;

	for my $j (@SortedBlocks ) {
		$j->write($out, $mode, $self->leftmargin, $cur,$valuetab);
	}

	#print "..done!\n";
	close ($out);
	
	
	return $fn;
}

sub full_filename {
    my ($self, $dir, $mn, $f, $ext) = @_;
    my $fn = $dir.$mn."_".$f .$ext;
    return $fn;
}
## Implement for specific file

sub filename {
	return "Testfile";
}
sub required {
	return 1;
}

sub prepare_shared {
	#Implement if new file does preparations which are relevant for other files as well
	return 1;
}

sub prepare_special {
	#Implement for new file
	return 1;
}

1;
