package MpmasNetworkSet;

use strict;
use warnings;
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use SimpleObjectRegistry;
use MpmasInputBlock;
use SimpleObject;
use IndexedTable;
use MpmasNetwork;
use CtUtilities qw(queryhash strip_sql);
our @ISA = qw(MpmasInputfileSet);

sub class_of_files{
	return "MpmasNetwork";
 }

sub determine_files {
	my $self = shift;
		
	for my $n (0..$self->config_entry("NETWORKS")-1 ) {
		my $fname = sprintf("%01d", $n);
		$self->add_file($n,$fname);
	}

}
sub define_specobjs {
      my $self = shift;
      my %specobjname = (
		femhead => "Female Head of Household",
		innovativeness => "Innovativeness",
		expectations => "Form of expectations",
		liquidity => "Liquidity",
		leverage =>	"Leverage",
		zero_savings_prob	=>	"Random number for logit zero savings model",
		timepref => 	"Monthly time preference of agent",
		which_market => "Market the agent ahs access to"
		) ;

	my %specobjtype = (
		femhead => -1,
		innovativeness => -2,
		expectations => -3,
		liquidity => -4,
		leverage =>	-5, 
		zero_savings_prob	=>	-6,
		timepref => 	-7,
		which_market	=> -8
	) ;
	my %specobjindex = (
			femhead => 1,
			innovativeness => 2,
			expectations => 3,
			liquidity => 4,
			leverage =>	5,
			zero_savings_prob	=>	6,
			timepref => 	7,
			which_market => 8
			  ) ;
      $self->set_specobj_info(\%specobjname,\%specobjindex,\%specobjtype);

}
sub prepare_shared_files {
	my $self = shift;
	my $dbh = $self->get_dbh;

	$self->define_specobjs;
	## Determine number of Segments by counting the thresholds in the list
	my @thresholds = $self->get_cfg_list("THRESHOLDS");
	$self->set_config_entry("SEGMENTS",  scalar @thresholds);
	
	### ObjectIDs for Network, Demography and Population
	# Distribute ObjectIDs and replace activity and constraint names by LP Column and Row Indices
	my $ki = 0;
	my @temp = sort {$a->id cmp $b->id} $self->networkobjects;
	my $counter = 0;
	for my $k (@temp){
		$k->set_oid($ki++);
		my $ac = $self->activity($k->act);
		my $co = $self->constraint($k->con);
		print STDERR  "Error in asset ".$k->id.": No such activity  (".$k->act.") \n" and exit(1) unless (defined $ac);
		print STDERR  "Error in asset ".$k->id.": No such constraint (endowment) (".$k->con.") \n" and exit(1) unless (defined $co);

		print STDERR "Error in asset ".$k->id.": Terrain out of range.\n" and exit(1) unless ($k->{terrain} < $self->config_entry("NSOIL"));

		my $lpc = $ac->lp_column;
		my $lpr = $co->lp_row;
		$k->set_act($lpc);
		$k->set_con($lpr);
		if ($k->perm ne "-1" and $k->perm ne "0" and defined $k->perm ) {
			my $perm = $self->constraint($k->perm);
			
			print STDERR "Error in asset ".$k->id.": No such constraint (yield) (".$k->perm.") \n" and exit(1) unless (defined $perm);
			my $lpr2 = $perm->lp_row;
			$k->set_perm($lpr2);
		}
		$counter++ if ($k->type == 2);

	}
	$self->set_config_entry("SYMBOLICS", $counter);

	#Distribute ObjectIDs for Liquidity, Female Lab etc.
	my %specindices = %{$self->get_specobj_indices()};
	my $temp2 = [];
	for my $k (keys %specindices ) {
	    
		$temp2->[ $specindices{$k} ] = $ki -1 + $specindices{$k};
	}
	$self->set_config_entry("SPECOBJS", $temp2);


	##Retrieve innohash from database or create empty 
	my $innohash;
	if (defined $self->config_entry("INNOVATIONS") ) {
		$innohash = $self->config_entry("INNOVATIONS")->transform($dbh, 3);
	}
	else {
	 	$innohash= IndexedTable->create_empty(3);
	}
	##Add innogrps to innohash that are not listed in innovations
	for my $k ($self->networkobjects){
		my $innogrp = $k->innogrp;
		for my $n (0.. $self->config_entry("NETWORKS") -1) {
			for my $s (0.. $self->config_entry("SEGMENTS") -1) {
				unless (defined $innohash->query([$n, $s, $innogrp], 4) ) {
					my @row;
					if (defined $innohash->query([$n, $s - 1, $innogrp], 4) ) {
						@row = ($n,$s,$innogrp, $innohash->query([$n, $s - 1, $innogrp], 4), $innohash->query([$n, $s - 1, $innogrp], 5) );
					}
					else {
						@row = ($n,$s,$innogrp, 0,1);
					}	
					$innohash->add_row(\@row);	
				}
			}
		}
	}
	# Determine InnovationIDs for Network and Segment
	my @innolist = $innohash->query([0,0],0);
	$ki = 0;
	my $innos ={};
	for my $k (0..$#innolist){
		my $in = "$innolist[$k]";
		$innos->{$in} = ++$ki;
	}

	for my $k ($self->networkobjects){
		my $f1 = $k->oid;
		my $r1 = $k->innogrp;
		my $f2 = $innos->{$r1};
		$k->set_innovid($f2);
	}	

	##Determine Order of Objects		
	my @SortedObjects = sort {$a->innovid <=> $b->innovid || $a->oid <=> $b->oid}  $self->networkobjects; ##
	my $g = -1;
	my @SortedObjectIds;
	for my $i (@SortedObjects){
		$i->set_network_row(++$g);
		push @SortedObjectIds, $i->id;
	}
	$self->set_sorted_networkobjects(\@SortedObjectIds);

	my $NN = 0; # Number of Investments in Network
	my @SortedObjectsInCurrentNetwork; 
	my %InnovationsInCurrentNetwork; # using hash instead of array  to ensure uniqueness
	for my $ddd (@SortedObjects) {
		for my $j (0..$self->config_entry("NETWORKS")-1){
				my $innofound = $ddd->innogrp;
				$InnovationsInCurrentNetwork{$innofound} = 1;
				push @SortedObjectsInCurrentNetwork, $ddd;
				$NN += 1;
				last;					
		}
	}
	#Sort InnovationsInCurrentNetwork
	my @SortedInnovationsInCurrentNetwork = sort {$innos->{$a} <=> $innos->{$b}} keys %InnovationsInCurrentNetwork;
	$self->set_global("Innovations", $innos);
	$self->set_global("InnovationInfo", $innohash);
	$self->set_global("InnovationsInCurrentNetwork", \%InnovationsInCurrentNetwork);
	$self->set_global("SortedInnovationsInCurrentNetwork", \@SortedInnovationsInCurrentNetwork);
}

sub prepare_special_files {
	my $self = shift;
	for my $f ($self->files) {
		$f->prepare_special();
	}
}
