package MpmasQLClassDefinitionLanguage;
use strict;
use warnings ;# FATAL => qw(all);
use Exporter;
use Storable qw(dclone);
use Data::Dumper;
use Carp qw(croak); #qw(croak verbose);
use CtUtilities qw(queryhash extqueryhash is_in_list);
use IndexedTable;
use Scalar::Util qw(looks_like_number);
use EmbeddedArray;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw();

#$SIG{__DIE__} = \&Carp::confess;

use constant {
 PI => 4 * atan2(1, 1),
};
my $singlemode;
my $dbh;
my $setup;
my $tableregistry;
my $verbose;
my $ciTempArrays;
my $TempArrays;
my $mapregistry;
my $fast = 0;
#my $forvalues = [];
my $locloudeb1 = 0;
my $locals_hash = {};
sub fast {
    $fast = 1;
}
sub set_globals {
	shift;
	$singlemode = shift;
	$dbh = shift;
	$setup = shift;
	$tableregistry = shift;	
	$verbose = shift||0;
}
sub set_mapregistry {
	shift;
	$mapregistry = shift;
}

sub convert {
	shift;
	my $coded = shift;
	my $locvarhash = shift;
	my @temp;
	eval { 	@temp = MpmasQLClassDefinitionLanguage->interprete($coded,$locvarhash); };
	croak "$@" if $@;
	if ($#temp > 0 or defined $temp[0][1] ) {
		my @array;
		for my $i (0..$#temp) {
			$array[$i] = $temp[$i][0] if (defined $temp[$i][0]) ;
		}
		return wantarray ? @array : \@array;
	}
	else {
		 return wantarray ? ($temp[0][0]) : $temp[0][0] ;
	}
}
sub convert_to_arrayrefs {
	shift;
	my $coded = shift;
	my $locvarhash = shift;
	my @temp;
	eval { 	@temp = MpmasQLClassDefinitionLanguage->interprete($coded,$locvarhash); };
	croak "$@" if $@;
	if ($#temp > 0 or defined $temp[0][1] ) {
		my @array;
		my @fors;
		for my $i (0..$#temp) {
			$array[$i] = $temp[$i][0] if (defined $temp[$i][0]) ;
			$fors[$i] =  @{$temp[$i]}[1..$#{$temp[$i]}] if (defined $temp[$i][1]) ;
		}
		return  \@array, \@fors;
	}
	else {
		 return [($temp[0][0])] , [(@ {$temp[0]}[1..$#{$temp[0]}])] ;
	}
}
sub interprete2 {	
	# read in variables passed to function

	my $class = shift;
	my $coded = shift;
	my $locvarhash = shift;
	croak "First argument to $class ->interprete should be string, not a reference!"  if (ref $coded);
	croak "Second argument to $class ->interprete should be a hash reference!"  unless (ref $locvarhash eq "HASH");
	my $tvars = $setup;
	my $ttabs = $tableregistry;

	my $locloudeb1 = $verbose;
	$ciTempArrays = 0;
	$TempArrays = {}; #ref to anonymous hash
	my @rt; #declare return array
	return ([("")]) unless (defined $coded);

	my @operators;
	my @operands;
	
	print "Starting to bind special variables in string '$coded' \n" if $locloudeb1;
	#check for debugging switch 
	if ($coded =~ s/@//g) {
		$locloudeb1 = 1;				
	}
	$coded =~ s/^\s*//;
	$coded =~ s/\s*$//;
	
	return ([("")]) unless (defined $coded);
	return ([("$coded")]) if ($coded =~ /^\w+$/);
	local $| = 1;
	#replace all special variables
	my @locvarlist = keys %{$locvarhash};
	@locvarlist = sort {loc_priority($locvarhash->{$a}) <=>  loc_priority($locvarhash->{$b}) } @locvarlist unless ($fast);
	for my $locvar (@locvarlist) {
		print "Looking for $locvar ". ref($locvarhash->{$locvar})."->\n" if $locloudeb1;
		
		while (ref $locvarhash->{$locvar} eq "ARRAY" and defined $coded and $coded =~ /#$locvar\[(\d+)\]/  ){
			my $x = $locvarhash->{$locvar}->[$1];
			$coded =~ s/#$locvar\[($1)\]/$x/;
			print "#$locvar"."[$1] replaced by $locvarhash->{$locvar}->[$1]". "-> $coded\n" if $locloudeb1;
		}
		if ((not ref $locvarhash->{$locvar}) and defined $coded and $coded =~ s/#$locvar/$locvarhash->{$locvar}/g  ) {
			print "#$locvar replaced by $locvarhash->{$locvar} -> $coded\n" if $locloudeb1;
		}
		while (ref $locvarhash->{$locvar}  and not (ref $locvarhash->{$locvar} eq "ARRAY") and defined $coded and $coded =~ /#$locvar\$(\w+)/  ){
			my $x = $locvarhash->{$locvar}->get_attribute($1);
			$x = "##NA" unless defined $x;
			$coded =~ s/#$locvar\$($1)/$x/;
			print "#$locvar"."\$$1 replaced by $x". "-> $coded\n" if $locloudeb1;
		}
		
	}
	return ([("")]) unless (defined $coded);
	return ([("$coded")]) if ($coded =~ /^\w+$/);
	#shunting_yard  algorithm to parse and process the code
	my $next_minus_is_unary = 1;
	while (length($coded) ) {
		#find an operator
		if ($coded =~ /^\s*([-*\/+~^\\]|>=|==|<=|<|>|!=)/) {	
				my $op = $1;
				my $x = "";
				$x = "\\" if ($op =~ /^([*\/+^\\])/);
				$coded =~ s/^\s*$x$op//;
				if ($op eq "-" and $next_minus_is_unary) {
					$op = "unary_minus" ;
				}
				elsif ($op eq "-" and !$next_minus_is_unary) {
					$op = "minus" ;
				}
				elsif ($op eq "+") {
					$op = "plus" ;
				}
				elsif ($op eq "*") {
					$op = "multiply" ;
				}
				elsif ($op eq "/") {
					$op = "divide" ;
				}
				elsif ($op eq "\\") {
					$op = "protected_divide" ;
				}
				elsif ($op eq "^") {
					$op = "power" ;
				}
				elsif ($op eq "~") {
					$op = "concatenate" ;
				}
				elsif ($coded =~ /^(>=|==|<=|<|>|!=)/) {
					my $op = $1;
					
				}
				my $do = 1;
				while ($do) {
					print "Operator Stack: ".join(",", @operators)."\n" if $locloudeb1;
					
					my $op_on_stack = pop @operators;
					if (not defined $op_on_stack) {
						$do = 0; last;
					}
					elsif (is_a_function($op_on_stack) or $op_on_stack eq "left_parenthesis" or $op_on_stack eq "left_bracket") {
						push @operators, $op_on_stack;
						$do = 0; last;
					}
					elsif (associativity($op) eq "left" and precedence($op) <= precedence($op_on_stack) ) {
						process_operator($op_on_stack, \@operands,$locloudeb1);
					}
					elsif (associativity($op) eq "right" and precedence($op) < precedence($op_on_stack) ) {
						process_operator($op_on_stack, \@operands,$locloudeb1);
					}
					else {
						push @operators, $op_on_stack;
						$do = 0; last;
					}
					
				}
				push @operators, $op;
				print "Pushed $op to Operator Stack\n" if $locloudeb1;
				print "Operator Stack: ".join(",", @operators)."\n" if $locloudeb1;
				$next_minus_is_unary = 1;
		}
		
		elsif ($coded =~ /^\s*#resolve\s*\(\s*(\w+)\s*\)/) {
				my $var = $1;
				croak "\$setup is not a HASH reference" unless (ref $setup eq "HASH");
				croak  "Error: Variable $var is  not defined or empty\n " unless (exists $setup->{$var});
				my $insert = $setup->{$var};
				$coded =~ s/^\s*#resolve\s*\(\s*($var)\s*\)/$insert/;
				print "\tCode insertion $setup->{$var} for #resolve($var) \n" if $locloudeb1;				
		}
		elsif ($coded =~ /^\s*#(\w+)\s*\(/) {
				my $op = $1;
				$coded =~ s/^\s*#$op\s*//;
				push @operators, $op;
				print "Pushed $op to Operator Stack\n" if $locloudeb1;
				$next_minus_is_unary = 1;
		}
		elsif ($coded =~ /^\s*,\s*/) {
				my $do = 1;
				while ($do) {
					print "Operator Stack: ".join(",", @operators)."\n" if $locloudeb1;
					my $op = pop @operators;
					
					if ($op eq "left_parenthesis") {
						push @operators, $op;
						$do = 0;last;
					}
					elsif ($op eq "left_bracket") {
						push @operators, $op;
						$do = 0;last;
					}
					elsif (not defined $op ) {
						croak "Misplaced comma in $coded\n"; 
					}
					else {
						process_operator($op, \@operands,$locloudeb1);
						
					}
				}
				$coded =~ s/^\s*,\s*//;
				$next_minus_is_unary = 1;
		}
		elsif ($coded =~ /^\s*\[\s*/) {
				push @operands, "left_bracket";
				push @operators, "left_bracket";
				print "Pushed left_bracket to Operator Stack and Operand stack\n" if $locloudeb1;
				$coded =~ s/^\s*\[\s*//;
				$next_minus_is_unary = 1;
		}
		elsif ($coded =~ /^\s*\]\s*/) {
				$coded =~ s/^\s*\]\s*//;
				my @newarray;
				my $do = 1;
				print "Encountered right bracket ->\n" if $locloudeb1;
				while ($do) {
					print "Operator Stack: ".join(",", @operators)."\n" if $locloudeb1;
					my $op = pop @operators;
					if ($op eq "left_bracket") {
						$do = 0;last;
					}
					elsif (not defined $op ) {
						croak "Misplaced ] in $coded\n";
					}
					else {
						process_operator($op, \@operands,$locloudeb1);
						
					}
				}
				$do = 1;
				while ($do) {
					my $op = pop @operands;
					if ($op eq "left_bracket") {
						$do = 0;last;
					}
					elsif (not defined $op ) {
						croak "Misplaced ] in $coded\n";
					}
					else {
						unshift @newarray, $op;
					}
				}
				print "..formed array and pushed it on operand stack\n" if $locloudeb1;
				push @operands, \@newarray;
				
				$next_minus_is_unary = 0;
		}
		elsif ($coded =~ /^\s*\(\s*/) {
				push @operators, "left_parenthesis";
				print "Pushed left_parenthesis to Operator Stack\n" if $locloudeb1;
				$coded =~ s/^\s*\(\s*//;
				$next_minus_is_unary = 1;
		}
		elsif ($coded =~ /^\s*\)\s*/) {
				$coded =~ s/^\s*\)\s*//;
				my $do = 1;
				print "Encountered right parenthesis ->\n" if $locloudeb1;
			
				while ($do) {
					print "Operator Stack: ".join(",", @operators)."\n" if $locloudeb1;
					my $op = pop @operators;
					if ($op eq "left_parenthesis") {
						$do = 0;last;
					}
					elsif (not defined $op ) {
						croak "Misplaced ) in $coded\n";
					}
					else {
						process_operator($op, \@operands,$locloudeb1);
											}
				}
				my $nextop= $operators[$#operators];
				if (is_a_function ($nextop )) {
					my $op = pop @operators;
					process_operator($op, \@operands,$locloudeb1);
					
				}
				
				$next_minus_is_unary = 0;
		}
		elsif ($coded =~ s/^\s*"([^"]*)"//) {
				push @operands, $1;
				print "Pushed $1 to Operand Stack\n" if $locloudeb1;
				#$coded =~ s/^\s*"$1"\s*//;
				$next_minus_is_unary = 0;
		}
		
		elsif ($coded =~/^\s*(\?ta\d+)\s*/) {
				push @operands, $1;
				print "Pushed Temporary Array: $1 to Operand Stack\n" if $locloudeb1;
				$coded =~ s/^\s*$1\s*//;
				$next_minus_is_unary = 0;
		}
		elsif ($coded =~ /^\s*(\d+.\d+(E|e)(\+|-)\d+)\s*/) {
				my $x = "$1";
				$coded =~ s/^\s*\Q$x\E//;
				$x =~ s/\s+/ /g;
				$x =~ s/\s*$//;
				push @operands, $x;
				print "Scientific Number $x found\n" if $locloudeb1;
				print "Pushed  $x to Operand Stack\n" if $locloudeb1;
				$next_minus_is_unary = 0;
		}
		elsif ($coded =~ /^\s*([\w.\$]+)/) {
				my $x = $1;
				$coded =~ s/^\s*\Q$x\E//;
				$x =~ s/\s+/ /g;
				$x =~ s/\s*$//;
				push @operands, $x;
				print "Pushed  $x to Operand Stack\n" if $locloudeb1;
				$next_minus_is_unary = 0;
		}
		elsif ($coded =~ /^\s*(##NA)/) {
				my $x = $1;
				$coded =~ s/^\s*\Q$x\E//;
				$x =~ s/\s+/ /g;
				$x =~ s/\s*$//;
				push @operands, $x;
				print "Pushed  $x to Operand Stack\n" if $locloudeb1;
				$next_minus_is_unary = 0;
		}

		elsif ($coded =~ s/^\s*"[^"]*$//) {
				croak "Quotes unbalanced!\n"
		}
		else {
				croak "Encountered uninterpretable string: $coded\n";
		}
		print "Operand Stack: ".join(" | ", @operands)."\n" if $locloudeb1;
	}
	while (@operators) {
		my $op = pop @operators;
		process_operator($op, \@operands, $locloudeb1);
	}
	my $decoded = "";
	while (@operands) {
		my $s =  pop @operands;
		if (any_NULL_in($s)) {
			return undef;
		}
		elsif (ref $s eq "ARRAY") {
			#$decoded = $s;
			#last;
			my $c = "ta".$ciTempArrays;
			$TempArrays->{$c} = $s;
			$s = "?$c";
			$ciTempArrays = $ciTempArrays +1;
		}
		
		$decoded = "$s"."$decoded";
	}
	print "Performing final concatenation --> $decoded\n" if $locloudeb1;

	### Expand arrays that have not been expanded yet
	EXPANDARRAYS:{
		#if (ref $decoded eq "ARRAY") {
		 #     
		  #    return undef if (scalar @$decoded < 1);
		   #   for my $k (0..$#{$decoded}) {
			#    $rt[$k][0] = $decoded->[$k];
			 #   $rt[$k][1] = $decoded->[$k];
	#	      }
	#	      last EXPANDARRAYS;
	#	}
		if (keys %$TempArrays < 1) {			
			$rt[0][0] = "$decoded";
			last EXPANDARRAYS;
		}
		my $switch = 0;
		if (keys %{$TempArrays} > 0) { 
			for my $k (keys %{$TempArrays}) {
				if (any_NULL_in($TempArrays->{$k}->[0])){
					return undef;
					last EXPANDARRAYS;
				}
			}
		}
		my @prert; # temporary variable to store intermediate results
		$prert[0][0] = $decoded;
		my $kcount = 1;
		my @sortedkeys = sort{$a cmp $b } keys %{$TempArrays};
		for my $k (@sortedkeys) { #loop over all arrays that have not been expanded yet
			for my $kk (0..$#prert) { # loop over all strings resulting from previous expansions
				for my $l (0..$#{$TempArrays->{$k}}){
					my $curi = $kk * ($#{$TempArrays->{$k}}+1) + $l;
					my $ArrayEntry = $TempArrays->{$k}->[$l];
					print "Array Entry $l: $ArrayEntry\n" if $locloudeb1;
					my $manipulated =  $prert[$kk][0] ;			
					$manipulated =~ s/\?$k/$ArrayEntry/;
					$rt[$curi][0] = $manipulated;	#return decoded string
					print "Expanding Array - Index $l -> $manipulated\n" if $locloudeb1;
					for my $rr (1..$kcount) {
						if ($rr < $kcount) {
							$rt[$curi][1][$rr] = $prert[$kk][1][$rr];#copy results from previous round
						}
						else {
							$rt[$curi][1][$rr] = $ArrayEntry;   #return 'pure' array entry for use in #thisfor
						}
					}
				}
			}
			undef @prert;
			for my $i (0..$#rt) {  #deep copy results back to prert; deep copy is necessary otherwise
				$prert[$i][0] = $rt[$i][0]; 
				for my $j (0..$#{$rt[$i][1]}) {
					$prert[$i][1][$j] = $rt[$i][1][$j]; 
				}
			}
			$kcount += 1;
		}
		last EXPANDARRAYS;
	}
	print "Finished decoding string $rt[0][0] \n\n" unless (defined $rt[0][0]); # if $locloudeb1;
	return @rt; #rt has  two dimensions, but dimensions actually used depend on return value of sub interprete
					#scalar return -> only $rt[0][0] defined
						#return of field list -> $rt [..][0] field name, $rt[..][1..x] List of #thisfor values



	
}

sub precedence {
	my $op = shift;
	my %precedencetab = (
		unary_minus => -10,
		left_bracket => 0,
		function => 0,
		power => -20,
		divide => -30,
		protected_divide => -30,
		multiply => -30,
		modulus => -30,
		plus => -40,
		minus => -40,
		concatenate => -50,
		"<" => -60,
		"<=" => -60,
		">" => -60,
		">=" => -60,
		"==" => -60,
		"!=" => -60
	);
	return $precedencetab{function} unless exists $precedencetab{$op};
	return $precedencetab{$op};
}
sub associativity {
	my $op = shift;
	my %associativitytab = (
		unary_minus => "right",
		left_bracket => "left",
		function => "left",
		power => "right",
		divide => "left",
		protected_divide => "left",
		multiply => "left",
		modulus => "left",
		plus => "left",
		minus => "left",
		concatenate => "left",
		"<" => "left",
		"<=" => "left",
		">" => "left",
		">=" => "left",
		"==" => "left",
		"!=" => "left"
	);
	return $associativitytab{function} unless exists $associativitytab{$op};
	return $associativitytab{$op};
}
sub is_a_function {
	my $op = shift;
	return 0 unless (defined $op);
	my %operatortab = (
		unary_minus => 1,
		power => 1,
		divide => 1,
		protected_divide => 1,
		multiply => 1,
		modulus => 1,
		plus => 1,
		minus => 1,
		concatenate => 1,
		left_parenthesis => 1,
		"<" => 1,
		"<=" => 1,
		">" => 1,
		">=" => 1,
		"==" => 1,
		"!=" => 1
	);
	return 1 unless exists $operatortab{$op};
	return 0;
}

sub process_operator {
	my $op = shift;
	my $stack = shift;
	my $locloudeb1 = shift;
	print "Processing $op:" if $locloudeb1;
	if ($op eq "concatenate") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		print "Concatenating '$o1' and '$o2'\n" if $locloudeb1;
		my $res = (any_NULL_in($o1, $o2)) ? "##NA" : ($o2.$o1);
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
		
	}
	elsif ($op eq "plus") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		my $res = any_NULL_in($o1, $o2) ? "##NA" :$o2 + $o1;
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "minus") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		my $res = any_NULL_in($o1, $o2) ? "##NA" :$o2 -$o1;
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "modulus") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		
		my $res = any_NULL_in($o1, $o2) ? "##NA":$o2 %$o1;
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "multiply") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		
		my $res = any_NULL_in($o1, $o2) ? "##NA" :$o2 *$o1;
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "divide") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		my $res = any_NULL_in($o1, $o2) ? "##NA" :$o2 /$o1;
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "protected_divide") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		my $res = ( any_NULL_in($o1, $o2)) ? "##NA" : ($o1 == 0   ?  0 : ($o2 / $o1) );
		push @$stack, $res;
		print "\t Result:  ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "power") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		my $res = any_NULL_in($o1, $o2) ? "##NA" :$o2 ** $o1;
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "unary_minus") {
		my $o1 = pop @$stack;
		handle_if_array("croak",$o1);
		my $res = any_NULL_in($o1) ? "##NA" : (-1 * $o1);
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ( is_in_list( $op ,[(">", ">=", "<=", "==", "<","!=")])) {
		
		my $o2 = pop @$stack;
		my $o1 = pop @$stack;
		handle_if_array("croak",$o1,$o2);
		my $string = $o1 ." ". $op." ". $o2;
		my $res = eval $string;
		$res = $res  ? 1 : 0;
		push @$stack, $res;
		print "compare $string\n" if $locloudeb1;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;	
	}
	elsif ($op eq "round_to") {
		my $o1 = pop @$stack;
		my $o2 = pop @$stack;
		handle_if_array("croak",$o2);
		my $res = (any_NULL_in($o2) or any_NULL_in($o2)) ? "##NA" : sprintf("%.${o1}f",$o2);
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "floor") {
		my $o1 = pop @$stack;
		handle_if_array("croak",$o1);
		my $res;
		if (any_NULL_in($o1) ) {
			$res = "##NA" ;
		}
		else {
			$o1 =~ /^(\d+).?/;
			if (defined $1) {
				$res = $1;
			}
			else  {
				$res = 0;
			}
		}
		push @$stack, $res;
		print "\t Result: ".$res." written to operands\n" if $locloudeb1;
	}
	elsif ($op eq "steps") {
		my $step = pop @$stack;
		my $end = pop @$stack;
		my $start = pop @$stack;
		handle_if_array("croak",$step, $start, $end);
		if (any_NULL_in($step, $start, $end) or $start > $end or $step == 0) {
			push @$stack, '##NA';
			print "\t ##NA pushed to operands\n" if $locloudeb1;
		}
		else {
			my @newlist;
			my $le = $start;
			my $cc = 0;
			while ($le <= $end) {
				$newlist[$cc]	= $le;
				$le += $step;
				$cc++;
			}
			my $c = "ta".$ciTempArrays;
			$TempArrays->{$c} = \@newlist;
			push @$stack, "?$c";
			$ciTempArrays = $ciTempArrays +1;
			print "\t List $start to $end in steps of $step pushed to operands\n" if $locloudeb1;
		}
	}
	elsif ($op eq "normlike") {
	    my $sigma = pop @$stack;
	    my $sim = pop @$stack;
	    my $obs = pop @$stack;
	    handle_if_array("croak",$sigma, $sim, $obs);
	    if (any_NULL_in($sigma, $sim, $obs) ) {
			push @$stack, '##NA';
			print "\t ##NA pushed to operands\n" if $locloudeb1;
	    }
	    my $result = 1 / ( sqrt(2 * PI ) * $sigma ) * exp ( -1 * ( $sim - $obs  ) ** 2 /  (2 * $sigma ** 2));
	    push @$stack, $result;
	    print "\t Result $result pushed to operands\n"  if $locloudeb1;
	}
	elsif ($op eq "sql") {
		my $get1 = pop @$stack;
		print "\t Performing database query ->\n " if $locloudeb1;
		if (any_NULL_in($get1)) {
			push @$stack, '##NA';
			print "\t ##NA pushed to operands\n" if $locloudeb1;
		}		
		else {
	
	
			my $c = "ta".$ciTempArrays;
			#if you get an array use only the first entry and delete array
			($get1) = handle_if_array("use_first",$get1);
			
			my $query =[];
			eval{	#exception handling to avoid fatals because table column does not exist 
				#!!CT: revise as for now it captures all warnings, better solutions after restructuring of database
				local $SIG{__WARN__}=sub{}; #Temporarily suppress warnings
				$query = $dbh->selectcol_arrayref($get1);
			};
			if ($@) {	
				$TempArrays->{$c} =[('NA')];
				croak "Query error captured\n$@\nQuery string was : $get1\n\n";
				
			}
			elsif (not defined $query->[0]) {
				
				print "Query resulted in NULL\n" if $::loudcprep;
				$query->[0] = '##NA';
				push @$stack, $query->[0];
				print "\t Result (NULL) pushed to operands\n"  if $locloudeb1;
				
			}
			elsif ($#{$query} == 0) {
				push @$stack, $query->[0];
				print "\t Result ($query->[0]) pushed to operands\n"  if $locloudeb1;
			}
			else {
				$TempArrays->{$c} = [@{$query}];
				push @$stack, "?$c";
				$ciTempArrays = $ciTempArrays +1;
				print "\t Result ($query->[0],..) temporarily saved in ?$c\n"  if $locloudeb1;
			}
		}
	}
	elsif ($op eq "if") {
		my $else = pop @$stack;
		my $then = pop @$stack;
		my $condition = pop @$stack;
		if ($condition) {
			print "if ($condition) evaluated true, $then pushed to operands\n" if $locloudeb1;
			push @$stack, $then;
		}
		else {
			push @$stack, $else;
			print "if ($condition) evaluated false, $else pushed to operands\n" if $locloudeb1;
		}
		
		
	}
	elsif ($op eq "ifthenelse") {
		my $else = pop @$stack;
		my $then = pop @$stack;
		my $condition = pop @$stack;
		my $ev = eval  $condition;
		if ($ev) {
			print "ifthenelse evaluated true, $then pushed to operands\n" if $locloudeb1;
			push @$stack, $then;
		}
		else {
			push @$stack, $else;
			print "ifthenelse evaluated false, $else pushed to operands\n" if $locloudeb1;
		}
		
		
	}
	elsif ($op eq "replace_null") {
		print "\t\tResult of replace_null:\t" if $locloudeb1;
		my $replaceby = pop @$stack;
		my $what = pop @$stack;
		if (any_NULL_in($what)){
			push @$stack, $replaceby;
			print "\t\tReplaced: 	$replaceby" if $locloudeb1;
		}
		else{
			push @$stack, $what;
			print "\t\tKept:	$what" if $locloudeb1;
		}
		print "\n" if $locloudeb1;

	}
	elsif ($op eq "ifalone"){
		print "\t\tResult of ifalone:\t" if $locloudeb1;
		my $ifsmat2 = pop @$stack;
		my $ifsmat1 = pop @$stack;
		if ($singlemode){
			push @$stack, $ifsmat1;
			print "\t\tOption 1: 	$ifsmat1" if $locloudeb1;
		}
		else{
			push @$stack, $ifsmat2;
			print "\t\tOption 2:	$ifsmat2" if $locloudeb1;
		}
		print "\n" if $locloudeb1;
	}
	elsif ($op eq "annuity"){
		print "\t\tCalculating Annuity:\n"  if $locloudeb1;
		my $anLifespan = pop @$stack;
		my $anCapital = pop @$stack;
		my $anInterest = pop @$stack;
		#if you get an array use only the first entry and delete array
		if (any_NULL_in($anLifespan, $anCapital, $anInterest)) {
			push @$stack, '##NA';
			print "\t ##NA pushed to operands\n" if $locloudeb1;
		}		
		else {
			($anLifespan, $anCapital, $anInterest) = handle_if_array("use_first",$anLifespan, $anCapital, $anInterest);
			
			print "\t\tLifespan:  $anLifespan\n\t\tAcqCost: $anCapital\n\t\tInterest rate: $anInterest\n" if $locloudeb1;
			
			my $an = $anCapital * (( 1 + $anInterest)**$anLifespan *$anInterest) / ((1 +$anInterest)**$anLifespan -1);
			print "\t\tResult: $an pushed to operands\n" if $locloudeb1;
			push @$stack, $an;
		}
	}
	elsif ($op eq "foreach"){
		my $slist = pop @$stack;
		my $c = "ta".$ciTempArrays;
		croak  "Error: Foreach list $slist  is \n not defined or empty\n " unless (defined $setup->{$slist});
		$TempArrays->{$c} = [@{$setup->{$slist}}];
		push @$stack, "?$c";
		$ciTempArrays = $ciTempArrays +1;
		print "\t Expanded list $slist temporarily saved in ?$c\n" if $locloudeb1;
	}
	elsif ($op eq "count"){
		my $slist = pop @$stack;
		croak  "Error: List $slist  is \n not defined or empty\n " unless ($setup->{$slist});
		my @list =  @{$setup->{$slist}} ;
		my $listsize = scalar @list;
		push @$stack, $listsize;
		print "\t Count: $slist has $listsize elements\n" if $locloudeb1;
	}
	elsif ($op eq "random_number"){
		my $ub = pop @$stack;
		croak  "Error: Argument ($ub) to random_number invalid\n " unless (looks_like_number($ub));
		my $res = rand($ub);
		push @$stack, $res;
		print "\tRandom number drawn: $res\n" if $locloudeb1;
		
	}
	elsif ($op eq "listelement"){
		my $element = pop @$stack;
		my $slist = pop @$stack;
		croak  "Error: List $slist  is \n not defined or empty\n " unless $setup->{$slist};
		my $c = $setup->{$slist}->[$element -1]    ;
		push @$stack, $c;
		print "\t Inserted $c for #listelement($slist, $element)\n" if $locloudeb1;
	}
	elsif ($op eq "var") {
		my $var = pop @$stack;
		croak "\$setup is not a HASH reference" unless (ref $setup eq "HASH");
		croak  "Error: Variable $var is  not defined or empty\n " unless (exists $setup->{$var});
		push @$stack, $setup->{$var};
		print "\tInsertion $setup->{$var} for #var($var)  \n" if $locloudeb1;
	}
	elsif ($op eq "list") {
		my $slist = pop @$stack;
		my @list = @{$setup->{$slist}};
		my $qlist = join (", ", @list);
		push @$stack, $qlist;
		print "\tinserted $qlist for $slist \n" if $locloudeb1;
	}
	elsif ($op eq "qlist") {
		my $slist = pop @$stack;
		my @list = @{$setup->{$slist}};
		my $qlist = "'".join("', '", @list)  ."'"; 
		push @$stack, $qlist;
		print "\tinserted $qlist for $slist\n" if $locloudeb1;
	}
	elsif ($op eq "table") {
		
		my $keylist = pop @$stack;
		my $column = pop @$stack;
		my $tablename = pop @$stack;
		my @keyarray = @$keylist;
		if (any_NULL_in($tablename, $column, @keyarray)) {
			push @$stack, '##NA';
			print "\t ##NA pushed to operands\n" if $locloudeb1;
		}		
		else {
			
			my $ttabs = $tableregistry;
			print "Table query : $tablename, $column, [".join(",", @keyarray)."]\n" if $locloudeb1;
			croak "Error: No tables (#table) defined in this environment!\n" unless (defined $ttabs);
			my $tabtouse = $ttabs->get_object_by_id($tablename);
			croak "Table (#table) $tablename unknown" unless ( defined $tabtouse);
			my $qresult =  [ $tabtouse->query(\@keyarray,$column) ];
			my $repl;
			if (not defined $qresult->[0]) {
				#using the array method because then the NA can be caught in the EXPANDARRAYS routine and the whole field value is set to 0, not creating a field even if there were pure string elements
				print "Query resulted in NULL\n" if $locloudeb1;
				$repl = '##NA';
				
			}
			elsif ($#{$qresult} == 0 and scalar @keyarray == $tabtouse->numkeys) { #makes sure a scalar is only reported, when it was the intention to do so (otherwise 1 item results for list queries would not produce thisfor values
				$repl = $qresult->[0];
			}
			else {
				my $c = "ta".$ciTempArrays;
				$TempArrays->{$c} = [@{$qresult}];
				$repl = "?$c";
				$ciTempArrays = $ciTempArrays +1;
	
			}
			print "\t\tResult: $repl pushed to operands\n" if $locloudeb1;
			push @$stack, $repl ;
		} 
	}
	elsif ($op eq "mapcell") {
		my $y = pop @$stack;
		my $x = pop @$stack;
		my $map = pop @$stack;
		print "Map query: $map, $x, $y:\n" if $locloudeb1;
		croak ("Error: No maps loaded\n") unless (defined $mapregistry);
		my $mo = $mapregistry->get_object_by_id($map);
		croak ("Error: Map $map unknown\n") unless (defined $mo);
		my $repl = $mo->get_cell($x,$y);
		print "\t\tResult: $repl pushed to operands\n" if $locloudeb1;
		push @$stack, $repl;
	}
	else {
		croak "Error: unknown operator ('$op')\n";
		last OPSWITCH;
	}
	print "Operand Stack: ".join(" | ", @$stack)."\n" if $locloudeb1;
}

sub handle_if_array {
	my $whattodo = shift;
	my @list_of_values = @_;
	my @newvalues;
	for my $i (@list_of_values) {
		my $new;
		while ($i =~ /\?(ta\d+)/){
			
			if ($whattodo eq "use_first") {
				my $arrayentry =  $TempArrays->{$1}->[0];
				delete $TempArrays->{$1};
				$i =~ s/\?$1/$arrayentry/;
				print "Warning: Found an array where a scalar was expected. Used only first entry!\n";
			}
			elsif ($whattodo eq "croak") {
				croak "Operator cannot handle an array!\n"
			}
		}
		push @newvalues, $i;
	}
	return @newvalues;
}
sub any_NULL_in {
	while (@_) {		
		my $x = shift(@_);
		return 1 if (not defined $x or $x eq "##NA");
		
	}
	return 0;
}

###Function for parsing and interpreting of Class Definition Code
sub interprete_old {	# expected arguments:  #thisinst, #thiselement, #thisfield, #thisby, #thisfor[array reference], interpretable string}
		# read in variables passed to function
		#my $dbh = shift;
		#my $singlemode = shift;
		my $class = shift;
		my $coded = shift;
		my $locvarhash = shift;
		croak "Second argument to $class ->interprete should be a hash reference!"  unless (ref $locvarhash eq "HASH");
		my $tvars = $setup;
		my $ttabs = $tableregistry;

		my $locloudeb1 = $verbose;
		my $ciTempArrays = 0;
		my $TempArrays = {}; #ref to anonymous hash
		my @rt; #declare return array

		print "Starting to bind variables in string '$coded' \n" if $locloudeb1;
		
		
		my $checkagain = 1;
		while ($checkagain) {
			#check for debugging switch 
			if ($coded =~ s/@//g) {
				$locloudeb1 = 1;
				$coded =~ s/^\s*//;
				$coded =~ s/\s*$//;
							
			}
			#replace all variables
			for my $locvar (keys %{$locvarhash}) {
				if (ref $locvarhash->{$locvar} and $coded =~ s/#$locvar\[(\d+)\]/$locvarhash->{$locvar}->[$1]/g  ){
					print "#$locvar"."[$1] replaced by $locvarhash->{$locvar}->[$1]". "-> $coded\n" if $locloudeb1;
				}
				elsif ((not ref $locvarhash->{$locvar}) and $coded =~ s/#$locvar/$locvarhash->{$locvar}/g  ) {
					print "#$locvar replaced by $locvarhash->{$locvar} -> $coded\n" if $locloudeb1;
				}
			}


			if ($coded =~ /#var\(\s*(\w+)\s*\)/){
				my $var = $1;
				croak "\$tvars is not a HASH reference" unless (ref $tvars eq "HASH");
				croak  "Error: Variable $var used in $coded is \n not defined or empty\n " unless (exists $tvars->{$var});
				my $res = $tvars->{$var};
				
				$coded =~ s/#var\(\s*$var\s*\)/$res/;
				print "\tinserted $res for $var -> '$coded' \n" if $locloudeb1;
				$checkagain = 1;
			} 
			elsif ($coded =~ /#qlist\(\s*(\w+)\s*\)/) {
				my $slist = $1;
				my @list = @{$tvars->{$slist}};
				my $qlist = ""; 
				for my $el (0..$#list) {
					if ($el == 0) {
						$qlist = "'".$list[$el]."'";
					}else{
						$qlist = $qlist.", '".$list[$el]."'";
					}
				}
				$coded =~ s/#qlist\(\s*$slist\s*\)/$qlist/;
				print "\tinserted $qlist for $slist -> '$coded' \n" if $locloudeb1;
				$checkagain = 1;
			}
			elsif ($coded =~ /#list\(\s*(\w+)\s*\)/) {
				my $slist = $1;
				my @list = @{$tvars->{$slist}};
				my $qlist = ""; 
				for my $el (0..$#list) {
					if ($el == 0) {
						$qlist = $list[$el];
					}else{
						$qlist = $qlist.", ".$list[$el];
					}
				}
				$checkagain = 1;
				$coded =~ s/#list\((\s*$slist\s*)\)/$qlist/;
				print "\tinserted $qlist for $slist -> '$coded' \n" if $locloudeb1;
				
			}
			else{
				$checkagain = 0;
			}
		}
		#second loop for tables
		
		$checkagain = 1;
		while ($checkagain) {
			if ($coded =~ /\s*#table\(\s*(\w*)\s*,\s*(\d*)\s*,\s*\[([\w\s,-]*)\]\s*\)/){
				my $keylist = $3;
				my $column = $2;
				my $tablename = $1;
				my $tempkeylist = $keylist;
				$tempkeylist =~ s/^\s*//;
				$tempkeylist =~ s/\s*$//;
				my @keyarray = split /\s*,\s*/ , $tempkeylist;
				croak "Error: No tables (#table) defined in this environment!\n" unless (defined $ttabs);
				my $tabtouse = $ttabs->get_object_by_id($tablename);
				croak "Table (#table) $tablename unknown" unless ( defined $tabtouse);
				my $qresult =  [ $tabtouse->query(\@keyarray,$column) ];
				my $repl;
				if (not defined $qresult->[0]) {
					#using the array method because then the NA can be caught in the EXPANDARRAYS routine and the whole field value is set to 0, not creating a field even if there were pure string elements
					print "Query resulted in NULL\n" if $locloudeb1;
					$qresult->[0] = "NA";
					my $c = "ta".$ciTempArrays;
					$TempArrays->{$c} = [@{$qresult}];
					$repl = "?$c ";
					$ciTempArrays = $ciTempArrays +1;
					
				}
				
				else {
					my $c = "ta".$ciTempArrays;
					$TempArrays->{$c} = [@{$qresult}];
					$repl = "?$c";
					$ciTempArrays = $ciTempArrays +1;

				}
				$checkagain = 1;
				$coded =~ s/#table\(\s*$tablename\s*,\s*$column\s*,\s*\[$keylist\]\)/$repl/;
				next;

			} 
			elsif ($coded =~ /\s*#table\(/) {
				croak "Syntax error in #table statement in $coded \n" ;
			}
			else {
				$checkagain = 0;
			}
		}
		
		#return if no function in code
		if (not($coded =~ /#/) and not ($coded =~ /\?ta/)) {
			$rt[0][0] = $coded;
			return @rt;	
		}
		
		##initializations of control and stack variables for command parsing
		my $unsolved = 1;
		my @operands;
		my @operators;
		
		my $decoded = "";
		print "Start decoding of string:\t '$coded'\n" if $locloudeb1;
		while ($unsolved) {
			STRINGEVAL:{

				if ($coded =~ /^[\s,]*#(\w*)\(/ ){
					if ($1) {
						push @operators, $1;
						print "\t$1 pushed to operators\n" if $locloudeb1;
					}
					else {
						print "\tFound open bracket\n" if $locloudeb1;
					}
					
					$coded =~ s/^[\s,]*#(\w*)\(//;
					
					last STRINGEVAL;
				}
				if ($coded =~/^[\s,]*(\?ta\d+)[\s,]*/) {
					push @operands, $1;
					$coded =~ s/^[\s,]*(\?ta\d+)[\s,]*//;
					print "\t $1 pushed to operands\n" if $locloudeb1;
					last STRINGEVAL;
				}
				if ($coded =~ /^[\s,]*"([^"]*)"[\s,]*/){
					push @operands, $1;	
					$coded =~ s/^[\s,]*"([^"]*)"[\s,]*//;
					print "\t $1 pushed to operands\n" if $locloudeb1;
					last STRINGEVAL;
				}
				if ($coded =~ /^[\s,]*\[([^\]]*)\][\s,]*/){
					push @operands, $1;	
					$coded =~ s/^[\s,]*\[([^\]]*)\][\s,]*//;
					print "\t $1 pushed to operands\n" if $locloudeb1;
					last STRINGEVAL;
				}
				if ($coded =~ /^[\s,]*(\w+)[\s,]*/ ){
					push @operands, $1;	
					$coded =~ s/^[\s,]*(\w+)*//;
					print "\t $1 pushed to operands\n" if $locloudeb1;
					last STRINGEVAL;
				}
				if ($coded =~ /^[\s,]*(-?\d*\.\d+)[\s,]*/ ){
					push @operands, $1;	
					$coded =~ s/^[\s,]*(-?\d*\.\d+)//;
					print "\t $1 pushed to operands\n" if $locloudeb1;
					last STRINGEVAL;
				}
				if ($coded =~ /^\s*\)/){
					my $op = pop @operators;
					croak "Error in class file: more closing brackets than operators in $coded!!!!\n" unless $op;
					$coded =~ s/^\s*\)//;
					OPSWITCH: {
						if ($op eq "cc") {
							print "\t Performing concatenation ->\n " if $locloudeb1;
							my $o1 = pop @operands;
							my $o2 = pop @operands;
							push @operands, $o2.$o1;
							print "\t Result:  ".$o2.$o1." written to operands\n" if $locloudeb1;
							last OPSWITCH;
						}
						if ($op eq "sql") {
							my $get1 = pop @operands;
							print "\t Performing database query ->\n " if $locloudeb1;
							my $c = "ta".$ciTempArrays;
							#if you get an array use only the first entry and delete array
							while ($get1 =~ /\?(ta\d+)/){
								my $arrayentry =  $TempArrays->{$1}->[0];
								delete $TempArrays->{$1};
								$get1 =~ s/\?$1/$arrayentry/;
							}
							my $query =[];
							eval{	#exception handling to avoid fatals because table column does not exist 
								#!!CT: revise as for now it captures all warnings, better solutions after restructuring of database
								local $SIG{__WARN__}=sub{}; #Temporarily suppress warnings
								$query = $dbh->selectcol_arrayref($get1);
							};
							if ($@) {	
								$TempArrays->{$c} =[('NA')];
								print "Query error captured in $coded\n$@\n Error? or just dirty ini/class/database structure?\n";# if $locloudeb1;
								print "Result of query set to $TempArrays->{$c}->[0],\ as query returned error of $@\n Query string was : $get1\n\n" ;#if $::loudcprep;
								croak "\n\n";
							}
							elsif (not defined $query->[0]) {
								
								print "Query resulted in NULL\n" if $::loudcprep;
								$query->[0] = 'NA';
								$TempArrays->{$c} = [@{$query}];
							}	
						 	else {
								$TempArrays->{$c} = [@{$query}];
							}
							push @operands, "?$c";
							$ciTempArrays = $ciTempArrays +1;
							print "\t Result ($query->[0],..) temporarily saved in ?$c\n"  if $locloudeb1;
							last OPSWITCH;
						}
						if ($op eq "ifalone"){
							print "\t\tResult of ifalone:\t" if $locloudeb1;
							my $ifsmat2 = pop @operands;
							my $ifsmat1 = pop @operands;
							if ($singlemode){
								push @operands, $ifsmat1;
								print "\t\tOption 1: 	$ifsmat1" if $locloudeb1;
							}
							else{
								push @operands, $ifsmat2;
								print "\t\tOption 2:	$ifsmat2" if $locloudeb1;
							}
							print "\n" if $locloudeb1;
							last OPSWITCH;
						}
						if ($op eq "annuity"){
							print "\t\tCalculating Annuity:\n"  if $locloudeb1;
							my $anLifespan = pop @operands;
							my $anCapital = pop @operands;
							my $anInterest = pop @operands;
							#if you get an array use only the first entry and delete array
							if ($anLifespan =~ /^\s*\?(\w+)/){
								$anLifespan =  $TempArrays->{$1}->[0];
								delete $TempArrays->{$1};
							}
							if ($anCapital =~ /^\s*\?(\w+)/){
								$anCapital =  $TempArrays->{$1}->[0];
								delete $TempArrays->{$1};
							}
							if ($anInterest =~ /^\s*\?(\w+)/){
								$anInterest =  $TempArrays->{$1}->[0];
								delete $TempArrays->{$1};
							}
							print "\t\tLifespan:  $anLifespan\n\t\tAcqCost: $anCapital\n\t\tInterest rate: $anInterest\n" if $locloudeb1;
							
							my $an = $anCapital * (( 1 + $anInterest)**$anLifespan *$anInterest) / ((1 +$anInterest)**$anLifespan -1);
							print "\t\tResult: $an pushed to operands\n" if $locloudeb1;
							push @operands, $an;
							last OPSWITCH;
						}
						if ($op eq "foreach"){
							my $slist = pop @operands;
							my $c = "ta".$ciTempArrays;
							croak  "Error: Foreach list $slist used in $coded is \n not defined or empty\n " unless $tvars->{$slist};
							$TempArrays->{$c} = [@{$tvars->{$slist}}];
							push @operands, "?$c";
							$ciTempArrays = $ciTempArrays +1;
							print "\t Expanded list $slist temporarily saved in ?$c\n" if $locloudeb1;
							last OPSWITCH;
						}
					
						else {
							croak "Error in classfile: unknown operator ('$op') in $coded\n";
							last OPSWITCH;
						}
					} #end of OPSWITCH
					
					last STRINGEVAL;
				} 
				if ($coded =~ /^\s*$/)	{
						croak "Error in class file: more operators than closing brackets \n" if @operators;
						print "Reached end of string.\n" if $locloudeb1;
						
						while (@operands) {
							my $s =  pop @operands;
							$decoded = "$s"."$decoded";
						}
						print "Performing final concatenation --> $decoded\n" if $locloudeb1;
						$unsolved = 0;
						
					last STRINGEVAL;	
				}
				else {
						croak "Error in classfile: ! Character not interpretable!\n near\t $coded \n\nMaybe opening bracket without '#'\nor whitespace between '#' and '('?\n\n";
						last STRINGEVAL;
				}
			} # END of STRINGEVAL
		} 


		### Expand arrays that have not been expanded yet
		EXPANDARRAYS:{
			if (keys %$TempArrays < 1) {
				
				$rt[0][0] = "$decoded";
				
				last EXPANDARRAYS;
			}
			my $switch = 0;
			if (keys %{$TempArrays} > 0) { 
				for my $k (keys %{$TempArrays}) {
					if ($TempArrays->{$k}->[0] =~ /^\s*-?NA\s*/){
						$rt[0][0] = 0;
						
						last EXPANDARRAYS;
					}
				}

			}
			my @prert; # temporary variable to store intermediate results
			$prert[0][0] = $decoded;
			my $kcount = 1;

			my @sortedkeys = sort{$a cmp $b } keys %{$TempArrays};
			for my $k (@sortedkeys) { #loop over all arrays that have not been expanded yet
				#print "$k\n" if $switch == 1;
				
			      	for my $kk (0..$#prert) { # loop over all strings resulting from previous expansions
					
					for my $l (0..$#{$TempArrays->{$k}}){
						my $curi = $kk * ($#{$TempArrays->{$k}}+1) + $l;
						my $ArrayEntry = $TempArrays->{$k}->[$l];
						print "Array Entry $l: $ArrayEntry\n" if $locloudeb1;
						my $manipulated =  $prert[$kk][0] ;
						
						$manipulated =~ s/\?$k/$ArrayEntry/;
						
						$rt[$curi][0] = $manipulated;	#return decoded string
						
						print "Expanding Array - Index $l -> $manipulated\n" if $locloudeb1;
						for my $rr (1..$kcount) {
							if ($rr < $kcount) {
								$rt[$curi][1][$rr] = $prert[$kk][1][$rr];#copy results from previous round
							}
							else {
								$rt[$curi][1][$rr] = $ArrayEntry;   #return 'pure' array entry for use in #thisfor
							}
						}
						
					}
				}
				
				undef @prert;
				for my $i (0..$#rt) {  #deep copy results back to prert; deep copy is necessary otherwise
					$prert[$i][0] = $rt[$i][0]; 
					for my $j (0..$#{$rt[$i][1]}) {
						$prert[$i][1][$j] = $rt[$i][1][$j]; 
					}
				}
				$kcount += 1;
			}
			last EXPANDARRAYS;
		}
		print "Finished decoding string $rt[0][0] \n\n" unless (defined $rt[0][0]); # if $locloudeb1;
		return @rt; #rt has  two dimensions, but dimensions actually used depend on return value of sub interprete
						#scalar return -> only $rt[0][0] defined
						#return of field list -> $rt [..][0] field name, $rt[..][1..x] List of #thisfor values
}
sub loc_priority {
	my $el = shift;
	return 200 if (not ref $el);
	return 100 if (ref $el eq "ARRAY");
	return 300 if (ref $el);
}	

sub interprete {
	my ($class, $coded, $locvarhash) = @_;
	croak "First argument to $class ->interprete should be string, not a reference!"  if (ref $coded);
	croak "Second argument to $class ->interprete should be a hash reference!"  unless (ref $locvarhash eq "HASH");
	$locloudeb1 = $verbose;

	my @rt; #declare return array
	return ([("")]) unless (defined $coded);

	
	
	print "Starting to bind special variables in string '$coded' \n" if $locloudeb1;
	#check for debugging switch 
	if ($coded =~ s/@//g) {
		$locloudeb1 = 1;				
	}
	$coded =~ s/^\s*//;
	$coded =~ s/\s*$//;
	
	return ([("")]) unless (defined $coded);
	return ([("$coded")]) if ($coded =~ /^\w+$/);
	local $| = 1;
	#replace all special variables
	my @locvarlist = keys %{$locvarhash};
	@locvarlist = sort {loc_priority($locvarhash->{$a}) <=>  loc_priority($locvarhash->{$b}) } @locvarlist unless ($fast);
	for my $locvar (@locvarlist) {
		print "Looking for $locvar ". ref($locvarhash->{$locvar})."->\n" if $locloudeb1;
		if (ref $locvarhash->{$locvar} eq "ARRAY") {
		    while ( defined $coded and $coded =~ /#$locvar\[(\d+)\]/  ){
			    my $x = $locvarhash->{$locvar}->[$1];
			    $coded =~ s/#$locvar\[($1)\]/$x/;
			    print "#$locvar"."[$1] replaced by $locvarhash->{$locvar}->[$1]". "-> $coded\n" if $locloudeb1;
		    }
		}
		elsif ((not ref $locvarhash->{$locvar}) and defined $coded and $coded =~ s/#$locvar/$locvarhash->{$locvar}/g  ) {
			print "#$locvar replaced by $locvarhash->{$locvar} -> $coded\n" if $locloudeb1;
		}
		elsif (ref $locvarhash->{$locvar}  and not ref $locvarhash->{$locvar} eq "ARRAY"){
		      while ( defined $coded and $coded =~ /#$locvar\$(\w+)/  ){
			      my $x = $locvarhash->{$locvar}->get_attribute($1);
			      $x = "[]" unless (defined $x);
			      $coded =~ s/#$locvar\$($1)/$x/;
			      print "#$locvar"."\$$1 replaced by $x". "-> $coded\n" if $locloudeb1;
		      }
		}
	}
	return ([("")]) unless (defined $coded);
	return ([("$coded")]) if ($coded =~ /^\w+$/);
	
	my $string = $coded;
	#resolve ^ \\ ~ "" undef arrays
	#translation into perl code
	$string =~ s/\[/&embed([/g;
	$string =~ s/\]/\])/g;
	my @quot;
	my $cc= 0;
	while ($string =~ s/"([^"]*)"/\$quot[$cc]/) {
	  push @quot, $1;
	  $cc++;
	}
	#$string =~ s/#var\s*\(\s*(\w+)\s*\)/\$setup->{$1}/g;
#	$string =~ s/#var\s*\(\s*(\w+)\s*\)/\$setup->{$1}/g;
	$string =~ s/\^/**/g;
	$string =~ s/#/&/g;
	while ($string =~ s/(^|\s|\(|\[|,|\)|\]|~|\})(\w|\w[\w\s]*\w)(\)|\s|$|,|\]|&|~)/$1'$2'$3/){};#\d+[_\pL]\w*|[_\pL]\w+
	$string =~  s/~/./g;
	while ($string =~ s/(\)|\]|'[^']*'|\})\s*('[^']*'|\(|&\w|\$)/$1.$2/){};
	
	print $string ."\n" if $locloudeb1;
	my $result = eval $string;
	croak $@."\n$string\n" if $@;
	if (not defined $result ) {
	    print "Result: Undefined\n" if $locloudeb1;;
	    return undef;
	}
	elsif (ref($result) eq 'EmbeddedArray') {
	  foreach my $i (0..$result->mi) {
	    print "List entry $i: ".$result->i($i)."\t".join("|",@{ $result->forvalues($i) })."\n" if $locloudeb1;;
	    push @rt, [($result->i($i), [-1, @{ $result->forvalues($i) }]) ];#
	  }
	}
	else {
	    print "Result: ". $result . "\n" if $locloudeb1;
	    $rt[0][0] = $result;
	}

	#print "Finished decoding string $rt[0][0] \n\n" unless (defined $rt[0][0]); # if $locloudeb1;
	return @rt; #rt has  two dimensions, but dimensions actually used depend on return value of sub interprete
					#scalar return -> only $rt[0][0] defined
						#return of field list -> $rt [..][0] field name, $rt[..][1..x] List of #thisfor values

}

	
sub protected_divide {
	my ($o2, $o1, $o3) = @_;
	return ( any_undef_in($o1, $o2)) ? undef : ($o1 == 0   ?  $o3 : ($o2 / $o1) );		
}

sub round_to {
	return sprintf("%.$_[1]f",$_[0]);
}	
sub floor {
	$_[0] =~ /^(\d+).?/;
	defined $1 ? $1 : 0;
}
sub steps {
    my ($start, $end, $step) = @_;
    my @array = map {$_ * $step } ($start..$end);
      return EmbeddedArray->new(\@array);
}	
sub normlike {
      return  1 / ( sqrt(2 * PI ) * $_[2] ) * exp ( -1 * ( $_[1] - $_[0]  ) ** 2 /  (2 * $_[2] ** 2));
}	
sub sql {
  my $get1 = $_[0];
  my $query =[];
  eval{	#exception handling to avoid fatals because table column does not exist 
	  #!!CT: revise as for now it captures all warnings, better solutions after restructuring of database
	  local $SIG{__WARN__}=sub{}; #Temporarily suppress warnings
	  $query = $dbh->selectcol_arrayref($get1);
  };
  if ($@) {	
	 croak "Query error captured\n$@\nQuery string was : $get1\n\n";
	 return undef; 
  }
  elsif (not defined $query->[0]) {
	  #print "Query resulted in NULL\n" if $::loudcprep;
	  $query->[0] = undef;
	  # print "\t Result (NULL) pushed to operands\n"  if $locloudeb1;
	  return  EmbeddedArray->new([]);		
  }
  elsif ($#{$query} == 0) {
	  #print "\t Result ($query->[0]) pushed to operands\n"  if $locloudeb1;
	  return EmbeddedArray->new($query);		
  }
  else {
    return EmbeddedArray->new($query);
  }
 }
sub if {
  $_[0] ? $_[1] : $_[2];
}
sub ifthenelse {
  $_[0] ? $_[1] : $_[2];
}	
sub replace_null {
   if (UNIVERSAL::isa( $_[0], "EmbeddedArray") ){
	$_[0] ? $_[0]:$_[1];
   } 
  else {
    defined $_[0] ? $_[0]:$_[1];
  }
}	
sub ifalone {
  $singlemode ? $_[0] : $_[1];
}
sub annuity {
    my ($anInterest, $anCapital, $anLifespan) = @_;
    return $anCapital * (( 1 + $anInterest)**$anLifespan *$anInterest) / ((1 +$anInterest)**$anLifespan -1);
}	
sub foreach {
    croak  "Error: Foreach list $_[0]  is not defined\n " unless (exists $setup->{$_[0]});
    print "Foreach: $_[0]\n" if $locloudeb1;
    return EmbeddedArray->new([])  if (not defined ref $setup->{$_[0]} or ref $setup->{$_[0]} eq "ARRAY" and scalar @{$setup->{$_[0]}} == 0 or not defined $setup->{$_[0]}->[0] );
    return EmbeddedArray->new([$setup->{$_[0]}]) if (not ref $setup->{$_[0]} eq "ARRAY");
    return EmbeddedArray->new( $setup->{$_[0]} );
}	
sub count {
      croak  "Error: Foreach list $_[0]  is not defined\n " unless (exists $setup->{$_[0]});
      my @list =  @{$setup->{$_[0]}} ;
      return scalar @list;
}	
sub random_number {
      croak  "Error: Argument ($_[0]) to random_number invalid\n " unless (looks_like_number($_[0]));
      return rand($_[0]);
}	
sub  listelement{
    croak  "Error: List $_[0]  is not defined or empty\n " unless $setup->{$_[0]};
    return $setup->{$_[0]}->[$_[1] -1]    ;	
}
sub var {
	croak  "Error: Variable $_[0] is  not defined\n " unless (exists $setup->{$_[0]});
	return	$setup->{$_[0]};
}
sub regexp {
	return $_[0] =~ $_[1];
}
sub list {
	my @list;
	if (UNIVERSAL::isa( $_[0], "EmbeddedArray")) { 
	    @list = $_[0]->ealist;
	}
	else {
	    @list = @{$setup->{$_[0]}};
	}
	my $qlist = join (", ", @list);
	return $qlist;	
}
sub qlist {
	my @list;
	if (UNIVERSAL::isa( $_[0], "EmbeddedArray")) { 
	    @list = $_[0]->ealist;
	}
	else {
	    @list = @{$setup->{$_[0]}};
	    croak  "Error: List $_[0] is  not defined or empty\n " unless (exists $setup->{$_[0]});
	}
	
	my $qlist = "'".join("', '", @list)  ."'"; 
	return $qlist;	
}
sub list_concat {
       my $first = shift @_;
       unless (UNIVERSAL::isa( $first, "EmbeddedArray")) {
	  $first = EmbeddedArray->new($setup->{$first} );
       }
       while (@_) {#
	  my $second = shift @_;
	  unless (UNIVERSAL::isa( $second, "EmbeddedArray")) {
			$second = EmbeddedArray->new($setup->{$second} );
	  }
	$first->merge($second);
	}
	return $first;
}
sub table {
	my ($tablename, $column, $keylist) = @_;
	my @keyarray = @{$keylist->{array}};
	if (any_undef_in($tablename, $column, @keyarray)) {
		return undef;
	}		
	else {
		#print "Table query : $tablename, $column, [".join(",", @keyarray)."]\n" if $locloudeb1;
		croak "Error: No tables (#table) defined in this environment!\n" unless (defined $tableregistry);
		my $tabtouse = $tableregistry->get_object_by_id($tablename);
		croak "Table (#table) $tablename unknown" unless ( defined $tabtouse);
		my $qresult =  [ $tabtouse->query(\@keyarray,$column) ];
		my $repl;
		
		if (not defined $qresult->[0]) {
			#using the array method because then the NA can be caught in the EXPANDARRAYS routine and the whole field value is set to 0, not creating a field even if there were pure string elements
			#print "Query resulted in NULL\n" if $locloudeb1;
			$repl = EmbeddedArray->new([]);
			#print scalar @keyarray."\t". $tabtouse->numkeys ."\t@keyarray\t@{$repl->{array}}" . "\n" if ($tablename eq "fieldtest" or $tablename eq "workforhire");
		}
		elsif ($#{$qresult} == 0 and scalar @keyarray == $tabtouse->numkeys) { #makes sure a scalar is only reported, when it was the intention to do so (otherwise 1 item results for list queries would not produce thisfor values
			$repl = $qresult->[0];
			#print scalar @keyarray."\t". $tabtouse->numkeys ."\t@keyarray\t$repl" . "\n" if ($tablename eq "fieldtest" or $tablename eq "workforhire");
		}
		else {
			$repl = EmbeddedArray->new($qresult);
		      #print scalar @keyarray."\t". $tabtouse->numkeys ."\t@keyarray\t@{$repl->{array}}" . "\n" if ($tablename eq "fieldtest" or $tablename eq "workforhire");	
		}
		  	
		#print "\t\tResult: $repl pushed to operands\n" if $locloudeb1;
		return $repl;
	} 
}
sub mapcell {
	my ($map, $x, $y) = @_;
	#print "Map query: $map, $x, $y:\n" if $locloudeb1;
	croak ("Error: No maps loaded\n") unless (defined $mapregistry);
	my $mo = $mapregistry->get_object_by_id($map);
	croak ("Error: Map $map unknown\n") unless (defined $mo);
	my $repl = $mo->get_cell($x,$y);
	#print "\t\tResult: $repl pushed to operands\n" if $locloudeb1;
	return $repl;
}
sub embed {
  return EmbeddedArray->new($_[0]);
}
sub handle_if_array2 {
	my $whattodo = shift;
	my @list_of_values = @_;
	my @newvalues;
	for my $i (@list_of_values) {
		my $new;
		while ($i =~ /\?(ta\d+)/){
			
			if ($whattodo eq "use_first") {
				my $arrayentry =  $TempArrays->{$1}->[0];
				delete $TempArrays->{$1};
				$i =~ s/\?$1/$arrayentry/;
				print "Warning: Found an array where a scalar was expected. Used only first entry!\n";
			}
			elsif ($whattodo eq "croak") {
				croak "Operator cannot handle an array!\n"
			}
		}
		push @newvalues, $i;
	}
	return @newvalues;
}
sub any_undef_in {
	foreach (@_)	{	
		return 1 if (not defined $_ );
	}
	return 0;
}
sub interprete_new {
	my ($class, $coded, $locvarhash, $is_translated, $be_verbose) = @_;
	croak "First argument to $class ->interprete should be string, not a reference!"  if (ref $coded);
	croak "Second argument to $class ->interprete should be a hash reference!"  unless (ref $locvarhash eq "HASH");
	$locloudeb1 = $verbose;

	my @rt; #declare return array
	return ([("")]) unless (defined $coded);

      
	#check for debugging switch 
	my $string;
	unless ($is_translated) {
	  ($string, $locloudeb1) = translate_to_perl($coded);
	}
	else {
	  $string = $coded;
	  $locloudeb1 = $be_verbose;
	}
	unset_this();
	set_this($locvarhash);
	print $string ."\n" if $locloudeb1;
	my $result = eval $string;
	croak $@."\n$string\n" if $@;
	if (not defined $result ) {
	    print "Result: Undefined\n" if $locloudeb1;;
	    return undef;
	}
	elsif (ref($result) eq 'EmbeddedArray') {
	  foreach my $i (0..$result->mi) {
	    print "List entry $i: ".$result->i($i)."\t".join("|",@{ $result->forvalues($i) })."\n" if $locloudeb1;;
	    push @rt, [($result->i($i), [-1, @{ $result->forvalues($i) }]) ];#
	  }
	}
	else {
	    print "Result: ". $result . "\n" if $locloudeb1;
	    $rt[0][0] = $result;
	}
	unset_this();
	#print "Finished decoding string $rt[0][0] \n\n" unless (defined $rt[0][0]); # if $locloudeb1;
	return @rt; #rt has  two dimensions, but dimensions actually used depend on return value of sub interprete
					#scalar return -> only $rt[0][0] defined
						#return of field list -> $rt [..][0] field name, $rt[..][1..x] List of #thisfor values

}
sub translate_to_perl {
	my $string = shift;
	my $verbose_switch;
	if ($string =~ s/@//g) {
		$verbose_switch = 1;				
	}
	$string =~ s/^\s*//;
	$string =~ s/\s*$//;

	#resolve ^ \\ ~ "" undef arrays
	#translation into perl code
	$string =~ s/\[/&embed([/g;
	$string =~ s/\]/\])/g;
	my @quot;
	my $cc= 0;
	while ($string =~ s/"([^"]*)"/\$quot[$cc]/) {
	  push @quot, $1;
	  $cc++;
	}
	$string =~ s/\^/**/g;
	$string =~ s/#/&/g;
	while ($string =~ s/(^|\s|\(|\[|,|\)|\]|~|\})(\w|\w[\w\s]*\w)(\)|\s|$|,|\]|&|~)/$1'$2'$3/){};#\d+[_\pL]\w*|[_\pL]\w+
	$string =~  s/~/./g;
	while ($string =~ s/(\)|\]|'[^']*'|\})\s*('[^']*'|\(|&\w|\$)/$1.$2/){};
	while ($string =~ s/\$quot\['?(\d+)'?\]/\"$quot[$1]\"/){};
	return ($string, $verbose_switch);
}
sub set_this {
    $locals_hash = shift;
}
sub unset_this {
    $locals_hash = undef;
}
sub this {
    my ($var, $index) = @_;
    croak "Error: Local variable $var undefined."  unless (exists $locals_hash->{$var});
    return $locals_hash->{$var} unless (ref $locals_hash->{$var});
    if (ref $locals_hash->{$var} eq "ARRAY") {
	  croak "Error: invalid index $index for local variable $var." unless ($index =~ /^\d+$/ and $index > 0 and $index <= $#{$locals_hash->{$var}});
	  return $locals_hash->{$var}->[$index];
    }
    return $locals_hash->{$var}->{$index};
}
1;
